/*
*  deform.c
*
*  Mesh deformations
*
*/
#include <stdlib.h>

#include <math.h>
#include "triangles.h"

#include <time.h>
#include "allegro.h" /* for matrix functions */

/*external, in triangles.c */
int error_exit(char *m1,char *m2);
int esc_pressed();

/* external, in mesh.c */
int new_edge(MESH *m,int v1,int v2, int f);
int new_triangle(MESH *m,int v1, int v2, int v3,int oflag);
int new_vertex(MESH *m,VTX vtx,int f);
int make_edge_list(MESH *m);
int try_orient(MESH *m);
int find_edge_in_meshedges(MESH *m,int va,int vb, int *o);
/****/
//#define TEST1
#define TEST2
#define TEST0
/* C version of John Van Sickle's POV smooth macro */
int SmoothSubdivide(MESH *m, float smoothing, int avg, int ignore_singular){
  /* Loop subdivision */
  int numV=m->num_vertices,
      numT=m->num_faces,
      numE;
  int newnumE,
      newnumV,
      newnumT=4*numT;

  VTX *EdgePoints;
  VECTOR *PointSum;  
  EDGE *NewEdges;
  TRIANGLE *TrianglesV=m->faces,
           *TrianglesE,*NewTriV,*NewTriE;
  int * PointCnt;
  TRIANGLE *T;  
  int e,i,j,iN,iT,iEa,iEb,o;
  float sA;
  int okedges,ret=1;
  int pA,pB,pC,pD,tA,tB;
  
  try_orient(m);

  okedges=make_edge_list(m);
  if(!okedges && !ignore_singular){
           return 0;
  }
  if(okedges<0){
           return 0;
  }

  numE=m->num_edges;
  newnumE=numE*2+numT*3;
  newnumV=numV+numE;
  if(newnumE >= m->num_edges_allocated-1){
       return -3;
  }

  if(newnumV >= m->num_vertices_allocated-1){
       return -2;
  }

  if(newnumT >= m->num_faces_allocated-1){
       return -1;
  }

  PointCnt=(int *) malloc(sizeof(int)*numV);
  TrianglesE=(TRIANGLE *)malloc(sizeof(TRIANGLE)*numT);
  NewTriV=(TRIANGLE *)malloc(sizeof(TRIANGLE)*newnumT);
  NewTriE=(TRIANGLE *)malloc(sizeof(TRIANGLE)*newnumT);
  EdgePoints=(VTX *)malloc(sizeof(VTX)*numE);
  PointSum=(VECTOR *)malloc(sizeof(VECTOR)*numV);
  NewEdges=(EDGE *)malloc(sizeof(EDGE)*newnumE);
  
  if(PointCnt==NULL || TrianglesE==NULL || NewTriV ==NULL || NewTriE ==NULL || EdgePoints==NULL || NewEdges==NULL || PointSum==NULL){
          return -4;
  }
  
  for(i=0;i<numT;i++){
      T=&(TrianglesV[i]);
      pA=T->v1;pB=T->v2;pC=T->v3;
      TrianglesE[i].v1=find_edge_in_meshedges(m,pA,pB,&o);
      TrianglesE[i].v2=find_edge_in_meshedges(m,pB,pC,&o);
      TrianglesE[i].v3=find_edge_in_meshedges(m,pC,pA,&o);
  }
  for(i=0;i<numV;i++){
       PointSum[i].x=0;PointSum[i].y=0;PointSum[i].z=0;
       PointCnt[i]=0;
  }
  for(i=0;i<numE;i++){
       pA=m->edges[i].v1;
       pB=m->edges[i].v2;
       /* add the end points to the running sum for the opposite end
          and remember how many have been added.*/
#ifdef TEST1
if((m->points[pA].flags&1  && m->points[pB].flags&1) ){
       PointSum[pA].x=PointSum[pA].x+m->points[pB].x;
       PointSum[pA].y=PointSum[pA].y+m->points[pB].y;
       PointSum[pA].z=PointSum[pA].z+m->points[pB].z;
       PointSum[pB].x=PointSum[pB].x+m->points[pA].x;
       PointSum[pB].y=PointSum[pB].y+m->points[pA].y;
       PointSum[pB].z=PointSum[pB].z+m->points[pA].z;
       PointCnt[pA]++; PointCnt[pB]++;
}
else if((m->points[pA].flags&1 ) ){
       PointSum[pB].x=PointSum[pB].x+m->points[pA].x;
       PointSum[pB].y=PointSum[pB].y+m->points[pA].y;
       PointSum[pB].z=PointSum[pB].z+m->points[pA].z;
       PointCnt[pB]++;
}
else if((m->points[pB].flags&1 ) ){
       PointSum[pA].x=PointSum[pA].x+m->points[pB].x;
       PointSum[pA].y=PointSum[pA].y+m->points[pB].y;
       PointSum[pA].z=PointSum[pA].z+m->points[pB].z;
       PointCnt[pA]++;
}
else{
#endif

       PointSum[pA].x=PointSum[pA].x+m->points[pB].x;
       PointSum[pA].y=PointSum[pA].y+m->points[pB].y;
       PointSum[pA].z=PointSum[pA].z+m->points[pB].z;
       PointSum[pB].x=PointSum[pB].x+m->points[pA].x;
       PointSum[pB].y=PointSum[pB].y+m->points[pA].y;
       PointSum[pB].z=PointSum[pB].z+m->points[pA].z;
       PointCnt[pA]++; PointCnt[pB]++;
#ifdef TEST1
}
#endif
       /* get the far vertices from the bordering triangles */
       tA=m->edges[i].t1;
       if(tA==-1){ tA=m->edges[i].t2;}
       tB=m->edges[i].t2;
       if(tB==-1){ tB=m->edges[i].t1;}
       pC=TrianglesV[tA].v1;
       if(pC==pA || pC==pB){
            pC=TrianglesV[tA].v2;
       }
       if(pC==pA || pC==pB){
            pC=TrianglesV[tA].v3;
       }
       pD=TrianglesV[tB].v1;
       if(pD==pA || pC==pB){
            pD=TrianglesV[tB].v2;
       }
       if(pD==pA || pD==pB){
            pD=TrianglesV[tB].v3;
       }
/***/
#ifdef TEST0
if((m->points[pA].flags&1  && m->points[pB].flags&1)&&avg){
       /*
       EdgePoints[i].x=(m->points[pA].x+m->points[pB].x)/2;

       EdgePoints[i].y=(m->points[pA].y+m->points[pB].y)/2;

       EdgePoints[i].z=(m->points[pA].z+m->points[pB].z)/2;
       */
       EdgePoints[i].x=(m->points[pA].x*3+m->points[pB].x*3
           +m->points[pA].x*smoothing+m->points[pB].x*smoothing)/(smoothing*2+6);
       EdgePoints[i].y=(m->points[pA].y*3+m->points[pB].y*3
           +m->points[pA].y*smoothing+m->points[pB].y*smoothing)/(smoothing*2+6);
       EdgePoints[i].z=(m->points[pA].z*3+m->points[pB].z*3
           +m->points[pA].z*smoothing+m->points[pB].z*smoothing)/(smoothing*2+6);

}
else{
/***/
#endif
       EdgePoints[i].x=(m->points[pA].x*3+m->points[pB].x*3
           +m->points[pC].x*smoothing+m->points[pD].x*smoothing)/(smoothing*2+6);
       EdgePoints[i].y=(m->points[pA].y*3+m->points[pB].y*3
           +m->points[pC].y*smoothing+m->points[pD].y*smoothing)/(smoothing*2+6);
       EdgePoints[i].z=(m->points[pA].z*3+m->points[pB].z*3
           +m->points[pC].z*smoothing+m->points[pD].z*smoothing)/(smoothing*2+6);
/***/
#ifdef TEST0
}
#endif
/***/
  }

  /* calculate the vertex points */
  for(i=0;i<numV;i++){
      iN=PointCnt[i];
      sA=.625-(cos(2*PI/iN)*2+3)*(cos(2*PI/iN)*2+3)*.015625;
      sA=(iN*(1-sA))/sA;
      PointSum[i].x=(m->points[i].x*sA+PointSum[i].x*smoothing)/(sA+iN*smoothing);
      PointSum[i].y=(m->points[i].y*sA+PointSum[i].y*smoothing)/(sA+iN*smoothing);
      PointSum[i].z=(m->points[i].z*sA+PointSum[i].z*smoothing)/(sA+iN*smoothing);
  }
  /* create all of the edges */
  i=0;j=0;
  while(i<numE){
      /*  P0 ---------------- P1  old edge
          P0 ------ NP ------ P1  new edges made from edge */
      NewEdges[j].v1=m->edges[i].v1;
      NewEdges[j].v2=i+numV;
      NewEdges[j].t1=-1;
      NewEdges[j].t2=-1;
      j++;
      NewEdges[j].v1=m->edges[i].v2;
      NewEdges[j].v2=i+numV;
      NewEdges[j].t1=-1;
      NewEdges[j].t2=-1;
      j++;
      i++;
  }
  iT=0;
  for(i=0;i<numT;i++){
      NewEdges[j].v1=TrianglesE[i].v1+numV;
      NewEdges[j].v2=TrianglesE[i].v3+numV;
      NewEdges[j].t1=-1;
      NewEdges[j].t2=-1;
      j++;
      NewEdges[j].v1=TrianglesE[i].v2+numV;
      NewEdges[j].v2=TrianglesE[i].v1+numV;
      NewEdges[j].t1=-1;
      NewEdges[j].t2=-1;
      j++;
      NewEdges[j].v1=TrianglesE[i].v3+numV;
      NewEdges[j].v2=TrianglesE[i].v2+numV;
      NewEdges[j].t1=-1;
      NewEdges[j].t2=-1;
      j++;
      /* create interior triangle */
      NewTriV[iT].v1=TrianglesE[i].v1+numV;
      NewTriV[iT].v2=TrianglesE[i].v2+numV;
      NewTriV[iT].v3=TrianglesE[i].v3+numV;
      NewTriE[iT].v1=j-2;
      NewTriE[iT].v2=j-1;
      NewTriE[iT].v3=j-3;
      NewEdges[j-1].t1=iT;
      NewEdges[j-2].t1=iT;
      NewEdges[j-3].t1=iT;
      iT++;
      /* create triangle at first vertex */
      iEa=TrianglesE[i].v1*2;
      if(NewEdges[iEa].v1!=TrianglesV[i].v1)
          iEa++;
      iEb=TrianglesE[i].v3*2;
      if(NewEdges[iEb].v1!=TrianglesV[i].v1)
          iEb++;
      NewTriV[iT].v1=TrianglesV[i].v1;
      NewTriV[iT].v2=TrianglesE[i].v1+numV;
      NewTriV[iT].v3=TrianglesE[i].v3+numV;

      NewTriE[iT].v1=iEa;
      NewTriE[iT].v2=j-3;
      NewTriE[iT].v3=iEb;

      NewEdges[j-3].t2=iT;
      if (NewEdges[iEa].t1==-1)
          NewEdges[iEa].t1=iT;
      else
          NewEdges[iEa].t2=iT;
      if (NewEdges[iEb].t1==-1)
          NewEdges[iEb].t1=iT;
      else
          NewEdges[iEb].t2=iT;
      iT++;
     /* create triangle at second vertex */
     iEa=TrianglesE[i].v2*2;
     if(NewEdges[iEa].v1 != TrianglesV[i].v2)
            iEa++;
     iEb=TrianglesE[i].v1*2;
     if(NewEdges[iEb].v1 != TrianglesV[i].v2)
            iEb++;

     NewTriV[iT].v1=TrianglesV[i].v2;
     NewTriV[iT].v2=TrianglesE[i].v2+numV;
     NewTriV[iT].v3=TrianglesE[i].v1+numV;

     NewTriE[iT].v1=iEa;
     NewTriE[iT].v2=j-2;
     NewTriE[iT].v3=iEb;
     NewEdges[j-2].t2=iT;
     if (NewEdges[iEa].t1==-1)
         NewEdges[iEa].t1=iT;
     else
         NewEdges[iEa].t2=iT;
    if (NewEdges[iEb].t1==-1)
         NewEdges[iEb].t1=iT;
    else
         NewEdges[iEb].t2=iT;
    iT++;
     /* create triangle at third vertex */
    iEa=TrianglesE[i].v3*2;
    if(NewEdges[iEa].v1!=TrianglesV[i].v3)
        iEa++;
    iEb=TrianglesE[i].v2*2;
    if(NewEdges[iEb].v1!=TrianglesV[i].v3)
        iEb++;

    NewTriV[iT].v1=TrianglesV[i].v3;
    NewTriV[iT].v2=TrianglesE[i].v3+numV;
    NewTriV[iT].v3=TrianglesE[i].v2+numV;

    NewTriE[iT].v1=iEa;
    NewTriE[iT].v2=j-1;
    NewTriE[iT].v3=iEb;

    NewEdges[j-1].t2=iT;
    if (NewEdges[iEa].t1==-1)
       NewEdges[iEa].t1=iT;
    else
       NewEdges[iEa].t2=iT;

    if (NewEdges[iEb].t1==-1)
       NewEdges[iEb].t1=iT;
    else
       NewEdges[iEb].t2=iT;
    iT++;

  }
  /* All triangles are generated;
     copy the data over */

  newnumV=numV+numE;

  i=0;
  while(i<numV){
#ifdef TEST2
    if(!(m->points[i].flags & 1)){
#endif
        m->points[i].x=PointSum[i].x;
        m->points[i].y=PointSum[i].y;
        m->points[i].z=PointSum[i].z;
#ifdef TEST2
    }
#endif
    i++;
  }
  while (i<newnumV){
       new_vertex(m,EdgePoints[i-numV],0);
       i++;
  }
  m->num_edges=0;
  for(i=0;i<newnumE;i++){
      e=new_edge(m,NewEdges[i].v1,NewEdges[i].v2,0);
      if(e<0){
         ret=2;
         break;
      }
      m->edges[e].t1=NewEdges[i].t1;
      m->edges[e].t2=NewEdges[i].t2;
  }
  m->num_faces=0;
  for(i=0;i<newnumT;i++){
      new_triangle(m,NewTriV[i].v1,NewTriV[i].v2,NewTriV[i].v3,0);

  }
  
  free(PointCnt);
  free(TrianglesE);
  free(NewTriV);
  free(NewTriE);
  free(PointSum);
  free(NewEdges);
  free(EdgePoints);
  return ret;
}


/* C version of John Van Sickle's POV "roughen" macro */
int perturb_mesh(MESH *m, float N,int Seed){
    int i;
    int pA,pB,pC;
    TRIANGLE *T;
    VTX vA,vB,vC;
    VECTOR *Norm;
    float *Area;
    float sA,sN;
    VECTOR vN,nvN;

    Norm=(VECTOR *)malloc(sizeof(VECTOR) * m->num_vertices);
    Area=(float *)malloc(sizeof(float) * m->num_vertices);
    if(Norm==NULL || Area==NULL){
        return 0;
    }
    
    for(i=0;i<m->num_vertices;i++){
         Norm[i].x=0;Norm[i].y=0;Norm[i].z=0;
         Area[i]=0.0;
    }

    if(Seed==-1){
       srand(time(0));
    }
    else srand((unsigned)Seed);

    for (i=0;i<m->num_faces;i++){
         T=&(m->faces[i]);
         pA=T->v1; pB=T->v2; pC=T->v3;
         vA=m->points[pA];vB=m->points[pB];vC=m->points[pC];
         cross_product_f(vB.x-vA.x,vB.y-vA.y,vB.z-vA.z, vC.x-vA.x,vC.y-vA.y,vC.z-vA.z, &(vN.x),&(vN.y),&(vN.z));
         nvN.x=vN.x; nvN.y=vN.y; nvN.z=vN.z;
         if(nvN.x!=0 || nvN.y!=0 || nvN.z!=0){
            normalize_vector_f(&(nvN.x),&(nvN.y),&(nvN.z));
         }
         Norm[pA].x += nvN.x; Norm[pA].y += nvN.y; Norm[pA].z += nvN.z;
         Norm[pB].x += nvN.x; Norm[pB].y += nvN.y; Norm[pB].z += nvN.z;
         Norm[pC].x += nvN.x; Norm[pC].y += nvN.y; Norm[pC].z += nvN.z;
         sA=vector_length_f(vN.x,vN.y,vN.z)/2.0;
         Area[pA] += sA; Area[pB] += sA; Area[pC] += sA;
    }
    for(i=0;i<m->num_vertices;i++){
         if(m->points[i].flags & 1) continue;
         sN=(N)*sqrt(Area[i])*(RND - RND);
         vN.x=Norm[i].x;vN.y=Norm[i].y;vN.z=Norm[i].z;
         if(vN.x!=0 || vN.y!=0 || vN.z!=0){
             normalize_vector_f(&(vN.x),&(vN.y),&(vN.z));
         }
         m->points[i].x += sN*vN.x;
         m->points[i].y += sN*vN.y;
         m->points[i].z += sN*vN.z;
    }
    free(Norm);
    free(Area);
    return 1;
}


/* Allegro/C versions of four Chris Colefax POV macros */
void vtx_wave(VTX *v, float Frequency, float Phase, float WaveScale){
    v->x *= (1+sin(v->y*2*PI*Frequency+Phase)*WaveScale);
    v->z *= (1+sin(v->y*2*PI*Frequency+Phase)*WaveScale);
}

int wave_mesh(MESH *m, float Frequency, float Phase, float WaveScale){
    int vi;
    VTX *v;
    for(vi=0;vi<m->num_vertices;vi++){
       v=&(m->points[vi]);
       vtx_wave(v, Frequency, Phase, WaveScale);
    }
    return 1;
}

void vtx_ripple(VTX *v, VECTOR RippleAxis, float Frequency,float Phase, VECTOR RippleScale){
    VECTOR ripple;
    float rx,ry,rz,rl;
    
    rx=v->x*RippleAxis.x;
    ry=v->y*RippleAxis.y;
    rz=v->z*RippleAxis.z;
    rl=vector_length_f(rx,ry,rz);
    ripple.x=sin(rl*Frequency+Phase)*RippleScale.x;
    ripple.y=sin(rl*Frequency+Phase)*RippleScale.y;
    ripple.z=sin(rl*Frequency+Phase)*RippleScale.z;
    v->x+=ripple.x;
    v->y+=ripple.y;
    v->z+=ripple.z;
    
}
int ripple_mesh(MESH *m, VECTOR RippleAxis, float Frequency,float Phase, VECTOR RippleAmount){
    int vi;
    VTX *v;

    for(vi=0;vi<m->num_vertices;vi++){
       v=&(m->points[vi]);
       vtx_ripple(v, RippleAxis, Frequency,Phase, RippleAmount);
    }
    return 1;
}



#define DEG2FIX(d) (d*128/180)
void vtx_twist(VTX *v,float TwistAngle, VTX TwistCentre){
    float vx,vy,vz;
    MATRIX_f M;
    float tw_anglex,tw_angley,tw_anglez;
    
    v->x -= TwistCentre.x;
    v->y -= TwistCentre.y;
    v->z -= TwistCentre.z;
    tw_anglex=v->x*TwistAngle;
    tw_angley=v->y*TwistAngle;
    tw_anglez=v->z*TwistAngle;
    get_rotation_matrix_f(&M, DEG2FIX(tw_anglex), DEG2FIX(tw_angley), DEG2FIX(tw_anglez));
    apply_matrix_f(&M,v->x,v->y,v->z,&vx,&vy,&vz);
    v->x = vx+ TwistCentre.x;
    v->y = vy+ TwistCentre.y;
    v->z = vz+ TwistCentre.z;
}
int twist_mesh(MESH *m,float TwistAngle, VTX TwistCentre){
    int vi;
    VTX *v;
    for(vi=0;vi<m->num_vertices;vi++){
       v=&(m->points[vi]);
       vtx_twist(v,TwistAngle,TwistCentre);
    }
    return 1;
}

void vtx_twist_about(VTX *v,float TwistAngle, VTX TwistCentre,VECTOR TwistAxis){

    MATRIX_f M;

    float tw_angle;
    float vx,vy,vz;
    float d;
    vx = v->x - TwistCentre.x;
    vy = v->y - TwistCentre.y;
    vz = v->z - TwistCentre.z;
    d=dot_product_f(TwistAxis.x,TwistAxis.y,TwistAxis.z,vx,vy,vz);
    
    tw_angle=d*TwistAngle;

    get_vector_rotation_matrix_f(&M, TwistAxis.x,TwistAxis.y,TwistAxis.z, DEG2FIX(tw_angle));

    apply_matrix_f(&M,vx,vy,vz,&(v->x),&(v->y),&(v->z));
}
int twist_mesh_about(MESH *m,float TwistAngle, VTX TwistCentre,VECTOR TwistAxis){
    int vi;
    VTX *v;
    for(vi=0;vi<m->num_vertices;vi++){
       v=&(m->points[vi]);
       vtx_twist_about(v,TwistAngle,TwistCentre,TwistAxis);
    }
    return 1;
}

void vtx_bend(VTX *v,VTX BendPoint,VECTOR BX,float BA){
    float  vx,vy,vz,vlx,vly,vlz,A;
    MATRIX_f M;
    
    v->x-=BendPoint.x;
    v->y-=BendPoint.y;
    v->z-=BendPoint.z;
    vlx=v->x*(1-BX.x);
    vly=v->y*(1-BX.y);
    vlz=v->z*(1-BX.z);
    A=vector_length_f(vlx,vly,vlz)*BA;
    get_vector_rotation_matrix_f(&M, BX.x, BX.y, BX.z, DEG2FIX(A));
    apply_matrix_f(&M,v->x,v->y,v->z,&vx,&vy,&vz);
    v->x=vx+BendPoint.x;
    v->y=vy+BendPoint.y;
    v->z=vz+BendPoint.z;
}
int bend_mesh(MESH *m,VTX BendPoint,VECTOR BX,float BA){
    int vi;
    VTX *v;

    for(vi=0;vi<m->num_vertices;vi++){
       v=&(m->points[vi]);
       vtx_bend(v,BendPoint,BX,BA);
    }
    return 1;
}

void vtx_repulse_point(VTX *v,VTX RepulsePoint, float c,float a, float b){
    float  vx,vy,vz;
    float dist,exponent,mult;
    
    vx=v->x - RepulsePoint.x;
    vy=v->y - RepulsePoint.y;
    vz=v->z - RepulsePoint.z;
    dist=vector_length_f(vx,vy,vz);
    exponent=-a*pow(dist,b);
    mult=c*exp(exponent);
    v->x+= mult*vx;
    v->y+= mult*vy;
    v->z+= mult*vz;
}
int repulse_point_mesh(MESH *m,VTX RepulsePoint, float c,float a, float b){

    int vi;
    VTX *v;

    for(vi=0;vi<m->num_vertices;vi++){
       v=&(m->points[vi]);
       vtx_repulse_point(v,RepulsePoint,c, a, b);
    }
    return 1;
}



