/*
*  files.c
*
*  Loading and saving meshes
*
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "allegro.h"
#include "triangles.h"
/* extern triangles.c */
void waiting_call();
void waiting_done();

int make_edge_list(MESH *m);
int try_orient(MESH *m);
int find_edge_in_meshedges(MESH *m,int va,int vb, int *o);

int new_triangle(MESH *m,int v1, int v2, int v3,int oflag);
int new_triangle_o(MESH *m,int v1, int v2, int v3);
int new_edge(MESH *m,int v1,int v2, int f);
int new_vertex(MESH *m,VTX vtx,int f);
void add_to_selected (int vtxnum, int removeiffound);
extern int *curVertex;
extern int num_selected;
/**/

/* external, in mesh.c */
MESH * copy_of_connected_component(MESH *m,int vtx);
void delete_mesh(MESH *m);
void get_mesh_extent(MESH *m, float *max_x,float *max_y,float *max_z,float *min_x,float *min_y,float *min_z);


#define DBG 220
DIALOG save_POV_RAW_dialog[]={
/* (dialog proc)     (x)   (y)   (w)   (h)     (fg)  (bg)  (key)    (flags)           (d1)  (d2)  (dp) */
  { d_box_proc,        0,   0,   320, 150,      0,   DBG,    0,         0,              0,    0,   NULL,NULL,NULL},
  
  { d_check_proc,     10,  30,    300,  10,      0,   DBG,    0,D_SELECTED,              0,    0,    "Save components as separate meshes" ,NULL,NULL},
  { d_check_proc,     10,  60,    150,  10,      0,   DBG,    0,D_SELECTED,              0,    0,    "Include texture" ,NULL,NULL},
  { d_check_proc,     10,  90,    300,  10,      0,   DBG,    0,D_SELECTED,              0,    0,    "Include #declared mesh extents" ,NULL,NULL},
  { d_button_proc,    10,  120,    20,  20,      0,   255,   13,    D_EXIT,              0,    0,   "OK",NULL,NULL},
  { d_button_proc,    40,  120,    60,  20,      0,   255,    0,    D_EXIT,              0,    0,   "Cancel",NULL,NULL},
  
  { NULL,              0,    0,    0,   0,      0,     0,    0,         0,              0,    0,    NULL ,NULL,NULL}
};
#define SAVEPOVSEPARATECOMPONENTSCHECK 1
#define SAVEPOVTEXTURECHECK 2
#define SAVEPOVEXTENTSCHECK 3
#define SAVEPOVCANCEL 5


int write_POV_triangles(MESH* m,FILE *fp,VECTOR *VNormals){
     int i;
     TRIANGLE *T;
     char outstring[512];
     float x1,y1,z1,x2,y2,z2,x3,y3,z3;
     float nx1,ny1,nz1,nx2,ny2,nz2,nx3,ny3,nz3;
     for (i=0;i<m->num_faces;i++){
        T=&(m->faces[i]);
        x1=T->vtxlist[T->v1].x;
        y1=T->vtxlist[T->v1].y;
        z1=T->vtxlist[T->v1].z;
        x2=T->vtxlist[T->v2].x;
        y2=T->vtxlist[T->v2].y;
        z2=T->vtxlist[T->v2].z;
        x3=T->vtxlist[T->v3].x;
        y3=T->vtxlist[T->v3].y;
        z3=T->vtxlist[T->v3].z;
        
        if(VNormals){
           nx1=VNormals[T->v1].x;
           ny1=VNormals[T->v1].y;
           nz1=VNormals[T->v1].z;
           nx2=VNormals[T->v2].x;
           ny2=VNormals[T->v2].y;
           nz2=VNormals[T->v2].z;
           nx3=VNormals[T->v3].x;
           ny3=VNormals[T->v3].y;
           nz3=VNormals[T->v3].z;
           sprintf(outstring,"smooth_triangle{<%f,%f,%f>,<%f,%f,%f>,<%f,%f,%f>,<%f,%f,%f>,<%f,%f,%f>,<%f,%f,%f>}\n",x1,-y1,z1,nx1,-ny1,nz1, x2,-y2,z2,nx2,-ny2,nz2,x3,-y3,z3,nx3,-ny3,nz3);
        }
        else{
           sprintf(outstring,"triangle{<%f,%f,%f>,<%f,%f,%f>,<%f,%f,%f>}\n",x1,-y1,z1,x2,-y2,z2,x3,-y3,z3);
        }
        fprintf(fp,outstring);
     }
     return 1;

}

int separate_components=1;
int include_textures=1;
int declare_extents=1;

int save_POV_triangle_mesh(MESH* m,char *filename,VECTOR *VNormals){
     int i;
     int savedialogret;
     int c,component_number=1;
     FILE *fp;
     MESH *cmesh;
     char message1[256];
     //char outstring[256];
     float maxx,maxy,maxz,minx,miny,minz;
     save_POV_RAW_dialog[SAVEPOVTEXTURECHECK].flags&=~D_HIDDEN;
     save_POV_RAW_dialog[SAVEPOVEXTENTSCHECK].flags&=~D_HIDDEN;
     centre_dialog(save_POV_RAW_dialog);
     savedialogret=popup_dialog(save_POV_RAW_dialog,-1);
     if(savedialogret==SAVEPOVCANCEL)
         return 0;

     separate_components=save_POV_RAW_dialog[SAVEPOVSEPARATECOMPONENTSCHECK].flags&D_SELECTED;
     include_textures=save_POV_RAW_dialog[SAVEPOVTEXTURECHECK].flags&D_SELECTED;
     declare_extents=save_POV_RAW_dialog[SAVEPOVEXTENTSCHECK].flags&D_SELECTED;
     


     fp=fopen(filename,"wt");
     if(fp==NULL)return -1;
     waiting_call();
     if(separate_components){
         for(i=0;i<m->num_vertices;i++){
              m->points[i].flags&=~2;
         }
         c=0;
         component_number=0;
         while(c<m->num_vertices){
           if(!(m->points[c].flags&2)){
             cmesh=copy_of_connected_component(m,c);
             if(cmesh==NULL){
               break;

             }
             if(cmesh->num_faces>0){

                if(declare_extents){
                    get_mesh_extent(cmesh, &maxx,&maxy,&maxz,&minx,&miny,&minz);
                    fprintf(fp,"#declare COMPONENT%d_MIN_X=%f;\n",component_number+1,minx);
                    fprintf(fp,"#declare COMPONENT%d_MIN_Y=%f;\n",component_number+1,-miny);
                    fprintf(fp,"#declare COMPONENT%d_MIN_Z=%f;\n",component_number+1,minz);
                    fprintf(fp,"#declare COMPONENT%d_MAX_X=%f;\n",component_number+1,maxx);
                    fprintf(fp,"#declare COMPONENT%d_MAX_Y=%f;\n",component_number+1,-maxy);
                    fprintf(fp,"#declare COMPONENT%d_MAX_Z=%f;\n",component_number+1,maxz);
                }
                fprintf(fp,"mesh{\n");
       
                write_POV_triangles(cmesh,fp,VNormals);
                if(include_textures)
                   fprintf(fp,"texture{TEXTURE%d}\n",component_number+1);
                fprintf(fp,"}\n");
                component_number++;
             }
             delete_mesh(cmesh);cmesh=NULL;

           }
           c++;
         }
         for(i=0;i<m->num_vertices;i++){
              m->points[i].flags&=~2;
         }
     }
     else{
             if(declare_extents){
                 fprintf(fp,"#declare COMPONENT%d_MIN_X=%f;\n",component_number+1,minx);
                 fprintf(fp,"#declare COMPONENT%d_MIN_Y=%f;\n",component_number+1,-miny);
                 fprintf(fp,"#declare COMPONENT%d_MIN_Z=%f;\n",component_number+1,minz);
                 fprintf(fp,"#declare COMPONENT%d_MAX_X=%f;\n",component_number+1,maxx);
                 fprintf(fp,"#declare COMPONENT%d_MAX_Y=%f;\n",component_number+1,-maxy);
                 fprintf(fp,"#declare COMPONENT%d_MAX_Z=%f;\n",component_number+1,maxz);
             }
             fprintf(fp,"mesh{\n");

             write_POV_triangles(m,fp,VNormals);
             if(include_textures)
               fprintf(fp,"texture{TEXTURE%d}\n",component_number+1);
             fprintf(fp,"}\n");

     }
     fclose(fp);
     waiting_done();
     if(separate_components){
          sprintf(message1,"Saved %d component%c",component_number,component_number==1?0:'s');
     }
     else
          sprintf(message1,"Saved mesh");
     alert(message1,NULL,NULL,"OK",NULL, 13, 0);

     return 0;
}


int write_RAW_triangles(MESH* m,FILE *fp, VECTOR *VNormals){
     TRIANGLE *T;
     int i;
     char outstring[512];
     float x1,y1,z1,x2,y2,z2,x3,y3,z3;
     float nx1,ny1,nz1,nx2,ny2,nz2,nx3,ny3,nz3;
     for (i=0;i<m->num_faces;i++){
        T=&(m->faces[i]);
        x1=T->vtxlist[T->v1].x;
        y1=T->vtxlist[T->v1].y;
        z1=T->vtxlist[T->v1].z;
        x2=T->vtxlist[T->v2].x;
        y2=T->vtxlist[T->v2].y;
        z2=T->vtxlist[T->v2].z;
        x3=T->vtxlist[T->v3].x;
        y3=T->vtxlist[T->v3].y;
        z3=T->vtxlist[T->v3].z;
        
        if(VNormals){
           nx1=VNormals[T->v1].x;
           ny1=VNormals[T->v1].y;
           nz1=VNormals[T->v1].z;
           nx2=VNormals[T->v2].x;
           ny2=VNormals[T->v2].y;
           nz2=VNormals[T->v2].z;
           nx3=VNormals[T->v3].x;
           ny3=VNormals[T->v3].y;
           nz3=VNormals[T->v3].z;
           sprintf(outstring,"%f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f %f}\n",x1,-y1,z1,x2,-y2,z2,x3,-y3,z3,nx1,-ny1,nz1,nx2,-ny2,nz2,nx3,-ny3,nz3);
        }
        else
           sprintf(outstring,"%f %f %f %f %f %f %f %f %f\n",x1,-y1,z1,x2,-y2,z2,x3,-y3,z3);
        fprintf(fp,outstring);
     }

     return 1;
}
int save_RAW_triangle_mesh(MESH* m,char *filename, VECTOR *VNormals){
     int i;
     int savedialogret;
     int c,component_number=1;
     FILE *fp;
     MESH *cmesh;
     char message1[256];

     save_POV_RAW_dialog[SAVEPOVTEXTURECHECK].flags|=D_HIDDEN;
     save_POV_RAW_dialog[SAVEPOVEXTENTSCHECK].flags|=D_HIDDEN;
     centre_dialog(save_POV_RAW_dialog);
     savedialogret=popup_dialog(save_POV_RAW_dialog,-1);
     if(savedialogret==SAVEPOVCANCEL)
         return 0;

     separate_components=save_POV_RAW_dialog[SAVEPOVSEPARATECOMPONENTSCHECK].flags&D_SELECTED;


     fp=fopen(filename,"wt");
     if(fp==NULL)return -1;

     waiting_call();
     if(separate_components){
         for(i=0;i<m->num_vertices;i++){
              m->points[i].flags&=~2;
         }
         c=0;
         component_number=0;
         while(c<m->num_vertices){
           if(!(m->points[c].flags&2)){
             cmesh=copy_of_connected_component(m,c);
             if(cmesh==NULL){
               break;

             }
             fprintf(fp,"COMPONENT%d\n",component_number+1);
    
             write_RAW_triangles(cmesh,fp,VNormals);

             fprintf(fp,"\n");
             delete_mesh(cmesh);cmesh=NULL;
             component_number++;
           }
           c++;
         }
         for(i=0;i<m->num_vertices;i++){
              m->points[i].flags&=~2;
         }
     }
     else{


             write_RAW_triangles(m,fp,VNormals);



     }
     waiting_done();
     fclose(fp);

     if(separate_components){
          sprintf(message1,"Saved %d component%c",component_number,component_number==1?0:'s');
     }
     else
          sprintf(message1,"Saved mesh");
     alert(message1,NULL,NULL,"OK",NULL, 13, 0);


     return 0;
}

/*
Mesh2
Author: Nathan Kopp e-mail: <Nathan@kopp.com>

URL: http://nathan.kopp.com/patched.htm

The new mesh syntax is designed for use in conversion from other file formats. This format was developed by Nathan Kopp and Thomas Baier.

Syntax
mesh2 {
  vertex_vectors {
    number_of_vertices,
    <vertex1>, <vertex2>, ...
  }
  normal_vectors {
    number_of_normals,
    <normal1>, <normal2>, ...
  }
  uv_vectors {
    number_of_uv_vectors,
    <uv_vect1>, <uv_vect2>, ...
  }
  texture_list {
    number_of_textures,
    texture { Texture1 },
    texture { Texture2 }, ...
  }
  face_indices {
    number_of_faces,
    <index_a, index_b, index_c> [,texture_index [, texture_index, texture_index]],
    <index_d, index_e, index_f> [,texture_index [, texture_index, texture_index]],
    ...
  }
  normal_indices {
    number_of_faces,
    <index_a, index_b, index_c>,
    <index_d, index_e, index_f>,
    ...
  }
  uv_indices {
    number_of_faces,
    <index_a, index_b, index_c>,
    <index_d, index_e, index_f>,
    ...
  }
  [object modifiers]
}
The normal_vectors, uv_vectors, and texture_list sections are optional. If the number of normals equals the number of vertices then the normal_indices section is optional and the indexes from the face_indices section are used instead. Likewise for the uv_indices section.

The indexes are ZERO-BASED! So the first item in each list has an index of zero.


mesh2
{
  normal_vectors{
    6
    <-1,0,0>,<0,-1,0>,<0,0,-1>
    < 1,0,0>,<0, 1,0>,<0,0, 1>
  }  
  vertex_vectors{
    6
    <-1,0,0>,<0,-1,0>,<0,0,-1>
    < 1,0,0>,<0, 1,0>,<0,0, 1>
  }

  face_indices{
    8
    <0,2,4>,<3,5,4>,<0,3,4>,<3,2,4>
    <0,2,1>,<3,5,1>,<0,3,1>,<3,2,1>
  }
  normal_indices{
    8
    <0,2,4>,<3,5,4>,<0,3,4>,<3,2,4>
    <0,2,1>,<3,5,1>,<0,3,1>,<3,2,1>
  }
  texture {pigment {radial frequency 8} finish{specular 1}}
}
*/
int write_mesh2(MESH* m,FILE *fp,VECTOR *VNormals){
     int i;
     TRIANGLE T;
     VTX v;
     VECTOR N;
     if(VNormals){
        fprintf(fp," normal_vectors{\n  %d,\n   ",m->num_vertices);
        for(i=0;i<m->num_vertices;i++){
            N=VNormals[i];
            fprintf(fp,"<%f,%f,%f>",N.x,-N.y,N.z);
            if(i<m->num_vertices-1){
               fprintf(fp,", ");
            }
   
            if((i%6)==5){
               fprintf(fp,"\n   ");
            }
        }
        fprintf(fp," }\n");

     }

     fprintf(fp," vertex_vectors{\n  %d,\n   ",m->num_vertices);
     for(i=0;i<m->num_vertices;i++){
         v=m->points[i];
         fprintf(fp,"<%f,%f,%f>",v.x,-v.y,v.z);
         if(i<m->num_vertices-1){
            fprintf(fp,", ");
         }

         if((i%6)==5){
            fprintf(fp,"\n   ");
         }
     }
     fprintf(fp," }\n");

     fprintf(fp," face_indices{\n  %d,\n   ",m->num_faces);
     for (i=0;i<m->num_faces;i++){
         T=m->faces[i];
         fprintf(fp,"<%d,%d,%d>",T.v1,T.v2,T.v3);
         if(i<m->num_faces-1){
            fprintf(fp,", ");
         }
         if((i%6)==5){
            fprintf(fp,"\n");
            if(i<m->num_faces-1){
               fprintf(fp,"   ");
            }
         }

     }
     fprintf(fp," }\n");

     return 1;

}

int save_mesh2(MESH* m,char *filename,VECTOR *VNormals){
     int i;
     int savedialogret;
     int c,component_number=1;
     FILE *fp;
     MESH *cmesh;
     char message1[256];

     float maxx,maxy,maxz,minx,miny,minz;

     save_POV_RAW_dialog[SAVEPOVTEXTURECHECK].flags&=~D_HIDDEN;
     save_POV_RAW_dialog[SAVEPOVEXTENTSCHECK].flags&=~D_HIDDEN;
     centre_dialog(save_POV_RAW_dialog);
     savedialogret=popup_dialog(save_POV_RAW_dialog,-1);
     if(savedialogret==SAVEPOVCANCEL)
         return 0;

     separate_components=save_POV_RAW_dialog[SAVEPOVSEPARATECOMPONENTSCHECK].flags&D_SELECTED;
     include_textures=save_POV_RAW_dialog[SAVEPOVTEXTURECHECK].flags&D_SELECTED;
     declare_extents=save_POV_RAW_dialog[SAVEPOVEXTENTSCHECK].flags&D_SELECTED;



     fp=fopen(filename,"wt");
     if(fp==NULL)return -1;
     waiting_call();
     fprintf(fp,"#version unofficial MegaPOV 0.4;\n");
     if(separate_components){
         for(i=0;i<m->num_vertices;i++){
              m->points[i].flags&=~2;
         }
         c=0;
         component_number=0;
         while(c<m->num_vertices){
           if(!(m->points[c].flags&2)){
             cmesh=copy_of_connected_component(m,c);
             if(cmesh==NULL){
               break;

             }
             if(cmesh->num_faces>0){

                if(declare_extents){
                    get_mesh_extent(cmesh, &maxx,&maxy,&maxz,&minx,&miny,&minz);
                    fprintf(fp,"#declare COMPONENT%d_MIN_X=%f;\n",component_number+1,minx);
                    fprintf(fp,"#declare COMPONENT%d_MIN_Y=%f;\n",component_number+1,-miny);
                    fprintf(fp,"#declare COMPONENT%d_MIN_Z=%f;\n",component_number+1,minz);
                    fprintf(fp,"#declare COMPONENT%d_MAX_X=%f;\n",component_number+1,maxx);
                    fprintf(fp,"#declare COMPONENT%d_MAX_Y=%f;\n",component_number+1,-maxy);
                    fprintf(fp,"#declare COMPONENT%d_MAX_Z=%f;\n",component_number+1,maxz);
                }
                fprintf(fp,"mesh2{\n");
       
                write_mesh2(cmesh,fp,VNormals);
                if(include_textures)
                   fprintf(fp,"texture{TEXTURE%d}\n",component_number+1);
                fprintf(fp,"}\n");
                component_number++;
             }
             delete_mesh(cmesh);cmesh=NULL;

           }
           c++;
         }
         for(i=0;i<m->num_vertices;i++){
              m->points[i].flags&=~2;
         }
     }
     else{
             if(declare_extents){
                 fprintf(fp,"#declare COMPONENT%d_MIN_X=%f;\n",component_number+1,minx);
                 fprintf(fp,"#declare COMPONENT%d_MIN_Y=%f;\n",component_number+1,-miny);
                 fprintf(fp,"#declare COMPONENT%d_MIN_Z=%f;\n",component_number+1,minz);
                 fprintf(fp,"#declare COMPONENT%d_MAX_X=%f;\n",component_number+1,maxx);
                 fprintf(fp,"#declare COMPONENT%d_MAX_Y=%f;\n",component_number+1,-maxy);
                 fprintf(fp,"#declare COMPONENT%d_MAX_Z=%f;\n",component_number+1,maxz);
             }
             fprintf(fp,"mesh2{\n");

             write_mesh2(m,fp,VNormals);
             if(include_textures)
               fprintf(fp,"texture{TEXTURE%d}\n",component_number+1);
             fprintf(fp,"}\n");

     }
     waiting_done();
     fclose(fp);

     if(separate_components){
          sprintf(message1,"Saved %d component%c",component_number,component_number==1?0:'s');
     }
     else
          sprintf(message1,"Saved mesh");
     alert(message1,NULL,NULL,"OK",NULL, 13, 0);

     return 0;
}

int save_FE_triangle_mesh(MESH* m,char *filename){
     EDGE E;
     TRIANGLE T;
     int i;
     int o1,o2,o3;
     char outstring[512];
     float x,y,z;
     int v1,v2,v3;
     int e1,e2,e3;
     VTX v;
     char fstring[128];
     
     FILE *fp=fopen(filename,"wt");
     if(fp==NULL)return -1;

     fprintf(fp,"vertices\n");
     for (i=0;i<m->num_vertices;i++){
        v=m->points[i];
        x=v.x; y=v.y; z=v.z;
        sprintf(outstring,"%d %f %f %f%s\n",i+1,x,-y,z,v.flags&1? "  fixed":"");
        fprintf(fp,outstring);
     }

     make_edge_list(m);
     try_orient(m);
     fprintf(fp,"edges\n");
     for (i=0;i<m->num_edges;i++){
        E=m->edges[i];
        v1=E.v1;
        v2=E.v2;
        sprintf(fstring,"%s",(m->points[v1].flags&1 && m->points[v2].flags&1)?" fixed":"");
        sprintf(outstring,"%d %d %d%s\n",i+1,v1+1,v2+1,fstring);
        fprintf(fp,outstring);
     }
     
     fprintf(fp,"faces\n");
     for (i=0;i<m->num_faces;i++){
        T=m->faces[i];
        v1=T.v1;
        v2=T.v2;
        v3=T.v3;
        e1=find_edge_in_meshedges(m,v1,v2,&o1);
        e2=find_edge_in_meshedges(m,v2,v3,&o2);
        e3=find_edge_in_meshedges(m,v3,v1,&o3);

        sprintf(outstring,"%d %d %d %d\n",i+1,o1*(e1+1),o2*(e2+1),o3*(e3+1));
        fprintf(fp,outstring);
     }
     fprintf(fp,"bodies\n1 ");
     for(i=0;i<m->num_faces;i++){
        sprintf(outstring,"%d ",i+1);
        fprintf(fp,outstring);        
     }
     fputc('\n',fp);

     fclose(fp);
     return 0;
}

int save_FEs_triangle_mesh(MESH* m,char *filename){

     TRIANGLE T;
     int i;
     char outstring[512];
     float x,y,z;
     int v1,v2,v3;
     VTX v;

     FILE *fp=fopen(filename,"wt");
     try_orient(m);
     if(fp==NULL)return -1;
     fprintf(fp,"SIMPLEX_REPRESENTATION\n");
     fprintf(fp,"vertices\n");
     for (i=0;i<m->num_vertices;i++){
        v=m->points[i];
        x=v.x; y=v.y; z=v.z;
        sprintf(outstring,"%d %f %f %f%s\n",i+1,x,-y,z,v.flags&1? "  fixed":"");
        fprintf(fp,outstring);


     }
     
     fprintf(fp,"faces\n");
     for (i=0;i<m->num_faces;i++){
        T=m->faces[i];
        v1=T.v1;
        v2=T.v2;
        v3=T.v3;

        sprintf(outstring,"%d %d %d %d\n",i+1,v1+1,v2+1,v3+1);
        fprintf(fp,outstring);
     }
     fprintf(fp,"bodies\n1 ");
     for(i=0;i<m->num_faces;i++){
        sprintf(outstring,"%d ",i+1);
        fprintf(fp,outstring);        
     }
     fputc('\n',fp);

     fclose(fp);
     return 0;
}
/*
"PCM1",   (include quotes)
PointCount,
#[X,Y,Z,]  (for each point)
FlatTriangleCount,SmoothTriangleCount,
#[PA,PB,PC,] (for each flat triangle)
#[PA,NA,PB,NB,PC,NC,]  (for each smooth triangle)
*/
int write_PCM_triangles(MESH* m,FILE *fp,VECTOR *VNormals){

     TRIANGLE T;
     int i;
     char outstring[512];
     float x,y,z;
     int v1,v2,v3;
     int n1,n2,n3;
     VTX v;
     VECTOR N;

     if(VNormals==NULL)
       fprintf(fp,"%d,\n",m->num_vertices);
     else
       fprintf(fp,"%d,\n",m->num_vertices + m->num_vertices);

     for (i=0;i<m->num_vertices;i++){
        v=m->points[i];
        x=v.x; y=v.y; z=v.z;

        sprintf(outstring,"%f,%f,%f,\n",x,-y,z);
        fprintf(fp,outstring);
     }
     if(VNormals!=NULL){
        for (i=0;i<m->num_vertices;i++){
            N=VNormals[i];
            x=N.x; y=N.y; z=N.z;

            sprintf(outstring,"%f,%f,%f,\n",x,-y,z);
            fprintf(fp,outstring);
        }
     }

     if(VNormals!=NULL){
        fprintf(fp,"0,%d,\n",m->num_faces);
     }
     else{
        fprintf(fp,"%d,0,\n",m->num_faces);
     }
     for (i=0;i<m->num_faces;i++){
        T=m->faces[i];
        v1=T.v1;
        v2=T.v2;
        v3=T.v3;
        if(VNormals!=NULL){
           n1=v1+m->num_vertices;
           n2=v2+m->num_vertices;
           n3=v3+m->num_vertices;
           sprintf(outstring,"%d,%d,%d,%d,%d,%d,\n",v1,n1,v2,n2,v3,n3);
        }
        else{
           sprintf(outstring,"%d,%d,%d,\n",v1,v2,v3);
        }

        fprintf(fp,outstring);
     }

     return 1;
}
int save_PCM_triangle_mesh(MESH* m,char *filename,VECTOR *VNormals){

     FILE *fp=fopen(filename,"wt");
     if(fp==NULL)return -1;
     fprintf(fp,"\"PCM1\",\n");
     waiting_call();
     write_PCM_triangles(m,fp,VNormals);
     waiting_done();
     fclose(fp);
     return 0;
}

int save_WRL_triangle_mesh(MESH* m,char *filename){

     TRIANGLE T;
     int i;
     char outstring[512];
     float x,y,z;
     int v1,v2,v3;
     VTX v;

     FILE *fp=fopen(filename,"wt");
     if(fp==NULL)return -1;
     fprintf(fp,
     "#VRML V1.0 ascii\n\nSeparator\n{"
     "DEF NONAME Separator\n{\nMaterial\n{\n"
     "diffuseColor 1.000 0.000 0.000\n"
     "ambientColor 0.100 0.000 0.000\n"
     "}\nCoordinate3\n{\npoint\n[\n");
     for (i=0;i<m->num_vertices;i++){
        v=m->points[i];
        x=v.x; y=v.y; z=v.z;
        sprintf(outstring,"%f %f %f",x,-y,z);
        fprintf(fp,outstring);
        if(i<m->num_vertices-1)
           fputc(',',fp);
        fputc('\n',fp);

     }
     fprintf(fp,"]\n}\nIndexedFaceSet\n{\ncoordIndex\n[\n");

     for (i=0;i<m->num_faces;i++){
        T=m->faces[i];
        v1=T.v1;
        v2=T.v2;
        v3=T.v3;

        sprintf(outstring,"%d, %d, %d, -1",v1,v2,v3);
        fprintf(fp,outstring);
        if(i<m->num_faces-1)
           fputc(',',fp);
        fputc('\n',fp);
     }
     fprintf(fp,
     "]\n}\n}\n}\n");
     
     fclose(fp);
     return 0;
}

int save_TRI_triangle_mesh(MESH* m,char *filename, int save_sm){

     TRIANGLE T;
     int i;
     char outstring[512];
     float x,y,z;
     int v1,v2,v3;
     VTX v;
     FILE *fp;

     waiting_call();
     fp=fopen(filename,"wt");
     if(fp==NULL)return -1;
     fprintf(fp,"vertices:\n");
     for (i=0;i<m->num_vertices;i++){
        v=m->points[i];
        x=v.x; y=v.y; z=v.z;
        sprintf(outstring,"%f %f %f\n",x,-y,z);
        fprintf(fp,outstring);
     }
     fprintf(fp,"triangles:\n");
     for (i=0;i<m->num_faces;i++){
        T=m->faces[i];
        v1=T.v1;
        v2=T.v2;
        v3=T.v3;

        sprintf(outstring,"%d %d %d\n",v1,v2,v3);
        fprintf(fp,outstring);
     }
     if(num_selected && save_sm&1){
         fprintf(fp,"selected:\n");
         for (i=0;i<num_selected;i++){
    
            v1=curVertex[i];
            sprintf(outstring,"%d\n",v1);
            fprintf(fp,outstring);
         }
     }
     if(save_sm&2){

         for (i=0;i<m->num_vertices;i++){
            if(m->points[i].flags&1)
              break;
         }
         if(i<m->num_vertices){
            fprintf(fp,"marked:\n");
            for (i=0;i<m->num_vertices;i++){
               if(m->points[i].flags&1){
                 sprintf(outstring,"%d\n",i);
                 fprintf(fp,outstring);
               }
            }

         }
     }
     fclose(fp);
     waiting_done();
     return 0;
}



int load_TRI_triangle_mesh(MESH* m,char *filename){

     char instring[512];
     char *tok;
     char *cptr;

     int v1,v2,v3,vt;
     int ea1,ea2,ea3,eb1,eb2,eb3;
     VTX v;

     int nv=m->num_vertices;
     int ne=m->num_edges;
     FILE *fp=fopen(filename,"rt");
     if(fp==NULL)return -1;
     if(fgets(instring, 511,fp)){
          cptr=strchr(instring, '#');
          if(cptr) *cptr='\0';
          if(!strncmp("vertices",instring,7)){
                while(fgets(instring, 511,fp)){
                     if(m->num_vertices>=m->num_vertices_allocated-1){
                         fclose(fp);
                         return -1;
                     }
                     cptr=strchr(instring, '#');
                     if(cptr) *cptr='\0';

                     if(!strncmp("triang",instring,5)||!strncmp("edge",instring,4)){
                        break;
                     }
                     tok = strtok(instring, " ,");
                     v.x=(float)atof(tok);
                     tok=strtok(0, " ,");
                     v.y=-(float)atof(tok);
                     tok=strtok(0, " ,\n");
                     v.z=(float)atof(tok);
                     new_vertex(m,v,0);
                }
                if(!strncmp("edge",instring,4)){
                   while(fgets(instring, 511,fp)){
                     if(m->num_edges>=m->num_edges_allocated-1){
                         fclose(fp);
                         return -2;
                     }
                     cptr=strchr(instring, '#');
                     if(cptr) *cptr='\0';

                     if(!strncmp("face",instring,4)){
                        break;
                     }
                     tok = strtok(instring, " ,");
                     v1=nv+atoi(tok);
                     tok=strtok(0, " ,\n");
                     v2=nv+atoi(tok);
                     if(tok==NULL)break;
                     new_edge(m,v1-1,v2-1,0);
                     
                   }
                   while(fgets(instring, 511,fp)){
                     if(m->num_faces>=m->num_faces_allocated-1){
                         fclose(fp);
                         return -3;
                     }
                     cptr=strchr(instring, '#');
                     if(cptr) *cptr='\0';

                     tok = strtok(instring, " ,");                     
                     ea1=atoi(tok);
                     tok=strtok(0, " ,");
                     ea2=atoi(tok);
                     tok=strtok(0, " ,\n");
                     ea3=atoi(tok);
                     if(tok==NULL)break;

                     eb1 = ea1<0 ? -ea1:ea1;
                     eb2 = ea2<0 ? -ea2:ea2;
                     eb3 = ea3<0 ? -ea3:ea3;
                     eb1=ne+eb1;
                     eb2=ne+eb2;
                     eb3=ne+eb3;

                     v1=m->edges[eb1-1].v1;
                     v2=m->edges[eb1-1].v2;
                     if(ea1<0){vt=v1;v1=v2;v2=vt;}

                     vt=m->edges[eb2-1].v2;
                     if(vt==v1||vt==v2)
                          v3=m->edges[eb2-1].v1;
                     else
                          v3=vt;
                     new_triangle_o(m,v1,v2,v3);
                   }
                   
                }
                else{
                   while(fgets(instring, 511,fp)){
                     if(m->num_faces>=m->num_faces_allocated-1){
                         fclose(fp);
                         return -3;
                     }
                     cptr=strchr(instring, '#');
                     if(cptr) *cptr='\0';

                     tok = strtok(instring, " ,");
                     v1=nv+atoi(tok);
                     tok=strtok(0, " ,");
                     v2=nv+atoi(tok);
                     tok=strtok(0, " ,\n");
                     v3=nv+atoi(tok);
                     if(tok==NULL)break;

                     new_triangle_o(m,v1,v2,v3);

                   }
                }
          }
     }
     fclose(fp);
     return 0;
}

#define PREGETTING -1
#define GETTINGNOTHING   0
#define GETTINGVERTICES  1
#define GETTINGEDGES     2
#define GETTINGTRIANGLES 3
#define GETTINGSELECTED  4
#define GETTINGMARKED    5
int load_TRI_triangle_mesh2(MESH* m,char *filename){

     char instring[512];
     char *tok;
     char *cptr,*iptr;
     int getting=PREGETTING,use_edges=0;
     int v1,v2,v3,vt;
     int ea1,ea2,ea3,eb1,eb2,eb3;
     VTX v;
     int nv=m->num_vertices;
     int ne=m->num_edges;
     FILE *fp;

     if( (fp=fopen(filename,"rt"))==NULL)
            return -1;
     while(fgets(instring, 511,fp)){
          cptr=instring;
          iptr=instring;
          while(isspace(*iptr)) iptr++;
          cptr=strchr(instring, '#');
          if(cptr) *cptr='\0';
          if(!strncmp("vertices",iptr,4)){
             getting=GETTINGVERTICES;
          }
          else if(!strncmp("edges",iptr,4)){
             getting=GETTINGEDGES;
             use_edges=1;
          }
          else if(!strncmp("triang",iptr,4)){
             getting=GETTINGTRIANGLES;
          }
          else if(!strncmp("marked",iptr,4)){
             getting=GETTINGMARKED;
          }
          else if(!strncmp("selected",iptr,4)){
             getting=GETTINGSELECTED;
          }
          
          else{
             switch(getting){

                 case GETTINGVERTICES:
                    if(m->num_vertices>=m->num_vertices_allocated-1){
                         fclose(fp);
                         return -1;
                    }
                    tok = strtok(iptr, " ,");
                    v.x=(float)atof(tok);
                    tok=strtok(0, " ,");
                    v.y=-(float)atof(tok);
                    tok=strtok(0, " ,\n");
                    v.z=(float)atof(tok);
                    new_vertex(m,v,0);
                 break;
                 case GETTINGEDGES:
                    if(m->num_edges>=m->num_edges_allocated-1){
                         fclose(fp);
                         return -2;
                    }
                    tok = strtok(iptr, " ,");
                    v1=nv+atoi(tok);
                    tok=strtok(0, " ,\n");
                    v2=nv+atoi(tok);
                    if(tok==NULL)break;
                    if(v1>=m->num_vertices||v2>=m->num_vertices)
                         return -5;
                    new_edge(m,v1,v2,0);
                     
                 break;
                 case GETTINGTRIANGLES:
                    if(m->num_faces>=m->num_faces_allocated-1){
                        fclose(fp);
                        return -3;
                    }
                    if(use_edges){
                        tok = strtok(iptr, " ,");
                        ea1=atoi(tok);
                        tok=strtok(0, " ,");
                        ea2=atoi(tok);
                        tok=strtok(0, " ,\n");
                        ea3=atoi(tok);
                        if(tok==NULL)break;
   
                        eb1 = ea1<0 ? -ea1:ea1;
                        eb2 = ea2<0 ? -ea2:ea2;
                        eb3 = ea3<0 ? -ea3:ea3;
                        eb1=ne+eb1;
                        eb2=ne+eb2;
                        eb3=ne+eb3;
   
                        v1=m->edges[eb1-1].v1;
                        v2=m->edges[eb1-1].v2;
                        if(ea1<0){vt=v1;v1=v2;v2=vt;}
   
                        vt=m->edges[eb2-1].v2;
                        if(vt==v1||vt==v2)
                             v3=m->edges[eb2-1].v1;
                        else
                             v3=vt;
                        if(v1>=m->num_edges||v2>=m->num_edges||v3>=m->num_edges)
                             return -5;
                        new_triangle_o(m,v1,v2,v3);                        
                    }
                    else{
                        tok = strtok(iptr, " ,");
                        v1=nv+atoi(tok);
                        tok=strtok(0, " ,");
                        v2=nv+atoi(tok);
                        tok=strtok(0, " ,\n");
                        v3=nv+atoi(tok);
                        if(tok==NULL)break;
                        if(v1>=m->num_vertices||v2>=m->num_vertices||v3>=m->num_vertices)
                            return -5;
                        new_triangle_o(m,v1,v2,v3);

                    }
                    
                 break;
                 case GETTINGMARKED:
                    tok = strtok(iptr, " ,\n");
                    while(tok){
                         v1=nv+atoi(tok);
                         if(v1>=m->num_vertices)
                            return -5;
                         m->points[v1].flags=1;
                         tok=strtok(0, " ,\n");
                    }
                 break;
                 case GETTINGSELECTED:
                    tok = strtok(iptr, " ,\n");
                    while(tok){
                         v1=nv+atoi(tok);
                         if(v1>=m->num_vertices)
                            return -5;
                         add_to_selected(v1,0);
                         tok=strtok(0, " ,\n");
                    }
                 break;

             }
          }
     }
     fclose(fp);
     return 0;
}


int load_RAW_triangle_mesh(MESH* m,char *filename){

     int i,vi;
     char instring[512];
     char c;
     char *tok;

     VTX p[3];
     int v[3];
     int nv=m->num_vertices;
     FILE *fp=fopen(filename,"rt");

     if(fp==NULL)return -1;
     c=getc(fp);
     while ( !strchr("-+0,1,2,3,4,5,6,7,8,9.", c)){
         c=getc(fp);
         if(c==EOF){
             fclose(fp);
             return 0;
         }
     }
     ungetc(c,fp);
     while(fgets(instring, 511,fp)){
       tok = strtok(instring, " ,");
       for(i=0;i<3;i++){

            p[i].x=(float)atof(tok);
            tok=strtok(0, " ,");
            p[i].y=-(float)atof(tok);
            if(i==2)
               tok=strtok(0, " ,\n");
            else
               tok=strtok(0, " ,");
            p[i].z=(float)atof(tok);
            v[i]=-1;
            for(vi=nv;vi<m->num_vertices;vi++){
               if(p[i].x==m->points[vi].x && p[i].y==m->points[vi].y && p[i].z==m->points[vi].z){
                  v[i]=vi; break;
               }           
            }
            if(v[i]==-1){
               if(m->num_vertices>=m->num_vertices_allocated-1){
                     fclose(fp);
                     return -1;
               }
               v[i]=new_vertex(m,p[i],0);
            }
            tok=strtok(0, " ,\n");
       }
       if(m->num_faces>=m->num_faces_allocated-1){
           fclose(fp);
           return -3;
       }
       new_triangle_o(m,v[0],v[1],v[2]);
       c=getc(fp);
       while ( !strchr("-+0,1,2,3,4,5,6,7,8,9.", c)&& c!=EOF){       
         c=getc(fp);
       }
       if(c==EOF)
          break;
       ungetc(c,fp);
       

     }
     fclose(fp);
     return 0;
}

int popup_file_select(char *message, char *path, char *ext){
    int fsret;
    //BITMAP * backbmp=create_bitmap(304,160);
    BITMAP * backbmp=create_bitmap(306,162);
    show_mouse(NULL);
    //blit(screen, backbmp, SCREEN_W/2-152,SCREEN_H/2-80,0,0,304,160);
    blit(screen, backbmp, SCREEN_W/2-153,SCREEN_H/2-81,0,0,306,162);
    show_mouse(screen);
    fsret=file_select(message, path, ext);
    show_mouse(NULL);
    //blit(backbmp,screen,0,0, SCREEN_W/2-152,SCREEN_H/2-80,304,160);
    blit(backbmp,screen,0,0, SCREEN_W/2-153,SCREEN_H/2-81,306,162);
    show_mouse(screen);
    destroy_bitmap(backbmp);
    return fsret;
}

int do_savefiledialog(char * message,char * filenamebuf, char * ext){
/* returns 0 for cancel, returns 1 for continue saving process*/
     int fsret=1,alertret;
     char buf1[355];
     char buf2[80]="Is it OK to overwrite it?";

     while (fsret){
          fsret=popup_file_select(message, filenamebuf, ext);
          if(fsret==0) return 0;
          if(exists(filenamebuf)){
               sprintf(buf1, "%s already exists.", filenamebuf);
               alertret=alert3(buf1, buf2, NULL, "Yes", "No","Cancel", 'y','n', 27) ;
               if(alertret==1) break;
               else if (alertret == 3)
                  return 0;
               else if(alertret==2){
                  fsret=1;
               }
          }
          else fsret=0;
     }
     return(1);
}


