/*
*  triangles.c
*
*  Main module
*
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "allegro.h"

#include "triangles.h"
#include "tridat.h"  /* mouse icons, etc */


int SCREENWIDE;
int SCREENHIGH;
int VIEWS_WIDTH;
int VIEWS_HEIGHT;

void set_dialog_view_sizes(int s_w,int s_h);

BEGIN_COLOR_DEPTH_LIST
  COLOR_DEPTH_8
END_COLOR_DEPTH_LIST
unsigned _stklen = 2*1048576; /* enlarge stack for the recursive triangle dividing */
unsigned stackbytes;

DATAFILE * allegrodata;
DIALOG the_dialog[];

int mesh_face_alloc_size=50000;
int mesh_vtx_alloc_size =50000;
int mesh_edge_alloc_size=100000;

int curVertex_alloc_size=50000;

#define XYVIEW XY
#define ZYVIEW ZY
#define XZVIEW XZ

#define VIEW_3D 4
/*
float xy_viewscale=1.0,
      zy_viewscale=1.0,
      xz_viewscale=1.0;
float xy_viewpan_x=0.0,xy_viewpan_y=0.0,
      zy_viewpan_x=0.0,zy_viewpan_y=0.0,
      xz_viewpan_x=0.0,xz_viewpan_y=0.0;
*/

char def_ini_filename[512]="triangles.ini";
int def_s_w=640,def_s_h=480;

char savepath[512]="";
char savetripath[512]="";

int error_exit(char *message1,char *message2);
int esc_pressed();
int escape=0;
void waiting_call();
void waiting_done();

/* external, in deform.c */
int SmoothSubdivide(MESH *m,float smoothing,int avg, int go_ahead);
int perturb_mesh(MESH *m, float N,int Seed);
void vtx_wave(VTX *v, float Frequency, float Phase, float WaveScale);
int wave_mesh(MESH *,float,float,float);
void vtx_ripple(VTX *v, VECTOR RippleAxis, float Frequency,float Phase, VECTOR RippleScale);

int ripple_mesh(MESH *m, VECTOR RippleAxis, float Frequency,float Phase, VECTOR RippleAmount);
void vtx_twist(VTX *v,float TwistAngle, VTX TwistCentre);
int twist_mesh(MESH *m,float TwistAngle, VTX TwistCentre);
int vtx_twist_about(VTX *v,float TwistAngle, VTX TwistCentre,VECTOR TwistAxis);
int twist_mesh_about(MESH *m,float TwistAngle, VTX TwistCentre,VECTOR TwistAxis);
void vtx_bend(VTX *v,VTX BendPoint,VECTOR BX,float BA);
int bend_mesh(MESH *m,VTX BendPoint,VECTOR BX,float BA);
void vtx_repulse_point(VTX *v,VTX RepulsePoint, float c,float a, float b);
int repulse_point_mesh(MESH *m,VTX RepulsePoint, float c,float a, float b);

/* external, in poly2tri.c */
int do_poly_dialog();

/* external, in reduce.c */
int ProgressiveMesh(MESH *m);

/* external, in file.c */
int save_TRI_triangle_mesh(MESH* m,char *filename,int  );
int save_POV_triangle_mesh(MESH* m,char *filename, VECTOR *VN);
int save_mesh2(MESH* m,char *filename, VECTOR *VN);
int save_RAW_triangle_mesh(MESH* m,char *filename, VECTOR *VN);
int save_PCM_triangle_mesh(MESH* m,char *filename,VECTOR *VNormals);
int save_WRL_triangle_mesh(MESH* m,char *filename);
int save_FE_triangle_mesh(MESH* m,char *filename);
int load_RAW_triangle_mesh(MESH* m,char *filename);
int load_TRI_triangle_mesh(MESH* m,char *filename);
int load_TRI_triangle_mesh2(MESH* m,char *filename);

int popup_file_select(char *message, char *path, char *ext);
int do_savefiledialog(char * message,char * filenamebuf, char * ext);

/* external, in mesh.c*/
void  triangle_normal_V(VTX v1, VTX v2, VTX v3,VECTOR *tnorm);
int get_vertex_normals(TRIANGLE *TArray,int TNum, int VNum,VECTOR *VNormals,int AreaAvg);
MESH *new_mesh(int n_v,int n_f,int n_e,VECTOR v, VECTOR r, float dz, VECTOR dr,float scale);
void delete_mesh(MESH *m);
void init_shape(MESH *m,VECTOR v, VECTOR r, float dz, VECTOR dr,float scale);
int new_vertex(MESH *m,VTX vtx,int f);
int new_triangle(MESH *m,int v1, int v2, int v3,int oflag);
int new_triangle_o(MESH *m,int v1, int v2, int v3);
int make_edge_list(MESH *m);
int find_edge_in_meshedges(MESH *m,int va,int vb, int *o);
int find_edge_in_triangle(TRIANGLE * T,  int va,int vb);
float vtx2vtx_dist(MESH *m,int v1,int v2);
int try_orient(MESH *m);
void delete_triangle(MESH *m,int trinum);
void delete_vertex(MESH *m,int vtxnum);
int triangle_has_vertex_p(MESH *m,int ti,int  vi);
void reverse_triangle(TRIANGLE * T);
int find_triangle_in_mesh(MESH *m,int v1, int v2, int v3);
int divide_triangle_mesh(MESH *m);
void insert_vertex(MESH *m,int va,int vb,VTX vtx);
int vertex_used_p(MESH *mesh,int vi);
int meld_vertices(MESH *m, int i1,int i2);
void animate_shape(MESH *m);
void ani_rotate(MESH *m,VECTOR dr);
void ani_translate(MESH *m,VECTOR dr);
void ani_scale(MESH *m,float scale);
float DistPointToSegment2d(float px, float py, float X1, float Y1, float X2, float Y2);
void mesh_CG(MESH *m,VTX *cg);
int add_mesh_to_mesh(MESH *dest,MESH *source);
void recurse_connected(MESH *m,int vtx);
/*-----------------------------*/


int GreyScaleBottom=128,
    GreyScaleTop=255,
    GreenScaleBottom=96,
    GreenScaleTop=127,
    RedScaleBottom=64,
    RedScaleTop=95;
    /* bottom of shading palette */
int ShadingPaletteBottom,
    /* top of shading palette */
    ShadingPaletteTop;

VTX Origin={0,0,0,0};
VTX light_pos={800,-800,-800,0};

int normal_orient=0;
int show_triangles=0;


BITMAP *texture=NULL;


/* animation rotation speed */
VECTOR Initial_dr={0,0,0};
/* animation z-motion speed */
float Initial_dz=0.0;
/* initial position */
VECTOR Initial_v={0,0,400};
/* initial rotation */
VECTOR Initial_r={0,0,0};
/*initial scale*/
float Initial_scale=1.5;


int show3daxes=1;
/* points used for drawing the 3d axes */
VTX xaxis3dend={100,  0,  0,0};
VTX yaxis3dend={  0,-100,  0,0};
VTX zaxis3dend={  0,  0,100,0};
VTX xaxis3dendout;
VTX yaxis3dendout;
VTX zaxis3dendout;
VTX Originout;


/* translate shapes from 3d world space to 2d screen space */
void translate_world_to_screen(MESH * m)
{
   int d;
   MATRIX_f matrix;

   /* build a transformation matrix */
   get_transformation_matrix_f(&matrix, m->shape.scale,
				m->shape.rx, m->shape.ry, m->shape.rz,
				m->shape.x, m->shape.y, m->shape.z);
   /* output the vertices */
   for (d=0; d<m->num_vertices; d++) {
	 apply_matrix_f(&matrix, m->points[d].x, m->points[d].y, m->points[d].z,
                        &(m->output_points[d].x), &(m->output_points[d].y), &(m->output_points[d].z));
	 persp_project_f(m->output_points[d].x, m->output_points[d].y, m->output_points[d].z,
                         &(m->output_points[d].x), &(m->output_points[d].y));
   }
   /* output the faces */
   for (d=0; d<m->num_faces; d++) {
	 m->output_faces[d] = m->faces[d];
	 m->output_faces[d].vtxlist = m->output_points;
   }

   if(show3daxes){
       apply_matrix_f(&matrix, xaxis3dend.x, xaxis3dend.y, xaxis3dend.z,
                            &(xaxis3dendout.x), &(xaxis3dendout.y), &(xaxis3dendout.z));
       persp_project_f(xaxis3dendout.x, xaxis3dendout.y, xaxis3dendout.z,
                             &(xaxis3dendout.x), &(xaxis3dendout.y));
       apply_matrix_f(&matrix, yaxis3dend.x, yaxis3dend.y, yaxis3dend.z,
                            &(yaxis3dendout.x), &(yaxis3dendout.y), &(yaxis3dendout.z));
       persp_project_f(yaxis3dendout.x, yaxis3dendout.y, yaxis3dendout.z,
                             &(yaxis3dendout.x), &(yaxis3dendout.y));
       apply_matrix_f(&matrix, zaxis3dend.x, zaxis3dend.y, zaxis3dend.z,
                            &(zaxis3dendout.x), &(zaxis3dendout.y), &(zaxis3dendout.z));
       persp_project_f(zaxis3dendout.x, zaxis3dendout.y, zaxis3dendout.z,
                             &(zaxis3dendout.x), &(zaxis3dendout.y));
                             
       apply_matrix_f(&matrix, Origin.x, Origin.y, Origin.z,
                            &(Originout.x), &(Originout.y), &(Originout.z));
       persp_project_f(Originout.x, Originout.y, Originout.z,
                             &(Originout.x), &(Originout.y));
   }
}


void triangle_draw(BITMAP *b, VTX *v1, VTX *v2, VTX *v3)
{
   int col;
   VTX fc={(v1->x+v2->x+v3->x)/3,(v1->y+v2->y+v3->y)/3,(v1->z+v2->z+v3->z)/3,0};
   VECTOR light_vec={light_pos.x-fc.x,light_pos.y-fc.y,light_pos.z-fc.z};
   VECTOR tnorm;
   float lightdot;
   /* V3D_f is an Allegro data type needed for triangle_3d() */
   V3D_f vtx1 = { v1->x, v1->y, v1->z, 0,  0,  0 };
   V3D_f vtx2 = { v2->x, v2->y, v2->z, 31, 0,  0 };
   V3D_f vtx3 = { v3->x, v3->y, v3->z, 0, 31, 0 };

   triangle_normal_V(*v1, *v2, *v3,&tnorm);
   normalize_vector_f(&(light_vec.x),&(light_vec.y),&(light_vec.z));
   lightdot=dot_product_f(tnorm.x,tnorm.y,tnorm.z,light_vec.x,light_vec.y,light_vec.z);
   if(!normal_orient){
       lightdot=fabs((double)lightdot);
       col = MID(ShadingPaletteBottom,
             (int) ((ShadingPaletteBottom+ShadingPaletteTop)+(ShadingPaletteTop-ShadingPaletteBottom)*lightdot*lightdot)/2,
             ShadingPaletteTop);
   }
   else{
       col = MID(ShadingPaletteBottom,
             (int) ( (ShadingPaletteTop+ShadingPaletteBottom)/2+(ShadingPaletteTop-ShadingPaletteBottom)*lightdot/2),
             ShadingPaletteTop);
   }
   vtx1.c = vtx2.c = vtx3.c = palette_color[col];

   triangle3d_f(b, POLYTYPE_FLAT, texture, &vtx1, &vtx2, &vtx3);
}


/* callback for qsort() */
int triangz_cmp(const void *e1, const void *e2)
{
   TRIANGLE *q1 = (TRIANGLE *)e1;
   TRIANGLE *q2 = (TRIANGLE *)e2;

   float d1 = q1->vtxlist[q1->v1].z + q1->vtxlist[q1->v2].z +
	      q1->vtxlist[q1->v3].z ;
              
   float d2 = q2->vtxlist[q2->v1].z + q2->vtxlist[q2->v2].z +
	      q2->vtxlist[q2->v3].z ;

   return (d2 > d1?1:-1);
}

/* draw the shapes calculated by translate_world_to_screen() */
void draw_3d(MESH *m,BITMAP *b)
{
   int c;
   TRIANGLE *face=m->output_faces;
   VTX *v1, *v2, *v3;

   /* depth sort */
   qsort(m->output_faces, m->num_faces, sizeof(TRIANGLE), triangz_cmp);

   for (c=0; c < m->num_faces ; c++) {
      /* find the vertices used by the face */
      v1 = face->vtxlist + face->v1;
      v2 = face->vtxlist + face->v2;
      v3 = face->vtxlist + face->v3;

      triangle_draw(b, v1, v2, v3);

      face++;
   }
   if(show3daxes){
      line(b,Originout.x,Originout.y,xaxis3dendout.x,xaxis3dendout.y,10);
      line(b,Originout.x,Originout.y,yaxis3dendout.x,yaxis3dendout.y,11);
      line(b,Originout.x,Originout.y,zaxis3dendout.x,zaxis3dendout.y,12);
   }
}

BITMAP *threenotribmp=NULL;
BITMAP *threetribmp=NULL;
BITMAP *notribmp=NULL;
BITMAP * xybmp;
BITMAP * zybmp;
BITMAP * xzbmp;
BITMAP * threedbmp;
BITMAP * waitmouse=NULL;
BITMAP * deform_icon;

MESH *mesh=NULL;
MESH *mesh_copy=NULL;
int *curVertex=NULL;
int num_selected=0;

int show_vertices=1;
#define VERTEXCOLOR 1
#define SELECTEDCOLOR 2
#define MARKEDCOLOR 4
#define WIRECOLOR 255
void draw_flat_wire_view(MESH *m,BITMAP *b,int view){
   TRIANGLE *face;
   VTX *v1, *v2, *v3;
   float x1=0,y1=0,x2=0,y2=0,x3=0,y3=0;
   int ix1,iy1,ix2,iy2,ix3,iy3;
   int c,s;
   face=m->faces;

   for (c=0; c < m->num_faces ; c++) {
          /* find the vertices used by the face */
             v1 = face->vtxlist + face->v1;
             v2 = face->vtxlist + face->v2;
             v3 = face->vtxlist + face->v3;

             if(view==XYVIEW){

                x1=v1->x;y1=v1->y;
                x2=v2->x;y2=v2->y;
                x3=v3->x;y3=v3->y;
             }
             else if(view==ZYVIEW){
                x1=v1->z;y1=v1->y;
                x2=v2->z;y2=v2->y;
                x3=v3->z;y3=v3->y;

             }

             else if(view==XZVIEW){
                x1=v1->x;y1=v1->z;
                x2=v2->x;y2=v2->z;
                x3=v3->x;y3=v3->z;

             }
             ix1=(int)x1; iy1=(int)y1;
             ix2=(int)x2; iy2=(int)y2;
             ix3=(int)x3; iy3=(int)y3;
             line(b,ix1+ VIEWS_WIDTH/2,iy1+ VIEWS_HEIGHT/2,ix2+ VIEWS_WIDTH/2,iy2+ VIEWS_HEIGHT/2,WIRECOLOR);
             line(b,ix2+ VIEWS_WIDTH/2,iy2+ VIEWS_HEIGHT/2,ix3+ VIEWS_WIDTH/2,iy3+ VIEWS_HEIGHT/2,WIRECOLOR);
             line(b,ix3+ VIEWS_WIDTH/2,iy3+ VIEWS_HEIGHT/2,ix1+ VIEWS_WIDTH/2,iy1+ VIEWS_HEIGHT/2,WIRECOLOR);
             face++;
   }
   if(show_vertices){
             for (c=0; c < m->num_vertices ; c++) {
                 v1=&(m->points[c]);
                 if(view==XYVIEW){
                     x1=v1->x;y1=v1->y;
                 }
                 else if(view==ZYVIEW){
                     x1=v1->z;y1=v1->y;
                 }
                 else if(view==XZVIEW){
                     x1=v1->x;y1=v1->z;
                 }
                 
                 ix1=(int)x1; iy1=(int)y1;

                 circle(b,ix1+ VIEWS_WIDTH/2,iy1+ VIEWS_HEIGHT/2,2,VERTEXCOLOR);
                 if(v1->flags &1){
                    rectfill(b,ix1+ VIEWS_WIDTH/2-1,iy1+ VIEWS_HEIGHT/2-1,ix1+ VIEWS_WIDTH/2+1,iy1+ VIEWS_HEIGHT/2+1,MARKEDCOLOR);
                 }
             }
             for(s=0;s<num_selected;s++){
                 v1=&(m->points[curVertex[s]]);
                 if(view==XYVIEW){
                     x1=v1->x;y1=v1->y;
                 }
                 else if(view==ZYVIEW){
                     x1=v1->z;y1=v1->y;
                 }
                 else if(view==XZVIEW){
                     x1=v1->x;y1=v1->z;
                 }
                 
                 ix1=(int)x1; iy1=(int)y1;
                 circlefill(b,ix1+ VIEWS_WIDTH/2,iy1+ VIEWS_HEIGHT/2,2,SELECTEDCOLOR);
                 if(v1->flags &1){

                    rectfill(b,ix1+ VIEWS_WIDTH/2-1,iy1+ VIEWS_HEIGHT/2-1,ix1+ VIEWS_WIDTH/2+1,iy1+ VIEWS_HEIGHT/2+1,MARKEDCOLOR);
                 }


             }
   }
}

int vertex_in_selected(int vtxnum){
    int ret=-1;
    int s;
    for(s=0;s<num_selected;s++){
       if(vtxnum==curVertex[s])
          return s;
    }
    return ret;
}
void unselect_vertex(int snum){
     int i;
     for(i=snum;i<num_selected;i++){
               curVertex[i]=curVertex[i+1];
     }
     num_selected--;
}

void add_to_selected (int vtxnum, int removeiffound){
   int found=0;
   int i;
   for(i=0;i<num_selected;i++){
        if(curVertex[i]==vtxnum){
             if(removeiffound)
                unselect_vertex(i);
             
             found=1;
             break;
        }
   }
   if(!found){
     if(num_selected<curVertex_alloc_size-1){
        curVertex[num_selected]=vtxnum;
     }
     num_selected++;
   }
}
int select_in_rectangle(int view,int rectstartx,int rectstarty,int mx,int my, int mode){
       VTX v;
       int vi,s;
       float x=0,y=0;
       int x1= (mx<rectstartx)?mx:rectstartx;
       int y1= (my<rectstarty)?my:rectstarty;
       int x2= (x1==mx)?rectstartx:mx;
       int y2= (y1==my)?rectstarty:my;
       for (vi=0; vi < mesh->num_vertices ; vi++) {
                 v=mesh->points[vi];
                 switch(view){
                   case XYVIEW:
                     x=v.x;y=v.y;
                   break;
                   case ZYVIEW:
                     x=v.z;y=v.y;
                   break;
                   case XZVIEW:
                     x=v.x;y=v.z;
                   break;
                 }
                 if(x>=x1 && x<=x2 && y>=y1 && y<=y2){
                     if(mode==0){
                        for(s=0;s<num_selected;s++){
                           if(vi==curVertex[s]){
                               unselect_vertex(s);
                           }
                        }
                     }
                     else{
                        add_to_selected(vi,0);
                     }
                 }
                 
       }
       return 0;
}


extern int *connected;

int maxnedge=10;

int select_connected_component(int vtx){
       int vi;

       connected=malloc(sizeof(int)*mesh->num_vertices);
       if(connected==NULL)
           return -1;
       for(vi=0;vi<mesh->num_vertices;vi++){
           connected[vi]=0;
       }
       waiting_call();
       recurse_connected(mesh,vtx);
       waiting_done();
       for(vi=0;vi<mesh->num_vertices;vi++){
             if(connected[vi]){
                add_to_selected(vi,0);
             }
       }
       free(connected); connected=NULL;
       return 0;
}


int unselect_connected_component(int vtx){
       int vi;
       int s;
       connected=malloc(sizeof(int)*mesh->num_vertices);
       if(connected==NULL)
           return -1;
       for(vi=0;vi<mesh->num_vertices;vi++){
           connected[vi]=0;
       }
       waiting_call();
       recurse_connected(mesh,vtx);
       waiting_done();
       for(vi=0;vi<mesh->num_vertices;vi++){
             if(connected[vi]){
                for(s=0;s<num_selected;s++){
                   if(vi==curVertex[s]){
                       unselect_vertex(s);
                   }
                }

             }
       }
       free(connected); connected=NULL;
       return 0;
}

float *connection_dist;
int recurse_edge_dist_from_to(int v1,int v2,float *dist){

       int ti;
       int ret=0;
       TRIANGLE T;
       float edged;
       float d;
       connected[v1]=1;

       if(connection_dist[v1]<0 || *dist<connection_dist[v1]){
          connection_dist[v1] = *dist;
       }
       else{
          *dist=connection_dist[v1];
       }
       if(v1==v2){
          return 1;
       }
       for (ti=0; ti < mesh->num_faces ; ti++) {
           T=mesh->faces[ti];
           if(T.v1==v1||T.v2==v1||T.v3==v1){
                 if(T.v1!=v1){
                    edged=vtx2vtx_dist(mesh,v1,T.v1);
                    if(connection_dist[T.v1]<0 || edged+*dist<connection_dist[T.v1]){
                       d=edged+*dist;
                       if(recurse_edge_dist_from_to(T.v1,v2,&d))
                           return 1;
                    }
                    
                 }
                 if(T.v2!=v1){
                    edged=vtx2vtx_dist(mesh,v1,T.v2);
                    if(connection_dist[T.v2]<0 || edged+*dist<connection_dist[T.v2]){
                       d=edged+*dist;
                       if(recurse_edge_dist_from_to(T.v2,v2,&d))
                           return 1;
                    }

                 }
                 if(T.v3!=v1){
                    edged=vtx2vtx_dist(mesh,v1,T.v3);
                    if(connection_dist[T.v3]<0 || edged+*dist<connection_dist[T.v3]){
                       d=edged+*dist;
                       if(recurse_edge_dist_from_to(T.v3,v2,&d))
                           return 1;
                    }

                 }
           }
       }
       return ret;
}


int recurse_edge_dist_from(int v1,float *dist,int *nedge){

       int ti;
       int ret=0;
       TRIANGLE T;
       float edged;
       float d;
       int ne;

       if(*nedge<maxnedge && (connection_dist[v1]<0 || *dist<connection_dist[v1])){
          connection_dist[v1] = *dist;
          connected[v1]=*nedge;
       }
       else{
          return 1;
       }

       for (ti=0; ti < mesh->num_faces ; ti++) {
           T=mesh->faces[ti];
           if(T.v1==v1||T.v2==v1||T.v3==v1){
                 if(T.v1!=v1){
                    edged=vtx2vtx_dist(mesh,v1,T.v1);

                    if((*nedge+1<maxnedge) && (connection_dist[T.v1]<0 || edged+*dist<connection_dist[T.v1])){
                       d=edged+*dist;
                       ne=*nedge+1;
                       recurse_edge_dist_from(T.v1,&d,&ne);
                    }                    
                 }
                 if(T.v2!=v1){
                    edged=vtx2vtx_dist(mesh,v1,T.v2);
                    if((*nedge+1<maxnedge) && (connection_dist[T.v2]<0 || edged+*dist<connection_dist[T.v2])){
                       d=edged+*dist;
                       ne=*nedge+1;
                       recurse_edge_dist_from(T.v2,&d,&ne);
                    }
                 }
                 if(T.v3!=v1){
                    edged=vtx2vtx_dist(mesh,v1,T.v3);
                    if((*nedge+1<maxnedge) && (connection_dist[T.v3]<0 || edged+*dist<connection_dist[T.v3])){
                       d=edged+*dist;
                       ne=*nedge+1;
                       recurse_edge_dist_from(T.v3,&d,&ne);
                    }
                 }
           }
       }
       return ret;
}

int get_connected_distances(int v1){
/* allocates the array connected[] and connection_dist[], so
   caller function should also, later, call free_connected_distances() */
       int dret;
       float dist=0;
       int nedge=0;
       int vj;
       connection_dist=malloc(sizeof(float)*mesh->num_vertices);
       if(connection_dist==NULL)
           return -1;
       connected=malloc(sizeof(int)*mesh->num_vertices);
       if(connected==NULL)
           return -1;

       for(vj=0;vj<mesh->num_vertices;vj++){
           connected[vj]=-1;
       }    
       for(vj=0;vj<mesh->num_vertices;vj++){
           connection_dist[vj]=-1.0;
       }
       
       waiting_call();
       dret=recurse_edge_dist_from(v1,&dist,&nedge);
       waiting_done();

       return dret;

}
void free_connected_distances(){
     if(connection_dist)
       free(connection_dist);
     connection_dist=NULL;
     if(connected)
       free(connected);
     connected=NULL;
}

VTX repulse_mouse_vtx={0,0,0,0};
int moving_vertex=-1;
float pull_a=.00010;
#define EDGEPULL_A_MULT 100.0

void pull_by_edges(int view,int mx,int my){
       float dx=0,dy=0,dz=0;
       float exp_pull;
       int vi;
       float edgepull_a=pull_a*EDGEPULL_A_MULT;
       switch(view){
          case XYVIEW:
               dx=(float)mx;
               dy=(float)my;
               dz=0;
          break;
          case ZYVIEW:
               dz=(float)mx;
               dy=(float)my;
               dx=0;
          break;
          case XZVIEW:
               dx=(float)mx;
               dz=(float)my;
               dy=0;
          break;
       

       }
       for(vi=0;vi<mesh->num_vertices;vi++){
          if(connection_dist[vi]>=0 && vi!=moving_vertex){
             exp_pull=(float)exp((double)(-edgepull_a*connection_dist[vi]));
             mesh->points[vi].x+=dx*exp_pull;
             mesh->points[vi].y+=dy*exp_pull;
             mesh->points[vi].z+=dz*exp_pull;
          }
       }
}
void pull_selected_by_edges(int view,int mx,int my){
       float dx=0,dy=0,dz=0;
       float exp_pull;
       int si;
       float edgepull_a=pull_a*EDGEPULL_A_MULT;
       switch(view){
          case XYVIEW:
               dx=(float)mx;
               dy=(float)my;
               dz=0;
          break;
          case ZYVIEW:
               dz=(float)mx;
               dy=(float)my;
               dx=0;
          break;
          case XZVIEW:
               dx=(float)mx;
               dz=(float)my;
               dy=0;
          break;
       

       }
       for(si=0;si<num_selected;si++){
          if(connection_dist[curVertex[si]]>=0 && curVertex[si]!=moving_vertex){
             exp_pull=(float)exp((double)(-edgepull_a*connection_dist[curVertex[si]]));
             mesh->points[curVertex[si]].x+=dx*exp_pull;
             mesh->points[curVertex[si]].y+=dy*exp_pull;
             mesh->points[curVertex[si]].z+=dz*exp_pull;
          }
       }
}

int triangle_in_selected(TRIANGLE T,int *s1,int *s2,int*s3){
    int s;
    int ok=0;
    int v1=T.v1,v2=T.v2,v3=T.v3;
    for(s=0;s<num_selected;s++){
          if(v1==curVertex[s]){
                  *s1=s;
                  ok=1;
                  break;
          }
    }
    if(!ok)
       return 0;
    ok=0;
    for(s=0;s<num_selected;s++){
          if(v2==curVertex[s]){
                  *s2=s;
                  ok=1;
                  break;
          }
    }
    if(!ok)
       return 0;
    ok=0;
    for(s=0;s<num_selected;s++){
          if(v3==curVertex[s]){
                  *s3=s;
                  ok=1;
                  break;
          }
    }

    return ok;
}
void reverse_selected(){
   int s1,s2,s3;
   int t;
   TRIANGLE *T;
   for(t=0;t<mesh->num_faces;t++){
        T=&(mesh->faces[t]);
        if( triangle_in_selected(*T,&s1,&s2,&s3)){
           reverse_triangle(T);
        }
   }
}
int mark_selected(int mark,char op){
    VTX*v;
    int s;
    for(s=0;s<num_selected;s++){
         v=&(mesh->points[curVertex[s]]);
         if(op=='&')
            v->flags&=mark;
         else if(op=='|')
            v->flags|=mark;
    }
    return 1;
}

#define dist2_2d(x1,y1,x2,y2) (((float)x1-(float)x2)*((float)x1-(float)x2)+((float)y1-(float)y2)*((float)y1-(float)y2))
int mouse_in_this_vertex_p(int view,int i,int mx,int my){
       int f=0;
       VTX*p=mesh->points;
       switch(view){
         case XYVIEW:
            if(dist2_2d(mx-1,my-1,p[i].x,p[i].y)<10){
               f=1;
            }
         break;
         case ZYVIEW:
            if(dist2_2d(mx-1,my-1,p[i].z,p[i].y)<10){
               f=1;
            }
         break;
         case XZVIEW:
            if(dist2_2d(mx-1,my-1,p[i].x,p[i].z)<10){
               f=1;
            }
         break;
       }
       return f;
}
int mouse_in_vertex_number(int view,int mx,int my){
       int vi;
       int ret=-1;
       for(vi=0;vi<mesh->num_vertices;vi++){
           if(mouse_in_this_vertex_p(view,vi,mx,my))
               return vi;
       }
       return ret;
}

int select_mouse_vertex_selection(int view, int mx, int my){
   int f=0,i;
   for(i=0;i<mesh->num_vertices;i++){
       if(mouse_in_this_vertex_p(view,i,mx, my)){
           add_to_selected (i,1);
           f=1;
       }

   }
   return f;
}
int unselect_mouse_vertex_selection(int view, int mx, int my){
   int f=0,i;
   for(i=0;i<num_selected;i++){
       if(mouse_in_this_vertex_p(view,curVertex[i],mx, my)){
           unselect_vertex(i);
           f=1;
           break; /* unselect one at a time */
       }
   }
   return f;
}

void mouse_createvertex(int view,int mx,int my,int add){
  VTX v;
  switch(view){
        case XYVIEW:
             v.x = mx;v.y = my;
             if(num_selected)
                v.z = mesh->points[curVertex[num_selected-1]].z;
             else v.z=0;
             break;
        case ZYVIEW:
             v.z = mx;v.y = my;
             if(num_selected)
                v.x = mesh->points[curVertex[num_selected-1]].x;
             else v.x=0;
             break;
             
        case XZVIEW:
             v.x = mx;v.z = my;
             if(num_selected)
                v.y = mesh->points[curVertex[num_selected-1]].z;
             else v.y=0;
             break;
    }    
    new_vertex(mesh,v,0);
    if(!add)
       num_selected=0;
    add_to_selected(mesh->num_vertices-1,0);

}

int mouse_nearest_edge(int view,int mx,int my,int *va,int * vb){
/* mouse at mx,my in view is nearest edge va->vb */
   int fi;

   float vax=0,vay=0,vbx=0,vby=0;
   float mindist=0,dist;
   TRIANGLE T;
   
   for(fi=0;fi<mesh->num_faces;fi++){
         T=mesh->faces[fi];
      
         switch(view){
            case  XYVIEW:
               vax=mesh->points[T.v1].x;vay=mesh->points[T.v1].y;
               vbx=mesh->points[T.v2].x;vby=mesh->points[T.v2].y;
            break;
            case  ZYVIEW:
               vax=mesh->points[T.v1].z;vay=mesh->points[T.v1].y;
               vbx=mesh->points[T.v2].z;vby=mesh->points[T.v2].y;

            break;
            case  XZVIEW:
               vax=mesh->points[T.v1].x;vay=mesh->points[T.v1].z;
               vbx=mesh->points[T.v2].x;vby=mesh->points[T.v2].z;
            break;
            
         }
         dist=DistPointToSegment2d(mx,my,vax, vay, vbx, vby);
         if(fi==0){
            mindist=dist;
            *va=T.v1;*vb=T.v2;
         }
         else{
            if(dist<mindist){
               mindist=dist;
               *va=T.v1;*vb=T.v2;
            }
         }

         switch(view){
            case  XYVIEW:
               vax=mesh->points[T.v2].x;vay=mesh->points[T.v2].y;
               vbx=mesh->points[T.v3].x;vby=mesh->points[T.v3].y;
            break;
            case  ZYVIEW:
               vax=mesh->points[T.v2].z;vay=mesh->points[T.v2].y;
               vbx=mesh->points[T.v3].z;vby=mesh->points[T.v3].y;

            break;
            case  XZVIEW:
               vax=mesh->points[T.v2].x;vay=mesh->points[T.v2].z;
               vbx=mesh->points[T.v3].x;vby=mesh->points[T.v3].z;
            break;
            
         }
         dist=DistPointToSegment2d(mx,my,vax, vay, vbx, vby);
         if(dist<mindist){
               mindist=dist;
               *va=T.v2;*vb=T.v3;
         }
         switch(view){
            case  XYVIEW:
               vax=mesh->points[T.v3].x;vay=mesh->points[T.v3].y;
               vbx=mesh->points[T.v1].x;vby=mesh->points[T.v1].y;
            break;
            case  ZYVIEW:
               vax=mesh->points[T.v3].z;vay=mesh->points[T.v3].y;
               vbx=mesh->points[T.v1].z;vby=mesh->points[T.v1].y;

            break;
            case  XZVIEW:
               vax=mesh->points[T.v3].x;vay=mesh->points[T.v3].z;
               vbx=mesh->points[T.v1].x;vby=mesh->points[T.v1].z;
            break;
            
         }
         dist=DistPointToSegment2d(mx,my,vax, vay, vbx, vby);
         if(dist<mindist){
               mindist=dist;
               *va=T.v3;*vb=T.v1;
         }
   }

   return 0;

}
void mouse_insertvertex(int view,int mx,int my){
   int va,vb;
   VTX v={0,0,0,0};
   if(mesh->num_faces==0) return;
   mouse_nearest_edge(view,mx,my,&va,&vb);
   switch(view){
       case XYVIEW:
          v.x=mx;
          v.y=my;
          v.z=(mesh->points[va].z+mesh->points[vb].z)/2;
       break;
       case ZYVIEW:
          v.x=(mesh->points[va].x+mesh->points[vb].x)/2;
          v.y=my;
          v.z=mx;
       break;
       case XZVIEW:
          v.x=mx;
          v.y=(mesh->points[va].y+mesh->points[vb].y)/2;
          v.z=my;
       break;
       insert_vertex(mesh,va,vb,v);

   }
   insert_vertex(mesh,va,vb,v);
}

void mouse_delete(int view, int mx,int my){
   int i,s;
   for(i=0;i<mesh->num_vertices;i++){
       if(mouse_in_this_vertex_p(view,i,mx, my)){
           delete_vertex(mesh,i);
           for(s=0;s<num_selected;s++){
               if(curVertex[s]==i)
                    unselect_vertex(s);
           }
           for(s=0;s<num_selected;s++){
               if(curVertex[s]>i)
                    curVertex[s]--;
           }
           return;
       }
   }
}

void move_vertex(int view, int i,int x, int y){
       VTX *v=&(mesh->points[i]);
       switch(view){
           case XYVIEW:
                   v->x += x;
                   v->y += y;
                   break;
           case XZVIEW:
                   v->x += x;
                   v->z += y;
                   break;
           case ZYVIEW:
                   v->z += x;
                   v->y += y;
       }
}
void move_selected(int view,int x, int y){
    int i;

    for(i=0;i<num_selected;i++){
       move_vertex(view,curVertex[i],x,y);
    }
}

void move_all(int view,int x, int y){
    int i;
    for(i=0;i<mesh->num_vertices;i++){
       move_vertex(view,i,x,y);
    }
}


void pull_vertex(int vi,VTX pullpoint, float px,float py,float pz){
       VTX *v;
       float vx,vy,vz;
       float mult;

       v=&(mesh->points[vi]);
       vx=pullpoint.x-v->x;
       vy=pullpoint.y-v->y;
       vz=pullpoint.z-v->z;

       mult=exp(-pull_a*(vx*vx+vy*vy+vz*vz));
       
       v->x+= mult*px;
       v->y+= mult*py;
       v->z+= mult*pz;
}



void pull_selected(int view,VTX pullpoint,int px, int py){
    int i;
    float fpx=0,fpy=0,fpz=0;
    switch(view){
       case XY:
          fpx=px;fpy=py;fpz=0;
       break;
       case ZY:
          fpx=0;fpy=py;fpz=px;
       break;
       case XZ:
          fpx=px;fpy=0;fpz=py;
       break;
    }

    for(i=0;i<num_selected;i++){
       if(curVertex[i]!=moving_vertex){
            pull_vertex(curVertex[i],pullpoint,fpx,fpy,fpz);
       }
    }
}
void pull_all(int view,VTX pullpoint,int px, int py){
    int i;
    float fpx=0,fpy=0,fpz=0;
    switch(view){
       case XY:
          fpx=px;fpy=py;fpz=0;
       break;
       case ZY:
          fpx=0;fpy=py;fpz=px;
       break;
       case XZ:
          fpx=px;fpy=0;fpz=py;
       break;
    }

    for(i=0;i<mesh->num_vertices;i++){

       if(i!=moving_vertex)
         pull_vertex(i,pullpoint,fpx,fpy,fpz);
    }
}

void mirror_selected(int plane_norm){
     int i;
     VTX *v;
     for (i=0;i<num_selected;i++){
         v=&(mesh->points[curVertex[i]]);
         if(plane_norm==X)
             v->x=-(v->x);
         else if(plane_norm==Y)
             v->y=-(v->y);
         else if(plane_norm==Z)
             v->z=-(v->z);
     }
}

void rotate_vertex(int view, int i,float cosr, float sinr){
     float x,y;
     VTX *v=&(mesh->points[i]);
     switch (view){
          case XYVIEW:
               x =  (v->x)*cosr - (v->y)*sinr;
               y =  (v->x)*sinr + (v->y)*cosr;
               v->x=x; v->y=y;
          break;
          case ZYVIEW:
               x =  (v->z)*cosr - (v->y)*sinr;
               y =  (v->z)*sinr + (v->y)*cosr;
               v->z=x; v->y=y;
          break;
          case XZVIEW:
               x =  (v->x)*cosr - (v->z)*sinr;
               y =  (v->x)*sinr + (v->z)*cosr;
               v->x=x; v->z=y;
          break;

     }
     
}
void rotate_vertex_about(int view, int i,float cosr, float sinr,VTX center){
     float x,y;
     VTX *v=&(mesh->points[i]);
     v->x-=center.x;
     v->y-=center.y;
     v->z-=center.z;
     switch (view){
          case XYVIEW:
               x =  (v->x)*cosr - (v->y)*sinr;
               y =  (v->x)*sinr + (v->y)*cosr;
               v->x=x; v->y=y;
          break;
          case ZYVIEW:
               x =  (v->z)*cosr - (v->y)*sinr;
               y =  (v->z)*sinr + (v->y)*cosr;
               v->z=x; v->y=y;
          break;
          case XZVIEW:
               x =  (v->x)*cosr - (v->z)*sinr;
               y =  (v->x)*sinr + (v->z)*cosr;
               v->x=x; v->z=y;
          break;

     }
     v->x+=center.x;
     v->y+=center.y;
     v->z+=center.z;
     
}

void rotate_selected(int view,float radians){

     int i;
     float cosr=cos(radians),sinr=sin(radians);
     for (i=0;i<num_selected;i++){
             rotate_vertex(view,curVertex[i],cosr,sinr);
     }
}
void selected_CG(VTX *cg){
     int i;
     cg->x=0;cg->y=0;cg->z=0;
     for (i=0;i<num_selected;i++){
             cg->x+=mesh->points[curVertex[i]].x;
             cg->y+=mesh->points[curVertex[i]].y;
             cg->z+=mesh->points[curVertex[i]].z;
     }
     cg->x/=(float)num_selected;
     cg->y/=(float)num_selected;
     cg->z/=(float)num_selected;
     
}
void rotate_selected_cg(int view,float radians){

     int i;
     float cosr=cos(radians),sinr=sin(radians);
     VTX cg;
     selected_CG(&cg);
     for (i=0;i<num_selected;i++){
             rotate_vertex_about(view,curVertex[i],cosr,sinr,cg);
     }
}

void rotate_all(int view,float radians){

     int i;
     float cosr=cos(radians),sinr=sin(radians);     
     for (i=0;i<mesh->num_vertices;i++){
             rotate_vertex(view,i,cosr,sinr);
     }
}
void rotate_all_cg(int view,float radians){

     int i;
     int n_v=mesh->num_vertices;
     float cosr=cos(radians),sinr=sin(radians);
     VTX cg;
     mesh_CG(mesh,&cg);
     
     for (i=0;i<n_v;i++){
             rotate_vertex_about(view,i,cosr,sinr,cg);
     }
}

float nearest_snap(float val,float snap){
   float d=val/snap;
   float f=floorf(d);
   float ret;
   if(d-f<.5) ret=(snap*f);
   else ret=snap*(f+1);
   return ret;
}


void scale_vertex(int vtxnum, int direction,float scale ,VTX scenter){

     float x,y,z;

     VTX *v;

     v=&(mesh->points[vtxnum]);
     v->x-=scenter.x;
     v->y-=scenter.y;
     v->z-=scenter.z;
     switch (direction){
          case X:
               x =  (v->x)*scale;
               v->x=x;
               break;
          case Y:
               y =  (v->y)*scale;
               v->y=y;
               break;
          case Z:
               z = (v->z)*scale;
               v->z=z;
               break;
     }
     v->x+=scenter.x;
     v->y+=scenter.y;
     v->z+=scenter.z;
}

void meld_vertex_pair(int i,int j){
    int s,t,m,k,f=0;
    k=i;m=j;
    for(s=0;s<num_selected;s++){
        if(curVertex[s]==m){
        /* if the vertex that is being absorbed
           is a selected vertex, remove it from
           the list of selected, but put the
           absorbER in to the list */
                f=1;
                unselect_vertex(s);
        }
    }
    for(t=0;t<num_selected;t++){
         if(curVertex[t]>m)
             curVertex[t]--;
    }
    if (k>m) k--;
    if(f) add_to_selected(k,0);
    
    meld_vertices(mesh,i,j);

}
void scale_selected(int view,float xscale, float yscale, int unif_p){
      int s; float uscale;
      VTX cg;

      if(xscale==0 && yscale==0) return;

      selected_CG(&cg);
           
      for(s=0;s<num_selected;s++){
        if(unif_p){
               uscale=yscale+xscale;
               scale_vertex(curVertex[s], X,uscale/2,cg);
               scale_vertex(curVertex[s], Y,uscale/2,cg);
               scale_vertex(curVertex[s], Z,uscale/2,cg);
        
        }
        else{
          switch(view){
            case XYVIEW:
               scale_vertex(curVertex[s], X,xscale,cg);
               scale_vertex(curVertex[s], Y,yscale,cg);
               break;
            case ZY:
               scale_vertex(curVertex[s], Z,xscale,cg);
               scale_vertex(curVertex[s], Y,yscale,cg);
               break;

            case XZ:
               scale_vertex(curVertex[s], X,xscale,cg);
               scale_vertex(curVertex[s], Z,yscale,cg);
               break;
          }
        }
      }
}

void scale_all(int view,float xscale, float yscale, int unif_p){
      int vnum; float uscale;
      if(xscale==0 && yscale==0) return;
      for(vnum=0;vnum<mesh->num_vertices;vnum++){
        if(unif_p){
               uscale=yscale+xscale;
               scale_vertex(vnum, X,uscale/2, Origin);
               scale_vertex(vnum, Y,uscale/2, Origin);
               scale_vertex(vnum, Z,uscale/2, Origin);
        
        }
        else{
          switch(view){
            case XYVIEW:
               scale_vertex(vnum, X,xscale, Origin);
               scale_vertex(vnum, Y,yscale, Origin);
               break;
            case ZY:
               scale_vertex(vnum, Z,xscale, Origin);
               scale_vertex(vnum, Y,yscale, Origin);
               break;

            case XZ:
               scale_vertex(vnum, X,xscale, Origin);
               scale_vertex(vnum, Z,yscale, Origin);
               break;
          }
        }
      }
}

float snapval=10.0;
int snapx_p=1;
int snapy_p=1;
int snapz_p=1;
void snap_all(float snap){
   int vi;
   VTX *v;

   for(vi=0;vi<mesh->num_vertices;vi++){
       v=&(mesh->points[vi]);
       if(snapx_p)
          v->x=nearest_snap(v->x,snap);
       if(snapy_p)
          v->y=nearest_snap(v->y,snap);
       if(snapz_p)
          v->z=nearest_snap(v->z,snap);
   }
}
void snap_selected(float snap){
   int si;
   VTX *v;

   for(si=0;si<num_selected;si++){
       v=&(mesh->points[curVertex[si]]);
       if(snapx_p)
          v->x=nearest_snap(v->x,snap);
       if(snapy_p)
          v->y=nearest_snap(v->y,snap);
       if(snapz_p)
          v->z=nearest_snap(v->z,snap);
   }
}

/* just about the whole rest of this file
   is based around Allegro GUI routines */

int load_tri();
int save_triangles_as();
int save_tri();
int save_selected_tri();
int do_get_undo();
int quitter();

int do_delete_all();

MENU file_menu[]={
   {"&New",          do_delete_all,    NULL,0,NULL},
   {"",NULL,NULL,0,NULL},   
   {"&Open",         load_tri,         NULL,0,NULL},
   {"",NULL,NULL,0,NULL},   
   {"&Save",         save_tri,         NULL,0,NULL},
   {"Save se&Lected",save_selected_tri,NULL,0,NULL},
   {"",NULL,NULL,0,NULL},
   {"Save &As",      save_triangles_as,NULL,0,NULL},
   {"",NULL,NULL,0,NULL},
   {"&Get Undo",     do_get_undo,      NULL,0,NULL},
   {"",NULL,NULL,0,NULL},
   {"&Quit",         quitter,          NULL,0,NULL},
   { NULL,NULL,NULL,0,NULL}
};

int do_make_triangle();
int do_delete_triangle();
int do_delete_selected_triangles();
int do_dup_selected();
int do_extrude_selected();
int do_reflect_selected();
int do_delete_selected();
int do_meld_selected();
int do_meld_coincident();
int do_reduce_one();

int do_mark_selected();
int do_unmark_selected();
int do_select_all();
int do_unselect_all();
int do_invert_selection();
int do_try_orient();
int do_reverse_selected();
int do_cleanup_loose_vertices();
int do_deform_menu();
int do_divide();
int do_perturb();
int do_reduce();
int do_wave();
int do_ripples();
int do_twist();
int do_bend();
int do_repulse_point();
int do_smoothsubdivide();

int do_make_copy_of_selected();
int do_paste_copy();
int orientedlighting();
int tabbedshiftselected();
int do_snap();

int do_detach_selected();
int dummy_keyb_proc(int msg, DIALOG *d, int c);

MENU deform_menu[];
MENU edit_menu[]={

   {"make &Triangle\tCtrl-t",             do_make_triangle,         NULL,D_DISABLED,NULL},
   {"",NULL,NULL,0,NULL},
   {"&Select all\tCtrl-s",                do_select_all,            NULL,D_DISABLED,NULL},
   {"&Unselect all\tCtrl-u",              do_unselect_all,          NULL,D_DISABLED,NULL},
   {"&Invert selection\tCtrl-i",          do_invert_selection,      NULL,D_DISABLED,NULL},

   {"",NULL,NULL,0,NULL},   
   {"Copy selected\tCtrl-c",              do_make_copy_of_selected, NULL,D_DISABLED,NULL},
   {"paste copy\tCtrl-v",                 do_paste_copy,            NULL,D_DISABLED,NULL},
   {"du&P selected\tCtrl-p",              do_dup_selected,          NULL,D_DISABLED,NULL},
   {"detach selected\tCtrl-b",            do_detach_selected,       NULL,D_DISABLED,NULL},
   {"",NULL,NULL,0,NULL},      
   {"&Reflect selected\tCtrl-r",          do_reflect_selected,      NULL,D_DISABLED,NULL},

   {"",NULL,NULL,0,NULL},   
   {"&Meld selected\tCtrl-m",             do_meld_selected,         NULL,D_DISABLED,NULL},
   {"",NULL,NULL,0,NULL},   
   {"e&Xtrude selection\tCtrl-x",         do_extrude_selected,      NULL,D_DISABLED,NULL},
   {"",NULL,NULL,0,NULL},
   {"Smooth/Reduce/&Deform\tCtrl-d",      do_deform_menu,           NULL,D_DISABLED,NULL},
   {"",NULL,NULL,0,NULL},
   {"snap to grid\tCtrl-y",               do_snap,                  NULL,D_DISABLED,NULL},
   {"",NULL,NULL,0,NULL},
   {"remove 'loose' vertices\tCtrl-l",    do_cleanup_loose_vertices,NULL,D_DISABLED,NULL},
   {"meld coincident vertices\tCtrl-w",   do_meld_coincident,       NULL,D_DISABLED,NULL},
   {"",NULL,NULL,0,NULL},   

   {"m&Ark selected\tCtrl-a",             do_mark_selected,         NULL,D_DISABLED,NULL},
   {"unmar&K selected\tCtrl-k",           do_unmark_selected,       NULL,D_DISABLED,NULL},
   {"",NULL,NULL,0,NULL},   
   {"try Orient\to",                      do_try_orient,            NULL,D_DISABLED,NULL},
   {"&Evert selected\tCtrl-e",            do_reverse_selected,      NULL,D_DISABLED,NULL},

   {"",NULL,NULL,0,NULL},
   {"delete trian&Gle\tCtrl-g",           do_delete_triangle,       NULL,D_DISABLED,NULL},
   {"delete selected triangles\tCtrl-j",  do_delete_selected_triangles,      NULL,D_DISABLED,NULL},
   {"&Delete selection\tDel",             do_delete_selected,       NULL,D_DISABLED,NULL},
   {"delete all",                         do_delete_all,            NULL,0,NULL},
   
   { NULL,NULL,NULL,0,NULL}
};

#define MAKETRIANGLEITEM 0

#define SELECTALLITEM 2
#define UNSELECTALLITEM 3
#define INVERTSELECTIONITEM 4

#define COPYSELECTEDITEM 6
#define PASTECOPYITEM 7
#define DUPSELECTEDITEM 8
#define DETACHSELECTEDITEM 9

#define REFLECTSELECTEDITEM 11

#define MELDSELECTEDITEM 13
#define EXTRUDESELECTEDITEM 15
#define DEFORMITEM 17
#define SNAPITEM 19
#define REMOVELOOSEITEM 21
#define MELDCOINCIDENTITEM 22

#define MARKSELECTEDITEM 24
#define UNMARKSELECTEDITEM 25

#define ORIENTITEM 27
#define EVERTSELECTEDITEM 28

#define DELETETRIANGLEITEM 30
#define DELETESELECTEDTRIANGLESITEM 31
#define DELETESELECTEDITEM 32
#define DELETEALLITEM 32

int do_scaleuniformitemtoggle();
int do_rotateaboutitemtoggle();
int do_show3daxes_toggle();
int do_display_resolution_menu();
int do_showflatvertices_toggle();
int do_pull_dialog();
MENU shading_palette_menu[];

MENU display_resolution_menu[]={
   {"640x480",NULL,NULL,0,NULL},
   {"800x600",NULL,NULL,0,NULL},
   {"1024x768",NULL,NULL,0,NULL},
   {"1280x1024",NULL,NULL,0,NULL},
   {"1600x1200",NULL,NULL,0,NULL},
   { NULL,NULL,NULL,0,NULL}
};
#define RES640x480   0
#define RES800x600   1
#define RES1024x768  2
#define RES1280x1024 3
#define RES1600x1200 4

MENU options_menu[]={
   {"  &Uniform scaling",                    do_scaleuniformitemtoggle,NULL,D_SELECTED,NULL},
   {"  &Rotations about view center",do_rotateaboutitemtoggle, NULL,0,NULL},
   {"  show &Vertices",                      do_showflatvertices_toggle,     NULL,D_SELECTED,NULL},
   {"  show 3d &Axes",                       do_show3daxes_toggle,     NULL,D_SELECTED,NULL},
   { "",NULL,NULL,0,NULL},   
   {"&Pull parameters",                      do_pull_dialog,           NULL,0,NULL},

   { "",NULL,NULL,0,NULL},
   {"&Display Resolution",do_display_resolution_menu,NULL,0,NULL},
   { "",NULL,NULL,0,NULL},
   {"&ShadingPalette",     NULL,shading_palette_menu,0,NULL},

   { NULL,NULL,NULL,0,NULL}
};
#define UNIFORMSCALINGITEM 0
#define ROTATEABOUTITEM 1
#define SHOWFLATVERTICESITEM 2
#define SHOW3dAXESITEM 3

MENU deform_menu[]={
{"",NULL,NULL,0,NULL},
   {"divide (flat)\tCtrl-f",        do_divide,NULL,0,NULL},
   {"",NULL,NULL,0,NULL},   
   {"smoot&H subdivision\tCtrl-h",  do_smoothsubdivide,NULL,0,NULL},
   {"",NULL,NULL,0,NULL},
   {"mesh reductio&N\tCtrl-n",      do_reduce,NULL,0,NULL},
   {"",NULL,NULL,0,NULL},
   {"&Repulsive point",     do_repulse_point,NULL,0,NULL},
   {"rou&Ghen",             do_perturb,      NULL,0,NULL},
   {"&Wave",                do_wave,         NULL,0,NULL},
   {"ri&Pples",             do_ripples,      NULL,0,NULL},
   {"&Twist",               do_twist,        NULL,0,NULL},
   {"&Bend",                do_bend,         NULL,0,NULL},
{"",NULL,NULL,0,NULL},
   { NULL,NULL,NULL,0,NULL}
};

int do_deform_menu(){
      int mret;
      mret=do_menu(deform_menu,mouse_x,mouse_y);
      return D_O_K;
}

int do_scaleuniformitemtoggle(){
    if(options_menu[UNIFORMSCALINGITEM].flags & D_SELECTED){
       options_menu[UNIFORMSCALINGITEM].flags &= ~D_SELECTED;
    }
    else{
       options_menu[UNIFORMSCALINGITEM].flags |= D_SELECTED;
    }
    return D_O_K;
}

int do_rotateaboutitemtoggle(){
    if(options_menu[ROTATEABOUTITEM].flags & D_SELECTED){
       options_menu[ROTATEABOUTITEM].flags &= ~D_SELECTED;
    }
    else{
       options_menu[ROTATEABOUTITEM].flags |= D_SELECTED;
    }
    return D_O_K;
}
void redraw_3dview();
int do_show3daxes_toggle(){
    if(options_menu[SHOW3dAXESITEM].flags & D_SELECTED){
       options_menu[SHOW3dAXESITEM].flags &= ~D_SELECTED;
       show3daxes=0;
    }
    else{
       options_menu[SHOW3dAXESITEM].flags |= D_SELECTED;
       show3daxes=1;
    }
    redraw_3dview();
    return D_O_K;
}
void redraw_flatviews();
int do_showflatvertices_toggle(){
    if(options_menu[SHOWFLATVERTICESITEM].flags & D_SELECTED){
       options_menu[SHOWFLATVERTICESITEM].flags &= ~D_SELECTED;
       show_vertices=0;
    }
    else{
       options_menu[SHOWFLATVERTICESITEM].flags |= D_SELECTED;
       show_vertices=1;
    }
    redraw_flatviews();
    return D_O_K;
}

MENU top_menu[]={
   {"&File",NULL,file_menu,0,NULL},
   {"&Edit",NULL,edit_menu,0,NULL},   
   {"&Options",NULL,options_menu,0,NULL},
   { NULL,NULL,NULL,0,NULL}
};

char numverticestxt[50]="vertices:";
char numfacestxt[50]=   "triangles:";
char numseltxt[50]=     "selected:";

int layoutscreen_proc(int msg, DIALOG *d, int c);
int view_proc(int msg, DIALOG *d, int c);
int redraw_all_button_proc(int msg, DIALOG *d UNUSED, int c);
int poly_button_proc(int msg, DIALOG *d, int c);
int deform_button_proc(int msg, DIALOG *d, int c);
int deform_icon_proc(int msg, DIALOG *d, int c);
int dump_screen();
char mousetxt[256]="";
#define DBG 200
#define DFG 172
#define DButBG 220 /* dialog button background */
#define C(x) (x-'a'+1)

/* dialog element sizes are set in a 'startup' routine */
DIALOG the_dialog[] =
{
   // (dialog proc)       (x)  (y)   (w)  (h)   (fg) (bg)  (key)  (flags)  (d1)    (d2) (dp)

   { layoutscreen_proc,    0,   0,    0,   0,    DFG,  DBG,   0,      0,      0,       0, NULL ,NULL,NULL},

   { view_proc,            0,   0,    0,   0,   255,    0,   0,      0,     XYVIEW,   0, NULL, NULL,NULL},
   { view_proc,            0,   0,    0,   0,   255,    0,   0,      0,     ZYVIEW,   0, NULL, NULL ,NULL},
   { view_proc,            0,   0,    0,   0,   255,    0,   0,      0,     XZVIEW,   0, NULL, NULL ,NULL},
   { view_proc,            0,   0,    0,   0,   255,    0,   0,      0,    VIEW_3D,   0, NULL ,NULL,NULL},
   { d_box_proc,
                           0,   0,    0,   0,   DFG,  DBG,   0,      0,          0,   0,  NULL,NULL,NULL},
   { d_text_proc,
                           0,  60,  160,  12,     0,  DBG,   0,      0,          0,    0, numverticestxt ,NULL,NULL},
   { d_text_proc,
                           0,  88,  160,  12,     0,  DBG,   0,      0,          0,    0, numfacestxt ,NULL,NULL},
   { d_text_proc,
                           0,  74,  160,  12,     0,  DBG,   0,      0,          0,    0, numseltxt ,NULL,NULL},
   { d_textbox_proc,
                           0, 110,  140,  36,     0,  DBG,   0,      0,          0,    0,  mousetxt ,NULL,NULL},
   { d_bitmap_proc,
                           0, 160,   32,  32,     0,  DBG,   0,      0,          0,    0,  NULL ,NULL,NULL},
   { d_menu_proc,
                           0,   0,    0,   0,     0,    0,   0,      0,          0,    0,   top_menu,NULL,NULL},
   { poly_button_proc,
                           0, 160,   70,  20,     0,DButBG,  0, D_EXIT,          0,    0,  "Polygon" ,NULL,NULL},
   { deform_icon_proc,
                           0, 200,  170,  20,     0,DButBG,  0,D_EXIT|D_DISABLED,0,    0,  NULL,NULL,NULL},
   { redraw_all_button_proc,
                           0, 270,   70,  20,     0,DButBG,  0,D_EXIT,0,    0,  "Redraw",NULL,NULL},
                           
/* keyboard shortcuts for the edit menu items */
   { d_keyboard_proc,      0,   0,     0,   0,     0,    0,         0,    0,KEY_TAB,      0, tabbedshiftselected,NULL,NULL},
   { d_keyboard_proc,      0,   0,     0,   0,     0,    0,    C('t'),    0,      0,      0, do_make_triangle,NULL,NULL},
   { d_keyboard_proc,      0,   0,     0,   0,     0,    0,    C('g'),    0,      0,      0, do_delete_triangle,NULL,NULL},
   { d_keyboard_proc,      0,   0,     0,   0,     0,    0,    C('j'),    0,      0,      0, do_delete_selected_triangles,NULL,NULL},
   { d_keyboard_proc,      0,   0,     0,   0,     0,    0,    C('p'),    0,      0,      0, do_dup_selected,NULL,NULL},
   { d_keyboard_proc,      0,   0,     0,   0,     0,    0,    C('r'),    0,      0,      0, do_reflect_selected,NULL,NULL},
   { d_keyboard_proc,      0,   0,     0,   0,     0,    0,    C('q'),    0,      0,      0, do_reduce_one,NULL,NULL},
   { d_keyboard_proc,      0,   0,     0,   0,     0,    0,    C('m'),    0,      0,      0, do_meld_selected,NULL,NULL},
   { d_keyboard_proc,      0,   0,     0,   0,     0,    0,    C('x'),    0,      0,      0, do_extrude_selected,NULL,NULL},
   { d_keyboard_proc,      0,   0,     0,   0,     0,    0,    C('a'),    0,      0,      0, do_mark_selected,NULL,NULL},
   { d_keyboard_proc,      0,   0,     0,   0,     0,    0,    C('k'),    0,      0,      0, do_unmark_selected,NULL,NULL},
   { d_keyboard_proc,      0,   0,     0,   0,     0,    0,    C('s'),    0,      0,      0, do_select_all,NULL,NULL},
   { d_keyboard_proc,      0,   0,     0,   0,     0,    0,    C('u'),    0,      0,      0, do_unselect_all,NULL,NULL},
   { d_keyboard_proc,      0,   0,     0,   0,     0,    0,         0,    0,      0,      0, do_invert_selection,NULL,NULL},
   { d_keyboard_proc,      0,   0,     0,   0,     0,    0,       'o',    0,      0,      0, do_try_orient,NULL,NULL},
   { d_keyboard_proc,      0,   0,     0,   0,     0,    0,    C('e'),    0,      0,      0, do_reverse_selected,NULL,NULL},
   { d_keyboard_proc,      0,   0,     0,   0,     0,    0,    C('f'),    0,      0,      0, do_divide,NULL,NULL},
   { d_keyboard_proc,      0,   0,     0,   0,     0,    0,    C('d'),    0,      0,      0, do_deform_menu,NULL,NULL},
   { d_keyboard_proc,      0,   0,     0,   0,     0,    0,    C('h'),    0,      0,      0, do_smoothsubdivide,NULL,NULL},
   { d_keyboard_proc,      0,   0,     0,   0,     0,    0,    C('n'),    0,      0,      0, do_reduce,NULL,NULL},
   { d_keyboard_proc,      0,   0,     0,   0,     0,    0,       'f',    0,      0,      0, orientedlighting,NULL,NULL},
   { d_keyboard_proc,      0,   0,     0,   0,     0,    0,         0,    0,KEY_DEL,      0, do_delete_selected,NULL,NULL},
   { d_keyboard_proc,      0,   0,     0,   0,     0,    0,        27,    0,      0,      0, quitter,NULL,NULL},
   { d_keyboard_proc,      0,   0,     0,   0,     0,    0,    C('y'),    0,      0,      0, do_snap,NULL,NULL},
   { d_keyboard_proc,      0,   0,     0,   0,     0,    0,    C('l'),    0,      0,      0, do_cleanup_loose_vertices,NULL,NULL},
   { d_keyboard_proc,      0,   0,     0,   0,     0,    0,    C('w'),    0,      0,      0, do_meld_coincident,NULL,NULL},
   { d_keyboard_proc,      0,   0,     0,   0,     0,    0,    C('c'),    0,      0,      0, do_make_copy_of_selected,NULL,NULL},
   { d_keyboard_proc,      0,   0,     0,   0,     0,    0,    C('v'),    0,      0,      0, do_paste_copy,NULL,NULL},
   { d_keyboard_proc,      0,   0,     0,   0,     0,    0,    C('z'),    0,      0,      0, dump_screen,NULL,NULL},
   { d_keyboard_proc,      0,   0,     0,   0,     0,    0,    C('b'),    0,      0,      0, do_detach_selected,NULL,NULL},
   { dummy_keyb_proc,      0,   0,     0,   0,     0,    0,         0,    0,      0,      0, NULL,NULL,NULL},
   { NULL,                 0,   0,     0,   0,     0,    0,         0,    0,      0,      0, NULL ,NULL,NULL}
};
#define MAINLAYOUT 0
#define XYPROC 1
#define ZYPROC 2
#define XZPROC 3
#define THREEDPROC 4
#define RIGHTBIGBOX 5
#define NUMVTXTEXT 6
#define NUMTRITEXT 7
#define NUMSELTEXT 8
#define MOUSETEXT 9
#define TRISELBMP 10
#define MAINMENU 11
#define POLYGONBUTTON 12
#define DEFORMBUTTON 13
#define REDRAWBUTTON 14

MENU mousemodeflat_menu[]={
{"",NULL,NULL,0,NULL},
   {"Create vertex \tc",NULL,NULL,0,NULL},
   
{"",NULL,NULL,0,NULL},
   {"Select vertex \ts",NULL,NULL,0,NULL},
{"",NULL,NULL,0,NULL},   
   {"Unselect vertex \tu",NULL,NULL,0,NULL},
{"",NULL,NULL,0,NULL},
   {"select in rectanGle \tg",NULL,NULL,0,NULL},
{"",NULL,NULL,0,NULL},
   {"unselect in rectanGle \tAlt-g",NULL,NULL,0,NULL},

{"",NULL,NULL,0,NULL},
   {"selecT connecTed componenT \tt",NULL,NULL,0,NULL},
{"",NULL,NULL,0,NULL},
   {"unselecT connecTed componenT \tAlt-t",NULL,NULL,0,NULL},
   
{"",NULL,NULL,0,NULL},
   {"mousemoVe \tv",NULL,NULL,0,NULL},
{"",NULL,NULL,0,NULL},
   {"Move selected \tm",NULL,NULL,0,NULL},
{"",NULL,NULL,0,NULL},   
   {"Move all \tAlt-m",NULL,NULL,0,NULL},

{"",NULL,NULL,0,NULL},
   {"Pull selected \tp",NULL,NULL,0,NULL},
{"",NULL,NULL,0,NULL},   
   {"Pull all \tAlt-p",NULL,NULL,0,NULL},

{"",NULL,NULL,0,NULL},   
   {"pull By edges (selected)\tb",NULL,NULL,0,NULL},
{"",NULL,NULL,0,NULL},   
   {"pull By edges (all)\tAlt-b",NULL,NULL,0,NULL},
   
{"",NULL,NULL,0,NULL},   
   {"repulsive mouse (pusH)\th",NULL,NULL,0,NULL},
   
{"",NULL,NULL,0,NULL},
   {"Rotate selected \tr",NULL,NULL,0,NULL},
{"",NULL,NULL,0,NULL},   
   {"Rotate all \tAlt-r",NULL,NULL,0,NULL},
{"",NULL,NULL,0,NULL},   

   {"scaLe selected \tl",NULL,NULL,0,NULL},
{"",NULL,NULL,0,NULL},   
   {"scaLe all \tAlt-l",NULL,NULL,0,NULL},
{"",NULL,NULL,0,NULL},
   {"Insert vertex \ti",NULL,NULL,0,NULL},
{"",NULL,NULL,0,NULL},

   {"Delete vertex \td",NULL,NULL,0,NULL},
   { NULL,NULL,NULL,0,NULL}
};

#define MOUSEMODEFLAT_CREATEVERTEX 1

#define MOUSEMODEFLAT_SELECTVERTEX 3
#define MOUSEMODEFLAT_UNSELECTVERTEX 5
#define MOUSEMODEFLAT_SELECTRECTANGLE 7
#define MOUSEMODEFLAT_UNSELECTRECTANGLE 9

#define MOUSEMODEFLAT_SELECTCONNECTED 11
#define MOUSEMODEFLAT_UNSELECTCONNECTED 13

#define MOUSEMODEFLAT_MOUSEMOVEVERTEX 15
#define MOUSEMODEFLAT_MOVESELECTED 17
#define MOUSEMODEFLAT_MOVEALL 19

#define MOUSEMODEFLAT_PULLSELECTED 21
#define MOUSEMODEFLAT_PULLALL 23
#define MOUSEMODEFLAT_PULLEDGESSELECTED 25
#define MOUSEMODEFLAT_PULLEDGES 27

#define MOUSEMODEFLAT_REPULSE 29

#define MOUSEMODEFLAT_ROTATESELECTED 31
#define MOUSEMODEFLAT_ROTATEALL 33

#define MOUSEMODEFLAT_SCALESELECTED 35
#define MOUSEMODEFLAT_SCALEALL 37

#define MOUSEMODEFLAT_INSERTVERTEX 39
#define MOUSEMODEFLAT_DELETEVERTEX 41

MENU mousemode3d_menu[]={
   {"Rotate (1)",NULL,NULL,0,NULL},
{"",NULL,NULL,0,NULL},   
   {"Rotate (2)",NULL,NULL,0,NULL},
{"",NULL,NULL,0,NULL},   
   {"Pan left/right in/out",NULL,NULL,0,NULL},
{"",NULL,NULL,0,NULL},   
   {"Pan left/right up/down",NULL,NULL,0,NULL},
{"",NULL,NULL,0,NULL},   
   {"Scale",NULL,NULL,0,NULL},

   
   { NULL,NULL,NULL,0,NULL}
};

#define MOUSEMODE3d_ROTATE0 0
#define MOUSEMODE3d_ROTATE1 2
#define MOUSEMODE3d_TRANSLATEXZ 4
#define MOUSEMODE3d_TRANSLATEXY 6
#define MOUSEMODE3d_SCALE 8

int mouse_mode_3d=MOUSEMODE3d_ROTATE0;
int mouse_mode_flat;
void do_repulse_mouse_dialog();

void set_mouse_mode_flat(int mode){
    int mmi=0;
    mouse_mode_flat=mode;
    while(mousemodeflat_menu[mmi].text!=NULL){
        if(mmi==mode)
           mousemodeflat_menu[mmi].flags|=D_SELECTED;
        else
           mousemodeflat_menu[mmi].flags&=~D_SELECTED;
        mmi++;
    }
    if(mode==MOUSEMODEFLAT_REPULSE){
        do_repulse_mouse_dialog();

    }
}
void set_mouse_mode_3d(int mode){
    int mmi=0;
    mouse_mode_3d=mode;
    while(mousemode3d_menu[mmi].text!=NULL){
        if(mmi==mode)
           mousemode3d_menu[mmi].flags|=D_SELECTED;
        else
           mousemode3d_menu[mmi].flags&=~D_SELECTED;
        mmi++;
    }
}

int dummy_keyb_proc(int msg, DIALOG *d UNUSED, int c){
    int s,a,ctrl;
    if(msg==MSG_XCHAR){
    
            a=key_shifts&KB_ALT_FLAG;
            s=key_shifts&KB_SHIFT_FLAG;
            ctrl=key_shifts&KB_CTRL_FLAG;
            switch((c>>8)-1+'a'){
                case 'c':
                    set_mouse_mode_flat(MOUSEMODEFLAT_CREATEVERTEX);
                break;
                case 's':
                    set_mouse_mode_flat(MOUSEMODEFLAT_SELECTVERTEX);
                break;
                case 'u':
                    set_mouse_mode_flat(MOUSEMODEFLAT_UNSELECTVERTEX);
                break;                
                case 'g':
                    if(a)
                      set_mouse_mode_flat(MOUSEMODEFLAT_UNSELECTRECTANGLE);
                    else
                      set_mouse_mode_flat(MOUSEMODEFLAT_SELECTRECTANGLE);
                break;
                case 't':
                    if(a)
                      set_mouse_mode_flat(MOUSEMODEFLAT_UNSELECTCONNECTED);
                    else
                      set_mouse_mode_flat(MOUSEMODEFLAT_SELECTCONNECTED);
                break;
                case 'v':
                    set_mouse_mode_flat(MOUSEMODEFLAT_MOUSEMOVEVERTEX);
                break;
                case 'h':
                    set_mouse_mode_flat(MOUSEMODEFLAT_REPULSE);
                break;

                case 'm':
                    if(a)
                      set_mouse_mode_flat(MOUSEMODEFLAT_MOVEALL);
                    else
                      set_mouse_mode_flat(MOUSEMODEFLAT_MOVESELECTED);

                break;
                case 'p':
                    if(a)
                      set_mouse_mode_flat(MOUSEMODEFLAT_PULLALL);
                    else
                      set_mouse_mode_flat(MOUSEMODEFLAT_PULLSELECTED);

                break;
                case 'b':
                    if(a)
                      set_mouse_mode_flat(MOUSEMODEFLAT_PULLEDGES);
                    else
                      set_mouse_mode_flat(MOUSEMODEFLAT_PULLEDGESSELECTED);
                break;

                case 'r':
                    if(a)
                      set_mouse_mode_flat(MOUSEMODEFLAT_ROTATEALL);
                    else
                      set_mouse_mode_flat(MOUSEMODEFLAT_ROTATESELECTED);
                break;
                case 'l':
                    if(a)
                      set_mouse_mode_flat(MOUSEMODEFLAT_SCALEALL);
                    else
                      set_mouse_mode_flat(MOUSEMODEFLAT_SCALESELECTED);

                break;
                case 'i':
                   if(ctrl){
                     do_invert_selection();
                   }
                   else
                     set_mouse_mode_flat(MOUSEMODEFLAT_INSERTVERTEX);

                break;
                case 'd':
                   set_mouse_mode_flat(MOUSEMODEFLAT_DELETEVERTEX);

                break;
                default:
                   return D_O_K;
            }
            return D_USED_CHAR;
    }
    return D_O_K;
}

void enabledisableitems(){
/* When the mesh changes, some menu items and options become
   enabled or disabled */
   int f;
   if(mesh->num_vertices){
     edit_menu[SELECTALLITEM].flags  &=~D_DISABLED;
     edit_menu[UNSELECTALLITEM].flags  &=~D_DISABLED;
     edit_menu[INVERTSELECTIONITEM].flags  &=~D_DISABLED;
     edit_menu[DEFORMITEM].flags  &=~D_DISABLED;
     edit_menu[SNAPITEM].flags  &=~D_DISABLED;
     edit_menu[REMOVELOOSEITEM].flags  &=~D_DISABLED;
     edit_menu[MELDCOINCIDENTITEM].flags  &=~D_DISABLED;
     edit_menu[ORIENTITEM].flags  &= ~D_DISABLED;
     the_dialog[DEFORMBUTTON].flags &= ~D_DISABLED;
   }
   else{
     edit_menu[SELECTALLITEM].flags |= D_DISABLED;
     edit_menu[UNSELECTALLITEM].flags |=D_DISABLED;
     edit_menu[INVERTSELECTIONITEM].flags |=D_DISABLED;
     edit_menu[DEFORMITEM].flags |=D_DISABLED;
     edit_menu[SNAPITEM].flags  |=D_DISABLED;
     edit_menu[REMOVELOOSEITEM].flags  |=D_DISABLED;
     edit_menu[MELDCOINCIDENTITEM].flags  |=D_DISABLED;
     edit_menu[ORIENTITEM].flags  |=D_DISABLED;
     the_dialog[DEFORMBUTTON].flags |=D_DISABLED;
   }
   if(num_selected){
     edit_menu[COPYSELECTEDITEM].flags  &=~D_DISABLED;
     edit_menu[DUPSELECTEDITEM].flags   &=~D_DISABLED;
     edit_menu[DETACHSELECTEDITEM].flags&=~D_DISABLED;
     edit_menu[REFLECTSELECTEDITEM].flags&=~D_DISABLED;
     edit_menu[MELDSELECTEDITEM].flags&=~D_DISABLED;
     edit_menu[EXTRUDESELECTEDITEM].flags&=~D_DISABLED;
     edit_menu[MARKSELECTEDITEM].flags&=~D_DISABLED;
     edit_menu[UNMARKSELECTEDITEM].flags&=~D_DISABLED;
     edit_menu[DELETESELECTEDTRIANGLESITEM].flags&=~D_DISABLED;
     edit_menu[DELETESELECTEDITEM].flags&=~D_DISABLED;
     edit_menu[EVERTSELECTEDITEM].flags&=~D_DISABLED;

   }
   else {
     edit_menu[COPYSELECTEDITEM].flags  |= D_DISABLED;
     edit_menu[DUPSELECTEDITEM].flags   |= D_DISABLED;
     edit_menu[DETACHSELECTEDITEM].flags|= D_DISABLED;
     edit_menu[REFLECTSELECTEDITEM].flags|=D_DISABLED;
     edit_menu[MELDSELECTEDITEM].flags|=D_DISABLED;
     edit_menu[EXTRUDESELECTEDITEM].flags|=D_DISABLED;
     edit_menu[MARKSELECTEDITEM].flags|=D_DISABLED;
     edit_menu[UNMARKSELECTEDITEM].flags|=D_DISABLED;
     edit_menu[DELETESELECTEDTRIANGLESITEM].flags|=D_DISABLED;
     edit_menu[DELETESELECTEDITEM].flags|=D_DISABLED;
     edit_menu[EVERTSELECTEDITEM].flags|=D_DISABLED;
   }
   if(num_selected==3){
      edit_menu[MAKETRIANGLEITEM].flags&= ~D_DISABLED;
      the_dialog[TRISELBMP].dp=threenotribmp;
      f=find_triangle_in_mesh(mesh,curVertex[0],curVertex[1],curVertex[2]);
      if(f>=0){
          edit_menu[DELETETRIANGLEITEM].flags &= ~D_DISABLED;
          edit_menu[MAKETRIANGLEITEM].flags|= D_DISABLED;
          the_dialog[TRISELBMP].dp=threetribmp;
      }
      else{
          edit_menu[DELETETRIANGLEITEM].flags |= D_DISABLED;
      }
   }
   else{
      edit_menu[MAKETRIANGLEITEM].flags |= D_DISABLED;
      edit_menu[DELETETRIANGLEITEM].flags |= D_DISABLED;

      the_dialog[TRISELBMP].dp=notribmp;
      
   }
   SEND_MESSAGE(&(the_dialog[TRISELBMP]),MSG_DRAW,0);

}


void update_mouse_numbers(int view,int mx,int my){
    char mousextxt[50];
    char mouseytxt[50];
    char mouseztxt[50];

    DIALOG *dm=&(the_dialog[MOUSETEXT]);
    switch(view){

        case XYVIEW:
            sprintf(mousextxt,"x %+5.2f",(float)mx);
            sprintf(mouseytxt,"y %+5.2f",(float)-my);
            if(moving_vertex>=0){
               sprintf(mousextxt,"x %+5.2f", mesh->points[moving_vertex].x);
               sprintf(mouseytxt,"y %+5.2f",-mesh->points[moving_vertex].y);
               sprintf(mouseztxt,"z %+5.2f", mesh->points[moving_vertex].z);
            }
            else if(mouse_mode_flat==MOUSEMODEFLAT_REPULSE){
               sprintf(mouseztxt,"z %+5.2f", repulse_mouse_vtx.z);
            }

            else if(num_selected>0)
               sprintf(mouseztxt,"z %+5.2f",mesh->points[curVertex[num_selected-1]].z);
            else sprintf(mouseztxt,"z %+5.2f",0.0);

        break;
        case ZYVIEW:
            sprintf(mouseztxt,"z %+5.2f",(float)mx);
            sprintf(mouseytxt,"y %+5.2f",(float)-my);
            if(moving_vertex>=0){
               sprintf(mousextxt,"x %+5.2f", mesh->points[moving_vertex].x);
               sprintf(mousextxt,"y %+5.2f", mesh->points[moving_vertex].y);
               sprintf(mousextxt,"z %+5.2f", mesh->points[moving_vertex].z);
            }
            else if(mouse_mode_flat==MOUSEMODEFLAT_REPULSE){
               sprintf(mousextxt,"x %+5.2f", repulse_mouse_vtx.x);
            }

            else if(num_selected>0)
               sprintf(mousextxt,"x %+5.2f",mesh->points[curVertex[num_selected-1]].x);
            else sprintf(mousextxt,"x %+5.2f",0.0);
      
        break;
        case XZVIEW:
            sprintf(mousextxt,"x %+5.2f",(float)mx);
            sprintf(mouseztxt,"z %+5.2f",(float)my);
            if(moving_vertex>=0){
               sprintf(mousextxt,"x %+5.2f", mesh->points[moving_vertex].x);
               sprintf(mouseytxt,"y %+5.2f",-mesh->points[moving_vertex].y);
               sprintf(mouseztxt,"z %+5.2f", mesh->points[moving_vertex].z);

            }
            else if(mouse_mode_flat==MOUSEMODEFLAT_REPULSE){
               sprintf(mouseytxt,"y %+5.2f",-repulse_mouse_vtx.y);
            }
            else if(num_selected>0)
               sprintf(mouseytxt,"y %+5.2f",-mesh->points[curVertex[num_selected-1]].y);
            else sprintf(mouseytxt,"y %+5.2f",0.0);
      
        break;
    }
    sprintf(mousetxt,"%s\n%s\n%s",mousextxt,mouseytxt,mouseztxt);
    SEND_MESSAGE(dm,MSG_DRAW,0);

}
void update_numbers(){
    DIALOG *dv=&(the_dialog[NUMVTXTEXT]);
    DIALOG *dt=&(the_dialog[NUMTRITEXT]);
    DIALOG *ds=&(the_dialog[NUMSELTEXT]);
    sprintf(numverticestxt,"vertices:  %5d",mesh->num_vertices);
    sprintf(numfacestxt,   "triangles: %5d",mesh->num_faces);
    sprintf(numseltxt,     "selected:  %5d",num_selected);
    SEND_MESSAGE(dv,MSG_DRAW,0);
    SEND_MESSAGE(dt,MSG_DRAW,0);
    SEND_MESSAGE(ds,MSG_DRAW,0);
}

void redraw_flatviews(){

 int view;
 DIALOG *d;
 for (view=XYPROC;view<=XZPROC;view++){

    d=&(the_dialog[view]);
    clear(d->dp);
    draw_flat_wire_view(mesh,d->dp, view);
    show_mouse(NULL);
    blit(d->dp, screen, 0, 0, d->x, d->y, d->w, d->h);
    if(d->dp2){
      if(view==XZPROC)
        draw_sprite(screen,d->dp2, d->x+VIEWS_WIDTH/2-7, d->y+VIEWS_HEIGHT/2-9);
      else
        draw_sprite(screen,d->dp2, d->x+VIEWS_WIDTH/2-7, d->y+VIEWS_HEIGHT/2-11);
    }
    show_mouse(screen);
 }

}
void redraw_3dview(){
   DIALOG * d=&(the_dialog[THREEDPROC]);
   clear_to_color(d->dp,0);
   translate_world_to_screen(mesh);
   draw_3d(mesh,d->dp);
   show_mouse(NULL);
   blit(d->dp, screen, 0, 0, VIEWS_WIDTH+2, VIEWS_HEIGHT+2, VIEWS_WIDTH, VIEWS_HEIGHT);
   show_mouse(screen);
}

int repulse_mouse(int view, int mx,int my);
int view_proc(int msg, DIALOG *d, int c UNUSED){

   int unifsc_p;
   int view=d->d1;
   int mx,my,mx2,my2,vmx,vmy,xmickeys,ymickeys;
   int stat_direction=0;
   static int lastmx=0,lastmy=0;
   int rectstartx,rectstarty;
   int mm,i,j;

   VECTOR dr;


   double a,xscale,yscale;

   if (msg==MSG_DRAW){


       if(view==XYVIEW||view==ZYVIEW||view==XZVIEW){
           clear(d->dp);
           draw_flat_wire_view(mesh,d->dp, view);
           blit(d->dp, screen, 0, 0, d->x, d->y, d->w, d->h);
           if(d->dp2){
              if(view==XZVIEW)
                 draw_sprite(screen,d->dp2, d->x+VIEWS_WIDTH/2-7, d->y+VIEWS_HEIGHT/2-9);
              else
                 draw_sprite(screen,d->dp2, d->x+VIEWS_WIDTH/2-7, d->y+VIEWS_HEIGHT/2-11);
           }
       }       
       else if(view==VIEW_3D){
          clear_to_color(d->dp,0);
          translate_world_to_screen(mesh);
          draw_3d(mesh,d->dp);
          blit(d->dp, screen, 0, 0, VIEWS_WIDTH+2, VIEWS_HEIGHT+2, VIEWS_WIDTH, VIEWS_HEIGHT);
       }
       return D_O_K;
   }
   else if(msg==MSG_IDLE&&view!=THREEDPROC){
       mx=mouse_x;my=mouse_y;
       if( mx >= d->x && mx < d->x+d->w && my >= d->y && my < d->y+d->h){
           if(mx!=lastmx||my!=lastmy){
              lastmx=mx;lastmy=my;
              vmx = mx - (d->x+VIEWS_WIDTH/2);
              vmy = my - (d->y+VIEWS_HEIGHT/2);
              update_mouse_numbers(view,vmx,vmy);
           }

       }
       return D_O_K;
   }
   else if(msg==MSG_CLICK){
        get_mouse_mickeys(&xmickeys, &ymickeys);
        mx=mouse_x;my=mouse_y;
        
        if(view==XYVIEW||view==ZYVIEW||view==XZVIEW){
           if(mouse_b&2){
                  mm=do_menu(mousemodeflat_menu,mx,my);
                  if(mm>=0)
                       set_mouse_mode_flat(mm);
           }
           else{
              vmx = mx - (d->x+VIEWS_WIDTH/2);
              vmy = my - (d->y+VIEWS_HEIGHT/2);

              switch(mouse_mode_flat){
                case MOUSEMODEFLAT_CREATEVERTEX:
                  mouse_createvertex(view,vmx,vmy,(key[KEY_LCONTROL]||key[KEY_RCONTROL]));
                  update_numbers();
                  redraw_flatviews();
                  enabledisableitems();
                break;
                case MOUSEMODEFLAT_SELECTVERTEX:
                  if(!(key[KEY_LCONTROL]||key[KEY_RCONTROL]))
                          num_selected=0;
                  select_mouse_vertex_selection(view, vmx, vmy);
                  update_numbers();
                  redraw_flatviews();
                  enabledisableitems();
                break;
                case MOUSEMODEFLAT_UNSELECTVERTEX:
                  if(unselect_mouse_vertex_selection(view, vmx, vmy)){
                      redraw_flatviews();
                  }
                  update_numbers();
                  enabledisableitems();
                break;
                
                case MOUSEMODEFLAT_SELECTRECTANGLE:
                case MOUSEMODEFLAT_UNSELECTRECTANGLE:
                  rectstartx=mx-d->x;rectstarty=my-d->y;
                  mx -= d->x;
                  my -= d->y;
                  if(!(key[KEY_LCONTROL]||key[KEY_RCONTROL])){
                     if(mouse_mode_flat==MOUSEMODEFLAT_SELECTRECTANGLE)
                          num_selected=0;
                  }
                  show_mouse(NULL);
                  while(mouse_b&1){
                     mx2=mx;my2=my;
                     mx = mouse_x-d->x;
                     my = mouse_y-d->y;
                     if(mx2!=mx || my2!=my){
                       blit(d->dp, screen, 0, 0, d->x, d->y, d->w, d->h);

                       rect( screen,
                             MID(d->x,rectstartx+d->x,d->x+d->w),
                             MID(d->y,rectstarty+d->y,d->y+d->h),
                             MID(d->x,mx+d->x,d->x+d->w),
                             MID(d->y,my+d->y,d->y+d->h),
                             200);
                       update_mouse_numbers(view,mx-VIEWS_WIDTH/2,my-VIEWS_HEIGHT/2);
                     }
                  }
                  show_mouse(screen);
                  mx = mouse_x-d->x-VIEWS_WIDTH/2;
                  my = mouse_y-d->y-VIEWS_HEIGHT/2;
                  rectstartx -= VIEWS_WIDTH/2;
                  rectstarty -= VIEWS_HEIGHT/2;
                  if(mouse_mode_flat==MOUSEMODEFLAT_UNSELECTRECTANGLE){
                     select_in_rectangle(view,rectstartx,rectstarty,mx,my,0);
                  }
                  else{
                     select_in_rectangle(view,rectstartx,rectstarty,mx,my,1);
                  }
                  update_numbers();
                  redraw_flatviews();
                  enabledisableitems();
                break;
                case MOUSEMODEFLAT_SELECTCONNECTED:
                case MOUSEMODEFLAT_UNSELECTCONNECTED:
                  if((mouse_mode_flat==MOUSEMODEFLAT_SELECTCONNECTED)&&(!(key[KEY_LCONTROL]||key[KEY_RCONTROL])))
                        num_selected=0;
                  i=mouse_in_vertex_number(view,vmx,vmy);

                  if(i>=0){
                        if(mouse_mode_flat==MOUSEMODEFLAT_SELECTCONNECTED){
                          if(select_connected_component(i)>=0){
                             update_numbers();
                             redraw_flatviews();
                             enabledisableitems();
                          }
                        }
                        else if(mouse_mode_flat==MOUSEMODEFLAT_UNSELECTCONNECTED){
                          if(unselect_connected_component(i)>=0){
                             update_numbers();
                             redraw_flatviews();
                             enabledisableitems();
                          }
                        }
                  }

                break;
                
                case MOUSEMODEFLAT_PULLSELECTED:
                case MOUSEMODEFLAT_PULLALL:
                case MOUSEMODEFLAT_PULLEDGES:
                case MOUSEMODEFLAT_PULLEDGESSELECTED:
                case MOUSEMODEFLAT_MOUSEMOVEVERTEX:
                  moving_vertex=mouse_in_vertex_number(view,vmx,vmy);
                  if(moving_vertex==-1)
                        break;
                  if(mouse_mode_flat==MOUSEMODEFLAT_PULLEDGES||mouse_mode_flat==MOUSEMODEFLAT_PULLEDGESSELECTED){
                        if(get_connected_distances(moving_vertex)<0)
                           break;
                  }
                  while(mouse_b&1){
   
                        if(stat_direction==0&&(key[KEY_LSHIFT]||key[KEY_RSHIFT])){
                          mx2 = mouse_x-(d->x+VIEWS_WIDTH/2);
                          my2 = mouse_y-(d->y+VIEWS_HEIGHT/2);
                          if(abs(vmx-mx2)>abs(vmy-my2)){
                             stat_direction=1;
                          }
                          else if(abs(vmy-my2)>abs(vmx-mx2)){
                             stat_direction=-1;
                          }
                        }

                        mx2 = mouse_x-(d->x+VIEWS_WIDTH/2);
                        my2 = mouse_y-(d->y+VIEWS_HEIGHT/2);                        

                        if(stat_direction==1){
                             vmy=0;my2=0;
                        }
                        else if(stat_direction==-1){
                             vmx=0;mx2=0;
                        }
                        move_vertex(view,moving_vertex,mx2-vmx,my2-vmy);
                        if(mouse_mode_flat==MOUSEMODEFLAT_PULLSELECTED){
                           pull_selected(view,mesh->points[moving_vertex],(mx2-vmx),(my2-vmy));
                        }
                        else if(mouse_mode_flat==MOUSEMODEFLAT_PULLALL){
                           pull_all(view,mesh->points[moving_vertex],(mx2-vmx),(my2-vmy));
                        }
                        else if(mouse_mode_flat==MOUSEMODEFLAT_PULLEDGES){
                           pull_by_edges(view,mx2-vmx,my2-vmy);
                        }
                        else if(mouse_mode_flat==MOUSEMODEFLAT_PULLEDGESSELECTED){
                           pull_selected_by_edges(view,mx2-vmx,my2-vmy);
                        }
      
                        vmx=mx2;vmy=my2;
   
                        redraw_flatviews();
                        redraw_3dview();
                        update_mouse_numbers(view,mx2,my2);
                  }
                  if(mouse_b&4){

                        for(j=0;j<mesh->num_vertices;j++){
                           if(j!=moving_vertex){
                              if(mouse_in_this_vertex_p(view,j,vmx,vmy)){
                                 meld_vertex_pair(moving_vertex,j);
                                 break;
                              }
                           }
                        }
                        update_numbers();
                        redraw_flatviews();
                        redraw_3dview();

                  }

                  moving_vertex=-1;
                  if(mouse_mode_flat==MOUSEMODEFLAT_PULLEDGES||mouse_mode_flat==MOUSEMODEFLAT_PULLEDGESSELECTED){
                        free_connected_distances();
                  }

                break;
                case MOUSEMODEFLAT_REPULSE:
                  if(key[KEY_LSHIFT]||key[KEY_RSHIFT]){
                       switch(view){
                          case XYVIEW:
                              repulse_mouse_vtx.x=vmx;
                              repulse_mouse_vtx.y=vmy;
                              
                          break;
                          case ZYVIEW:
                              repulse_mouse_vtx.z=vmx;
                              repulse_mouse_vtx.y=vmy;

                          break;
                          case XZVIEW:
                              repulse_mouse_vtx.x=vmx;
                              repulse_mouse_vtx.z=vmy;
                          
                          break;


                       }
                  }
                  else{
                    while(mouse_b&1){
   
                        mx2 = mouse_x-(d->x+VIEWS_WIDTH/2);
                        my2 = mouse_y-(d->y+VIEWS_HEIGHT/2);


                        repulse_mouse(view,mx2,my2);

                        vmx=mx2;vmy=my2;
   
                        redraw_flatviews();
                        redraw_3dview();
                        update_mouse_numbers(view,mx2,my2);
                    }
                  }
                break;

                case MOUSEMODEFLAT_MOVESELECTED:
                case MOUSEMODEFLAT_MOVEALL:
                  while(mouse_b&1){
                        if(stat_direction==0&&(key[KEY_LSHIFT]||key[KEY_RSHIFT])){
                          mx2 = mouse_x-(d->x+VIEWS_WIDTH/2);
                          my2 = mouse_y-(d->y+VIEWS_HEIGHT/2);
                          if(abs(vmx-mx2)>abs(vmy-my2)){
                             stat_direction=1;
                          }
                          else if(abs(vmy-my2)>abs(vmx-mx2)){
                             stat_direction=-1;
                          }
                        }

                        mx2=mx;my2=my;
                        mx=mouse_x;my=mouse_y;
                        if(stat_direction==1){
                             my2=my;
                        }
                        else if(stat_direction==-1){
                             mx2=mx;
                        }

                        if(mouse_mode_flat==MOUSEMODEFLAT_MOVESELECTED){
                            move_selected(view,mx-mx2,my-my2);
                        }
                        else{
                            move_all(view,mx-mx2,my-my2);
                        }
                        update_mouse_numbers(view,mx-(d->x+VIEWS_WIDTH/2),my-(d->y+VIEWS_HEIGHT/2));
                        redraw_flatviews();
                        redraw_3dview();
                  }

                break;
                case MOUSEMODEFLAT_ROTATESELECTED:
                case MOUSEMODEFLAT_ROTATEALL:
                  mx -= d->x; my -= d->y;
                  while(mouse_b&1){
                         get_mouse_mickeys(&xmickeys, &ymickeys);
                         a=atan2((double)(my-VIEWS_HEIGHT/2+ymickeys),(double)(mx-VIEWS_WIDTH/2+xmickeys))-atan2((double)(my-VIEWS_HEIGHT/2),(double)(mx-VIEWS_WIDTH/2));
                         if(isnan(a)) break;
                         mx=mouse_x;my=mouse_y;
                         mx -= d->x; my -= d->y;
                         if(mouse_mode_flat==MOUSEMODEFLAT_ROTATESELECTED){
                            if(options_menu[ROTATEABOUTITEM].flags&D_SELECTED)
                               rotate_selected(view ,(float)a);
                            else
                               rotate_selected_cg(view ,(float)a);
                         }
                         
                         else{
                            if(options_menu[ROTATEABOUTITEM].flags&D_SELECTED)
                               rotate_all(view ,(float)a);
                            else
                               rotate_all_cg(view ,(float)a);
                         }
                         redraw_flatviews();
                         redraw_3dview();
                  }
                break;
                case MOUSEMODEFLAT_SCALESELECTED:
                case MOUSEMODEFLAT_SCALEALL:

                  unifsc_p=options_menu[UNIFORMSCALINGITEM].flags&D_SELECTED;
                  while(mouse_b&1){
                         if( (!unifsc_p)&&stat_direction==0&&(key[KEY_LSHIFT]||key[KEY_RSHIFT])){
                           mx2 = mouse_x-(d->x+VIEWS_WIDTH/2);
                           my2 = mouse_y-(d->y+VIEWS_HEIGHT/2);
                           if(abs(vmx-mx2)>abs(vmy-my2)){
                              stat_direction=1;
                           }
                           else if(abs(vmy-my2)>abs(vmx-mx2)){
                              stat_direction=-1;
                           }
                         }

                          
                          get_mouse_mickeys(&xmickeys, &ymickeys);
                          if(xmickeys || ymickeys){
                             if(stat_direction==-1)
                                xscale=1;
                             else
                                xscale=pow2((double)xmickeys/(double)VIEWS_WIDTH);
                             if(stat_direction==1)
                                yscale=1;
                             else
                                yscale=pow2((double)-ymickeys/(double)VIEWS_HEIGHT);
                             if(mouse_mode_flat==MOUSEMODEFLAT_SCALESELECTED)
                                  scale_selected(view,(float)xscale,(float)yscale,unifsc_p);
                             else scale_all(view,(float)xscale,(float)yscale, unifsc_p);

                             redraw_flatviews();
                             redraw_3dview();

                          }
                  }
                break;
                case MOUSEMODEFLAT_INSERTVERTEX:
                  mouse_insertvertex(view,vmx,vmy);
                  update_numbers();
                  redraw_flatviews();
                  enabledisableitems();
                break;
                case MOUSEMODEFLAT_DELETEVERTEX:
                  mouse_delete(view,vmx,vmy);
                  update_numbers();
                  redraw_flatviews();
                  redraw_3dview();
                  enabledisableitems();
                break;
              }
              
           }
           
        }
        if(view==VIEW_3D){
           if(mouse_b&2){
             mm=do_menu(mousemode3d_menu,mx,my);
             if(mm>=0)
                set_mouse_mode_3d(mm);
           }
           else{
              switch(mouse_mode_3d){
                case MOUSEMODE3d_ROTATE0:
                case MOUSEMODE3d_ROTATE1:
                   while(mouse_b&1){
                         get_mouse_mickeys(&xmickeys, &ymickeys);
                         if(mouse_mode_3d==MOUSEMODE3d_ROTATE0){
                           dr.x=-(float)ymickeys;dr.y=(float)xmickeys;dr.z=0;
                         }
                         else if(mouse_mode_3d==MOUSEMODE3d_ROTATE1){
                           dr.x=-(float)ymickeys;dr.y=0;dr.z=(float)xmickeys;
                         }
                         ani_rotate(mesh,dr);
                         //dialog_message(d, MSG_DRAW, 0, &obj);
                         redraw_3dview();
    
                   }
                break;
                case MOUSEMODE3d_TRANSLATEXY:
                   while(mouse_b&1){
                         get_mouse_mickeys(&xmickeys, &ymickeys);
    
                         dr.x=(float)xmickeys;dr.y=(float)ymickeys;dr.z=0;
                         ani_translate(mesh,dr);
                         //dialog_message(d, MSG_DRAW, 0, &obj);
                         redraw_3dview();
                   }
                   break;
                case MOUSEMODE3d_TRANSLATEXZ:
                   while(mouse_b&1){
                         get_mouse_mickeys(&xmickeys, &ymickeys);
    
                         dr.x=(float)xmickeys;dr.y=0;dr.z=(float)ymickeys;
                         ani_translate(mesh,dr);
                         //dialog_message(d, MSG_DRAW, 0, &obj);
                         redraw_3dview();
                   }
                break;
                case MOUSEMODE3d_SCALE:
                   while(mouse_b&1){
                         get_mouse_mickeys(&xmickeys, &ymickeys);
                         if(ymickeys){

                         yscale=pow2(((double)-ymickeys/(double)VIEWS_HEIGHT));

                         ani_scale(mesh,yscale);
                         //dialog_message(d, MSG_DRAW, 0, &obj);
                         redraw_3dview();
                         }
                   }
                break;
              }
           }
        }


        return D_O_K;
   }

   return D_O_K;

}


int layoutscreen_proc(int msg, DIALOG *d, int c){
   int cret=d_clear_proc(msg,d,c);

   if(msg==MSG_DRAW){
         line(screen,VIEWS_WIDTH+1,0,VIEWS_WIDTH+1,2*(VIEWS_HEIGHT),d->bg);
         line(screen,0,VIEWS_HEIGHT+1,2*(VIEWS_WIDTH+1),VIEWS_HEIGHT+1,d->bg);
         update_numbers();
         
         return D_O_K;
   }
   return cret;
}
int redraw_all_button_proc(int msg, DIALOG *d UNUSED, int c){
   int bret=d_button_proc(msg,d,c);
   if(bret==D_CLOSE){

        return D_REDRAW;
   }
   return bret;
}
int poly_button_proc(int msg, DIALOG *d UNUSED, int c){
   int bret=d_button_proc(msg,d,c);
   if(bret==D_CLOSE){
       if(do_poly_dialog()){
         update_numbers();
         redraw_flatviews();
         redraw_3dview();
         enabledisableitems();
         set_mouse_mode_flat(MOUSEMODEFLAT_MOVESELECTED);
       }
       return D_O_K;
   }
   return bret;
}


int deform_icon_proc(int msg, DIALOG *d UNUSED, int c){
   int bret;
   int obj;
   bret=d_icon_proc(msg,d,c);
   if(bret==D_CLOSE){
       do_deform_menu();
       dialog_message(&the_dialog[THREEDPROC], MSG_DRAW, 0, &obj);
       return D_O_K;
   }
   return bret;
}


int do_make_triangle(){
    int obj;
    if(num_selected !=3){
          return D_O_K;
    }
    if(find_triangle_in_mesh(mesh,curVertex[0],curVertex[1],curVertex[2])>=0){
          alert("The selected vertices are", "already making a triangle", NULL, "OK",NULL, 13, 0);
          return D_O_K;
    }

    //new_triangle(mesh,curVertex[0],curVertex[1],curVertex[2],0);
    new_triangle_o(mesh,curVertex[0],curVertex[1],curVertex[2]);
    update_numbers();
    redraw_flatviews();
    dialog_message(&the_dialog[THREEDPROC], MSG_DRAW, 0, &obj);
    //redraw_3dview();
    enabledisableitems();
    return D_O_K;
}

void delete_selected(){
    int s=0,i;

    while( s<num_selected){
          delete_vertex(mesh,curVertex[s]);

          for(i=0;i<num_selected;i++){
             if(curVertex[i]>curVertex[s])
                 curVertex[i]--;
          }
          s++;
    }

    num_selected=0;

}
int do_delete_selected(){
    int obj;

    waiting_call();
    delete_selected();
    waiting_done();

    update_numbers();
    redraw_flatviews();
    dialog_message(&the_dialog[THREEDPROC], MSG_DRAW, 0, &obj);
    enabledisableitems();
    return D_O_K;
}

int do_delete_all(){
   int obj;
   num_selected=0;
   mesh->num_vertices=0;
   mesh->num_faces=0;
   update_numbers();
   redraw_flatviews();
   enabledisableitems();
   init_shape(mesh,Initial_v,Initial_r,Initial_dz,Initial_dr,Initial_scale);
   dialog_message(&the_dialog[THREEDPROC], MSG_DRAW, 0, &obj);
   return D_O_K;
}

int do_mark_selected(){
     mark_selected(1,'|');
     redraw_flatviews();     
     return D_O_K;
}
int do_unmark_selected(){
     mark_selected(~1,'&');
     redraw_flatviews();     
     return D_O_K;
}

int do_meld_selected(){
    int obj;
    VTX cg;
    selected_CG(&cg);
    if(num_selected<2)
        return D_O_K;
    waiting_call();
    while(num_selected>1){
          meld_vertex_pair(curVertex[0],curVertex[1]);
    }
    waiting_done();
    
    mesh->points[curVertex[0]].x=cg.x;
    mesh->points[curVertex[0]].y=cg.y;
    mesh->points[curVertex[0]].z=cg.z;
    
    update_numbers();
    redraw_flatviews();
    dialog_message(&the_dialog[THREEDPROC], MSG_DRAW, 0, &obj);
    enabledisableitems();
    return D_O_K;
}

int do_delete_selected_triangles(){
   int obj;
   int f,s1,s2,s3;
   TRIANGLE *T;
   f=0;
   waiting_call();
   while(f<mesh->num_faces){

        T=&(mesh->faces[f]);
        if( triangle_in_selected(*T,&s1,&s2,&s3)){
           delete_triangle(mesh,f);
        }
        else f++;
   }
   waiting_done();
   update_numbers();
   redraw_flatviews();
   dialog_message(&the_dialog[THREEDPROC], MSG_DRAW, 0, &obj);
   enabledisableitems();
   return D_O_K;

}
int do_delete_triangle(){
    int obj;
    int f;
    if(num_selected != 3){
          return D_O_K;
    }
    f=find_triangle_in_mesh(mesh,curVertex[0],curVertex[1],curVertex[2]);
    if(f<0){
          alert("The selected vertices", "are not a triangle", NULL, "OK",NULL, 13, 0);
          return D_O_K;
    }
    delete_triangle(mesh,f);
    update_numbers();
    redraw_flatviews();
    dialog_message(&the_dialog[THREEDPROC], MSG_DRAW, 0, &obj);
    enabledisableitems();
    return D_O_K;
}

int save_undo();
MESH * deform_connected_selection_copy();


int divide_selected_p=0;
int do_divide_dialog();

int do_divide(){
    int obj,divret;
    int n_f=mesh->num_faces;
    int n_s=num_selected;
    int n_v,vi;
    MESH * divmesh;
    if(do_divide_dialog()){
    
        waiting_call();
    
        save_undo();
        if(divide_selected_p){
            divmesh=deform_connected_selection_copy();
            n_f=divmesh->num_faces;
        }
        else{
            divmesh=mesh;
        }
        if(4*n_f >= divmesh->num_faces_allocated-1){
           waiting_done();
           alert("Too many triangles to divide!",NULL,NULL,"OK",NULL,13,0);
           return D_O_K;
        }
    
        if(divmesh==NULL){
              waiting_done();
              return D_O_K;
        }
        divret=divide_triangle_mesh(divmesh);
        if(divret<0){
            waiting_done();
            switch(divret){
               case -1:
               alert("Too many triangles",NULL,NULL,"OK",NULL,13,0);
               break;
               case -2:
               alert("Too many vertices",NULL,NULL,"OK",NULL,13,0);
               break;
               case -3:
               alert("Too many edges",NULL,NULL,"OK",NULL,13,0);
               break;
               case -4:
               alert("Memory allocation error",NULL,NULL,"OK",NULL,13,0);
               break;
    
               default:
               alert("Error in mesh division","???",NULL,"OK",NULL,13,0);
               break;
            }
            do_get_undo();
            return D_O_K;
        }
        if(divide_selected_p){
             delete_selected();
             n_v=mesh->num_vertices;
           
             add_mesh_to_mesh(mesh,divmesh);
             for(vi=0;vi<n_s;vi++){
                add_to_selected(n_v+vi, 0);
             }
             delete_mesh(divmesh);
        }
    
        waiting_done();
        update_numbers();
        redraw_flatviews();
        dialog_message(&the_dialog[THREEDPROC], MSG_DRAW, 0, &obj);
        enabledisableitems();
    }
    return D_O_K;
}

int do_try_orient(){
    int obj;
    waiting_call();
    try_orient(mesh);
    waiting_done();
    dialog_message(&the_dialog[THREEDPROC], MSG_DRAW, 0, &obj);
    return D_O_K;
}

int do_reverse_selected(){
     int obj;
     reverse_selected();
     redraw_flatviews();
     dialog_message(&the_dialog[THREEDPROC], MSG_DRAW, 0, &obj);
     return D_O_K;
}

int do_cleanup_loose_vertices(){
     int obj;
     int vi,s;

     vi=0;
     while (vi<mesh->num_vertices){
        if(!vertex_used_p(mesh,vi)){
           delete_vertex(mesh,vi);
           for(s=0;s<num_selected;s++){
               if(curVertex[s]==vi)
                    unselect_vertex(s);
           }
           for(s=0;s<num_selected;s++){
               if(curVertex[s]>vi)
                    curVertex[s]--;
           }
        }
        else vi++;
        
        
     }
     update_numbers();
     redraw_flatviews();
     dialog_message(&the_dialog[THREEDPROC], MSG_DRAW, 0, &obj);
     enabledisableitems();
     return D_O_K;
}

int save_undo(){
      if(mesh->num_faces>0){
         save_TRI_triangle_mesh(mesh,"lasttri.tri",3);
         return 1;
      }
      return 0;
}

/* various 'get parameter' dialogs */
//#define DBG 200
DIALOG divide_dialog[]={
/* (dialog proc)     (x)   (y)   (w)   (h)     (fg)  (bg)  (key)    (flags)           (d1)  (d2)  (dp) */
  { d_box_proc,        0,   0,   240, 90,      0,   DBG,    0,         0,              0,    0,   NULL,NULL,NULL},
  

  { d_check_proc,     10,  30,    200,  12,      0,   DBG,   's',         0,              0,    0,    "apply only to Selected" ,NULL,NULL},

  { d_button_proc,    10,  60,    20,  20,      0,   255,   13,    D_EXIT,              0,    0,   "OK",NULL,NULL},
  { d_button_proc,    40,  60,    60,  20,      0,   255,    0,    D_EXIT,              0,    0,   "Cancel",NULL,NULL},
  
  { NULL,              0,    0,    0,   0,      0,     0,    0,         0,              0,    0,    NULL ,NULL,NULL}
};
#define DIVIDESELECTEDCHECK 1
#define DIVIDECANCEL 3

int do_divide_dialog(){
    int popupret;
    centre_dialog(divide_dialog);
    popupret=popup_dialog(divide_dialog,1);
    if(popupret==DIVIDECANCEL){
       return 0;
    }
    divide_selected_p=divide_dialog[DIVIDESELECTEDCHECK].flags&D_SELECTED;
    return 1;
}


int snapvalslidercallback(void *dp3 UNUSED, int d2);
char snapvalstring[80];
#define SNAPVALSLIDERLENGTH 99

DIALOG snap_dialog[]={
/* (dialog proc)     (x)   (y)   (w)   (h)     (fg)  (bg)  (key)    (flags)           (d1)  (d2)  (dp) */
  { d_box_proc,        0,   0,   180, 150,      0,   DBG,    0,         0,              0,    0,   NULL,NULL,NULL},
  

  { d_text_proc,      10,  22,     0,   0,      0,   DBG,    0,         0,              0,    0,    "Snap to nearest:" ,NULL,NULL},
  { d_slider_proc,    10,  32,   120,  10,      0,   DBG,    0,         0, SNAPVALSLIDERLENGTH,    0,    NULL,     snapvalslidercallback,NULL},
  { d_text_proc,      70,  44,    60,  10,      0,   DBG,    0,         0,              0,    0,    snapvalstring ,NULL,NULL},

  { d_check_proc,     10,  64,    30,  10,      0,   DBG,    'x',D_SELECTED,              0,    0,    "X" ,NULL,NULL},
  { d_check_proc,     50,  64,    30,  10,      0,   DBG,    'y',D_SELECTED,              0,    0,    "Y" ,NULL,NULL},
  { d_check_proc,     90,  64,    30,  10,      0,   DBG,    'z',D_SELECTED,              0,    0,    "Z" ,NULL,NULL},

  { d_check_proc,     10, 90,     170,  12,      0,   DBG,   's',         0,              0,    0,    "snap only Selected" ,NULL,NULL},
  { d_button_proc,    10,  120,    20,  20,      0,   255,   13,    D_EXIT,              0,    0,   "OK",NULL,NULL},
  { d_button_proc,    40,  120,    60,  20,      0,   255,    0,    D_EXIT,              0,    0,   "Cancel",NULL,NULL},
  
  { NULL,              0,    0,    0,   0,      0,     0,    0,         0,              0,    0,    NULL ,NULL,NULL}
};
#define SNAPVALSLIDER 2
#define SNAPVALTEXT 3
#define SNAPXCHECK 4
#define SNAPYCHECK 5
#define SNAPZCHECK 6
#define SNAPSELECTEDCHECK 7

#define SNAPVALCANCEL 9

int snapvalslidercallback(void *dp3 UNUSED, int d2)
{
        int obj;
        DIALOG *dtext=&snap_dialog[SNAPVALTEXT];
        DIALOG *dslider=&snap_dialog[SNAPVALSLIDER];

        snapval=50*(float)(d2)/(float)(dslider->d1);
        if (snapval==0)snapval=1.0;
        sprintf(dtext->dp,"%3.1f",snapval);
        rectfill(screen,dtext->x,dtext->y, dtext->x + dtext->w,dtext->y + dtext->h,dtext->bg);
        dialog_message(dtext, MSG_DRAW,0, &obj);
        return(D_O_K);
}

int snap_selected_p=0;
int do_snap_dialog(){
    int popupret;
    float holdsnapval=snapval;
    sprintf(snapvalstring,"%1.1f",snapval);
    
    snap_dialog[SNAPVALSLIDER].d2=(int)(snapval/50*SNAPVALSLIDERLENGTH);
    if(snapx_p)
      snap_dialog[SNAPXCHECK].flags|=D_SELECTED;
    else
      snap_dialog[SNAPXCHECK].flags&=~D_SELECTED;
    if(snapy_p)
      snap_dialog[SNAPYCHECK].flags|=D_SELECTED;
    else
      snap_dialog[SNAPYCHECK].flags&=~D_SELECTED;
    if(snapz_p)
      snap_dialog[SNAPZCHECK].flags|=D_SELECTED;
    else
      snap_dialog[SNAPZCHECK].flags&=~D_SELECTED;

    centre_dialog(snap_dialog);
    popupret=popup_dialog(snap_dialog,2);
    if(popupret==SNAPVALCANCEL){
       snapval=holdsnapval;
       return 0;
    }
    snap_selected_p=snap_dialog[SNAPSELECTEDCHECK].flags&D_SELECTED;
    snapx_p=snap_dialog[SNAPXCHECK].flags&D_SELECTED;
    snapy_p=snap_dialog[SNAPYCHECK].flags&D_SELECTED;
    snapz_p=snap_dialog[SNAPZCHECK].flags&D_SELECTED;
    
    return 1;
}

int do_snap(){
    int obj;
    if(do_snap_dialog()){
      if(snap_selected_p)
        snap_selected(snapval);
      else
        snap_all(snapval);
      redraw_flatviews();
      dialog_message(&the_dialog[THREEDPROC], MSG_DRAW, 0, &obj);
    }
    return D_O_K;
}


int repulsepoint_selected_p=0;
VTX repulse_point={0,0,0,0};
float repulse_point_gauss_exp=2.0;
float repulse_point_expmult=.001;
float repulse_point_c=1.0;


int repulse_point_selected(VTX RepulsePoint, float c,float a, float b){

    int si;
    VTX *v;

    for(si=0;si<num_selected;si++){
       v=&(mesh->points[curVertex[si]]);
       vtx_repulse_point(v,RepulsePoint,c, a, b);
    }
    return 1;
}


int do_repulse_point_dialog();

int do_repulse_point(){
    int obj;

    if(do_repulse_point_dialog()){
       save_undo();
       if(repulsepoint_selected_p)
          repulse_point_selected(repulse_point,repulse_point_c,repulse_point_expmult, repulse_point_gauss_exp);
       else
          repulse_point_mesh(mesh,repulse_point, repulse_point_c,repulse_point_expmult, repulse_point_gauss_exp);
       redraw_flatviews();
       dialog_message(&the_dialog[THREEDPROC], MSG_DRAW, 0, &obj);
    }
    return D_O_K;
}

int perturb_selected_p=0;
float perturbing=.1;
int do_perturb_dialog();
MESH * copy_selected_to_new_mesh();
int do_perturb(){
    int obj;
    MESH *pmesh;
    int vi;
    if(do_perturb_dialog()){
       save_undo();
       if(perturb_selected_p){

          pmesh=copy_selected_to_new_mesh();
          if(pmesh==NULL){
              return D_O_K;
          }
       }
       else{
          pmesh=mesh;
       }

       perturb_mesh(pmesh,perturbing,-1);

       if(perturb_selected_p){
           for(vi=0;vi<num_selected;vi++){
              mesh->points[curVertex[vi]].x=pmesh->points[vi].x;
              mesh->points[curVertex[vi]].y=pmesh->points[vi].y;
              mesh->points[curVertex[vi]].z=pmesh->points[vi].z;
           }
           delete_mesh(pmesh);
       }
       redraw_flatviews();
       dialog_message(&the_dialog[THREEDPROC], MSG_DRAW, 0, &obj);
    }
    return D_O_K;
}

int reduce_selected_p=0;
int do_reduce_dialog();
float reducepercent=.1;

int do_reduce(){
    int obj;
    int reduce_ret,alertret;
    int n_v;
    int reduceto;
    MESH *reducemesh=NULL;
    
    if(do_reduce_dialog()){
       save_undo();
       if(reduce_selected_p){
          reducemesh=deform_connected_selection_copy();
          n_v=num_selected;
       }
       else{
          reducemesh=mesh;
          n_v=mesh->num_vertices;
       }
       if(reducemesh==NULL){
          return D_O_K;
       }
       reduceto=n_v*(1-reducepercent);
       waiting_call();

       while(reducemesh->num_vertices>reduceto+1){
           reduce_ret=ProgressiveMesh(reducemesh);
           if(reduce_ret==-1){
               waiting_done();
               alert("An error occurred", "in reduction", NULL, "OK",NULL, 13, 0);
               do_get_undo();
               return D_O_K;
           }
           if(esc_pressed()){
               waiting_done();
               alertret=alert("Reduction aborted", "Keep reduction done so far,","or undo?" , "Keep","Undo", 13, 'u');
               if(alertret==2){
                  do_get_undo();
                  return D_O_K;
               }
               break;
           }
       }
       if(reduce_selected_p){
         delete_selected();       
         add_mesh_to_mesh(mesh,reducemesh);
         delete_mesh(reducemesh);
       }
       waiting_done();
       num_selected=0;
       update_numbers();
       redraw_flatviews();
       dialog_message(&the_dialog[THREEDPROC], MSG_DRAW, 0, &obj);
       enabledisableitems();
    }
    
    return D_O_K;

}

float wave_frequency=.01;
float wave_phase=0.0;
float wave_scale=.1;

int wave_selected(float Frequency, float Phase, float WaveScale){
    int si;
    VTX *v;
    
    for(si=0;si<num_selected;si++){
       v=&(mesh->points[curVertex[si]]);
       vtx_wave(v, Frequency, Phase, WaveScale);
    }
    return 1;
}

int do_wave_dialog();
int wave_selected_p=0;
int do_wave(){
    int obj;
    if(do_wave_dialog()){
      save_undo();
      if(wave_selected_p){
         wave_selected(wave_frequency,wave_phase,wave_scale);
      }
      else{
         wave_mesh(mesh,wave_frequency,wave_phase,wave_scale);
      }
      redraw_flatviews();
      dialog_message(&the_dialog[THREEDPROC], MSG_DRAW, 0, &obj);
    }
    return D_O_K;
}
int wavescaleslidercallback(void *dp3 UNUSED, int d2);
char wavescalestring[80];

char wavefrequencystring[12];
char wavephasestring[12];
#define WAVESCALESLIDERLENGTH 99

DIALOG wave_dialog[]={
/* (dialog proc)     (x)   (y)   (w)   (h)     (fg)  (bg)  (key)    (flags)           (d1)  (d2)  (dp) */
  { d_box_proc,        0,   0,   260, 220,      0,   DBG,    0,         0,              0,    0,   NULL,NULL,NULL},
  

  { d_text_proc,      10,  22,     0,   0,      0,   DBG,    0,         0,              0,    0,    "Amount:" ,NULL,NULL},
  { d_slider_proc,    10,  32,   120,  10,      0,   DBG,    0,         0, WAVESCALESLIDERLENGTH,    0,    NULL,     wavescaleslidercallback,NULL},
  { d_text_proc,      70,  44,    60,  10,      0,   DBG,    0,         0,              0,    0,    wavescalestring ,NULL,NULL},

  { d_text_proc,     80, 76,     0,   0,      0,   DBG,    0,         0,              0,    0,    "Freq" ,NULL,NULL},
  { d_edit_proc,     80, 90,    64,  12,      0,   255,     0,         0,               8,    0,   wavefrequencystring,NULL,NULL},
  
  { d_text_proc,     170, 76,     0,   0,      0,   DBG,    0,         0,              0,    0,    "Phase" ,NULL,NULL},
  { d_edit_proc,     170, 90,    50,  12,      0,   255,    0,         0,               8,    0,   wavephasestring,NULL,NULL},
  
  { d_check_proc,     80, 130,     150,  12,      0,   DBG,  's',         0,              0,    0,    "only Selected" ,NULL,NULL},

  { d_button_proc,    10,  190,    20,  20,      0,   255,   13,    D_EXIT,              0,    0,   "OK",NULL,NULL},
  { d_button_proc,    40,  190,    60,  20,      0,   255,    0,    D_EXIT,              0,    0,   "Cancel",NULL,NULL},
  
  { NULL,              0,    0,    0,   0,      0,     0,    0,         0,              0,    0,    NULL ,NULL,NULL}
};
#define WAVESCALESLIDER 2
#define WAVESCALETEXT 3
#define WAVESELECTEDCHECK 8
#define WAVECANCEL 10

int wavescaleslidercallback(void *dp3 UNUSED, int d2)
{
        int obj;
        DIALOG *dtext=&wave_dialog[WAVESCALETEXT];
        DIALOG *dslider=&wave_dialog[WAVESCALESLIDER];

        wave_scale=1*(float)(d2)/(float)(dslider->d1);
        sprintf(dtext->dp,"%1.3f",wave_scale);
        rectfill(screen,dtext->x,dtext->y, dtext->x + dtext->w,dtext->y + dtext->h,dtext->bg);
        dialog_message(dtext, MSG_DRAW,0, &obj);
        return(D_O_K);
}

int do_wave_dialog(){
    int popupret;
    float holdwavescale=wave_scale;
    sprintf(wavescalestring,"%1.3f",wave_scale);
    sprintf(wavefrequencystring,"%1.3f",wave_frequency);
    sprintf(wavephasestring,"%1.1f",wave_phase);
    
    wave_dialog[WAVESCALESLIDER].d2=(int)(wave_scale/1*WAVESCALESLIDERLENGTH);

    centre_dialog(wave_dialog);
    popupret=popup_dialog(wave_dialog,2);
    if(popupret==WAVECANCEL){
       wave_scale=holdwavescale;
       return 0;
    }
    wave_selected_p=wave_dialog[WAVESELECTEDCHECK].flags&D_SELECTED;
    wave_frequency=atof(wavefrequencystring);
    wave_phase=atof(wavephasestring);
    return 1;
}

int ripple_selected(VECTOR RippleAxis, float Frequency,float Phase, VECTOR RippleAmount){
    int si;
    VTX *v;

    for(si=0;si<num_selected;si++){
       v=&(mesh->points[curVertex[si]]);
       vtx_ripple(v, RippleAxis, Frequency,Phase, RippleAmount);
    }
    return 1;
}

VECTOR ripple_axis={0.0,1.0,0.0};
float ripple_frequency=1.0;
float ripple_phase=0.0;
float ripple_scale=1.0;
int ripple_selected_p=0;
int do_ripple_dialog();
int do_ripples(){
    int obj;
    VECTOR RA=ripple_axis;
    VECTOR RippleScale;

    if(do_ripple_dialog()){
       if(RA.x==0 && RA.y==0 && RA.z==0)
           return D_O_K; /* fail */
       normalize_vector_f(&RA.x,&RA.y,&RA.z);
       RippleScale.x=(1-RA.x)*ripple_scale;
       RippleScale.y=(1-RA.y)*ripple_scale;
       RippleScale.z=(1-RA.z)*ripple_scale;
       

       save_undo();
       if(ripple_selected_p)
          ripple_selected(ripple_axis, ripple_frequency,ripple_phase, RippleScale);
       else
          ripple_mesh(mesh, ripple_axis, ripple_frequency,ripple_phase, RippleScale);

       redraw_flatviews();
       dialog_message(&the_dialog[THREEDPROC], MSG_DRAW, 0, &obj);
    }
    return D_O_K;
}

int ripplescaleslidercallback(void *dp3 UNUSED, int d2);
char ripplescalestring[80];

char rippleaxisxstring[12];
char rippleaxisystring[12];
char rippleaxiszstring[12];
char ripplefrequencystring[12];
char ripplephasestring[12];
#define RIPPLESCALESLIDERLENGTH 99

DIALOG ripple_dialog[]={
/* (dialog proc)     (x)   (y)   (w)   (h)     (fg)  (bg)  (key)    (flags)           (d1)  (d2)  (dp) */
  { d_box_proc,        0,   0,   260, 220,      0,   DBG,    0,         0,              0,    0,   NULL,NULL,NULL},
  

  { d_text_proc,      10,  22,     0,   0,      0,   DBG,    0,         0,              0,    0,    "Amount:" ,NULL,NULL},
  { d_slider_proc,    10,  32,   120,  10,      0,   DBG,    0,         0, RIPPLESCALESLIDERLENGTH,    0,    NULL,     ripplescaleslidercallback,NULL},
  { d_text_proc,      70,  44,    60,  10,      0,   DBG,    0,         0,              0,    0,    ripplescalestring ,NULL,NULL},

  { d_text_proc,      10,  90,     0,   0,      0,   DBG,    0,         0,              0,    0,    "X:" ,NULL,NULL},
  { d_text_proc,      10, 106,     0,   0,      0,   DBG,    0,         0,              0,    0,    "Y:" ,NULL,NULL},
  { d_text_proc,      10, 122,     0,   0,      0,   DBG,    0,         0,              0,    0,    "Z:" ,NULL,NULL},
  
  { d_text_proc,     30,  76,     0,   0,      0,   DBG,    0,         0,              0,    0,    "Axis" ,NULL,NULL},
  { d_edit_proc,     30,  90,    50,  12,      0,   255,     0,         0,               8,    0,   rippleaxisxstring,NULL,NULL},
  { d_edit_proc,     30, 106,    50,  12,      0,   255,     0,         0,               8,    0,   rippleaxisystring,NULL,NULL},
  { d_edit_proc,     30, 122,    50,  12,      0,   255,     0,         0,               8,    0,   rippleaxiszstring,NULL,NULL},

  { d_text_proc,     100, 76,     0,   0,      0,   DBG,    0,         0,              0,    0,    "Freq" ,NULL,NULL},
  { d_edit_proc,     100, 90,    50,  12,      0,   255,     0,         0,               8,    0,   ripplefrequencystring,NULL,NULL},
  
  { d_text_proc,     170, 76,     0,   0,      0,   DBG,    0,         0,              0,    0,    "Phase" ,NULL,NULL},
  { d_edit_proc,     170, 90,    50,  12,      0,   255,    0,         0,               8,    0,   ripplephasestring,NULL,NULL},

  { d_check_proc,     10, 150,     150, 12,      0,   DBG,    's',         0,              0,    0,    "only Selected" ,NULL,NULL},

  { d_button_proc,    10,  190,    20,  20,      0,   255,   13,    D_EXIT,              0,    0,   "OK",NULL,NULL},
  { d_button_proc,    40,  190,    60,  20,      0,   255,    0,    D_EXIT,              0,    0,   "Cancel",NULL,NULL},
  
  { NULL,              0,    0,    0,   0,      0,     0,    0,         0,              0,    0,    NULL ,NULL,NULL}
};
#define RIPPLESCALESLIDER 2
#define RIPPLESCALETEXT 3
#define RIPPLESELECTEDCHECK 15
#define RIPPLECANCEL 17

int ripplescaleslidercallback(void *dp3 UNUSED, int d2)
{
        int obj;
        DIALOG *dtext=&ripple_dialog[RIPPLESCALETEXT];
        ripple_scale=10*(float)(d2)/(float)RIPPLESCALESLIDERLENGTH;
        sprintf(ripplescalestring,"%1.3f",ripple_scale);
        rectfill(screen,dtext->x,dtext->y, dtext->x + dtext->w,dtext->y + dtext->h,dtext->bg);
        dialog_message(dtext, MSG_DRAW,0, &obj);
        return(D_O_K);
}
int do_ripple_dialog(){
    int popupret;
    float holdripplescale=ripple_scale;
    sprintf(ripplescalestring,"%1.3f",ripple_scale);
    sprintf(rippleaxisxstring,"%1.1f",ripple_axis.x);
    sprintf(rippleaxisystring,"%1.1f",-ripple_axis.y);
    sprintf(rippleaxiszstring,"%1.1f",ripple_axis.z);
    sprintf(ripplefrequencystring,"%1.1f",ripple_frequency);
    sprintf(ripplephasestring,"%1.1f",ripple_phase);
    
    ripple_dialog[RIPPLESCALESLIDER].d2=(int)(ripple_scale/10*RIPPLESCALESLIDERLENGTH);

    centre_dialog(ripple_dialog);
    popupret=popup_dialog(ripple_dialog,2);
    if(popupret==RIPPLECANCEL){
       ripple_scale=holdripplescale;

       return 0;
    }
    ripple_selected_p=ripple_dialog[RIPPLESELECTEDCHECK].flags&D_SELECTED;
    
    ripple_axis.x=atof(rippleaxisxstring);
    ripple_axis.y=-atof(rippleaxisystring);
    ripple_axis.z=atof(rippleaxiszstring);
    ripple_frequency=atof(ripplefrequencystring);
    ripple_phase=atof(ripplephasestring);
    return 1;
}

int twist_selected_about(float TwistAngle, VTX TwistCentre,VECTOR TwistAxis){
    int si;
    VTX *v;
    for(si=0;si<num_selected;si++){
       v=&(mesh->points[curVertex[si]]);
       vtx_twist_about(v,TwistAngle,TwistCentre,TwistAxis);
    }
    return 1;
}


int twist_selected(float TwistAngle, VTX TwistCentre){
    int si;
    VTX *v;
    for(si=0;si<num_selected;si++){
       v=&(mesh->points[curVertex[si]]);
       vtx_twist(v,TwistAngle,TwistCentre);
    }
    return 1;
}

VECTOR twist_axis={0,1,0};
float twist_angle=1;
VTX twist_center={0,0,0,0};
int twist_about_axis_p=0;
int twist_selected_p=0;

int do_twist_dialog();
int do_twist(){
    int obj;
    if(do_twist_dialog()&&twist_angle!=0){
      save_undo();
      if(twist_selected_p){
         if(twist_about_axis_p&&(twist_axis.x!=0||twist_axis.y!=0||twist_axis.z!=0)){
           twist_selected_about(twist_angle, twist_center,twist_axis);
         }
         else{
           twist_selected(twist_angle, twist_center);
         }
      }
      else{
         if(twist_about_axis_p&&(twist_axis.x!=0||twist_axis.y!=0||twist_axis.z!=0)){
           twist_mesh_about(mesh,twist_angle, twist_center,twist_axis);
         }
         else{
           twist_mesh(mesh,twist_angle, twist_center);
         }
      }
      redraw_flatviews();
      dialog_message(&the_dialog[THREEDPROC], MSG_DRAW, 0, &obj);
    }
    return D_O_K;
}

int twistangleslidercallback(void *dp3 UNUSED, int d2);
char twistanglestring[80];
char twistcenterxstring[12];
char twistcenterystring[12];
char twistcenterzstring[12];

char twistaxisxstring[12];
char twistaxisystring[12];
char twistaxiszstring[12];

#define TWISTANGLESLIDERLENGTH 99

DIALOG twist_dialog[]={
/* (dialog proc)     (x)   (y)   (w)   (h)     (fg)  (bg)  (key)    (flags)           (d1)  (d2)  (dp) */
  { d_box_proc,        0,   0,   260, 220,      0,   DBG,    0,         0,              0,    0,   NULL,NULL,NULL},


  { d_text_proc,      10,  22,     0,   0,      0,   DBG,    0,         0,              0,    0,    "Angle:" ,NULL,NULL},
  { d_slider_proc,    10,  32,   120,  10,      0,   DBG,    0,         0, TWISTANGLESLIDERLENGTH,    0,    NULL,     twistangleslidercallback,NULL},
  { d_text_proc,      70,  44,    60,  10,      0,   DBG,    0,         0,              0,    0,    twistanglestring ,NULL,NULL},

  { d_text_proc,      10,  90,     0,   0,      0,   DBG,    0,         0,              0,    0,    "X:" ,NULL,NULL},
  { d_text_proc,      10, 106,     0,   0,      0,   DBG,    0,         0,              0,    0,    "Y:" ,NULL,NULL},
  { d_text_proc,      10, 122,     0,   0,      0,   DBG,    0,         0,              0,    0,    "Z:" ,NULL,NULL},
  
  { d_text_proc,     30,  76,     0,   0,      0,   DBG,    0,         0,              0,    0,    "Center" ,NULL,NULL},
  { d_edit_proc,     30,  90,    50,  12,      0,   255,     0,         0,               8,    0,   twistcenterxstring,NULL,NULL},
  { d_edit_proc,     30, 106,    50,  12,      0,   255,     0,         0,               8,    0,   twistcenterystring,NULL,NULL},
  { d_edit_proc,     30, 122,    50,  12,      0,   255,     0,         0,               8,    0,   twistcenterzstring,NULL,NULL},

/***/
  { d_check_proc,    120,  60,   150,  12,      0,   DBG,   'a',         0,              0,    0,    "about Axis" ,NULL,NULL},
  { d_text_proc,     150,  76,     0,   0,      0,   DBG,    0,         0,              0,    0,    "Axis" ,NULL,NULL},
  { d_edit_proc,     150,  90,    50,  12,      0,   255,     0,         0,               8,    0,   twistaxisxstring,NULL,NULL},
  { d_edit_proc,     150, 106,    50,  12,      0,   255,     0,         0,               8,    0,   twistaxisystring,NULL,NULL},
  { d_edit_proc,     150, 122,    50,  12,      0,   255,     0,         0,               8,    0,   twistaxiszstring,NULL,NULL},

  { d_check_proc,     10,  150,   150,  12,      0,   DBG,   's',         0,              0,    0,    "only Selected" ,NULL,NULL},
  { d_button_proc,    10,  190,    20,  20,      0,   255,   13,    D_EXIT,              0,    0,   "OK",NULL,NULL},
  { d_button_proc,    40,  190,    60,  20,      0,   255,    0,    D_EXIT,              0,    0,   "Cancel",NULL,NULL},
  
  { NULL,              0,    0,    0,   0,      0,     0,    0,         0,              0,    0,    NULL ,NULL,NULL}
};
#define TWISTANGLESLIDER 2
#define TWISTANGLETEXT 3
#define TWISTABOUTAXISCHECK 11
#define TWISTSELECTEDCHECK 16
#define TWISTCANCEL 18

int twistangleslidercallback(void *dp3 UNUSED, int d2)
{
        int obj;
        DIALOG *dtext=&twist_dialog[TWISTANGLETEXT];
        twist_angle=10*(float)(d2)/(float)TWISTANGLESLIDERLENGTH-5;
        sprintf(twistanglestring,"%1.3f",twist_angle);
        rectfill(screen,dtext->x,dtext->y, dtext->x + dtext->w,dtext->y + dtext->h,dtext->bg);
        dialog_message(dtext, MSG_DRAW,0, &obj);
        return(D_O_K);
}
int do_twist_dialog(){
    int popupret;
    float holdtwistangle=twist_angle;
    sprintf(twistanglestring,"%1.3f",twist_angle);
    sprintf(twistcenterxstring,"%1.1f",twist_center.x);
    sprintf(twistcenterystring,"%1.1f",-twist_center.y);
    sprintf(twistcenterzstring,"%1.1f",twist_center.z);

    sprintf(twistaxisxstring,"%1.1f",twist_axis.x);
    sprintf(twistaxisystring,"%1.1f",twist_axis.y);
    sprintf(twistaxiszstring,"%1.1f",twist_axis.z);
    
    twist_dialog[TWISTANGLESLIDER].d2=(int)((twist_angle+5)/(10)*TWISTANGLESLIDERLENGTH);

    centre_dialog(twist_dialog);
    popupret=popup_dialog(twist_dialog,2);
    if(popupret==TWISTCANCEL){
       twist_angle=holdtwistangle;

       return 0;
    }
    twist_selected_p=twist_dialog[TWISTSELECTEDCHECK].flags&D_SELECTED;
    
    twist_center.x=atof(twistcenterxstring);
    twist_center.y=-atof(twistcenterystring);
    twist_center.z=atof(twistcenterzstring);

    twist_about_axis_p=twist_dialog[TWISTABOUTAXISCHECK].flags&D_SELECTED;
    twist_axis.x=atof(twistaxisxstring);
    twist_axis.y=atof(twistaxisystring);
    twist_axis.z=atof(twistaxiszstring);

    return 1;
}




int bend_selected(VTX BendPoint,VECTOR BX,float BA){
    int si;
    VTX *v;

    for(si=0;si<num_selected;si++){
       v=&(mesh->points[curVertex[si]]);
       vtx_bend(v,BendPoint,BX,BA);
    }
    return 1;
}

VTX bend_point1={0,-10,0,0};
VTX bend_point2={0,10,0,0};
VECTOR bend_direction={1,0,0};
float bend_angle=1;
int bend_selected_p=0;
int do_bend_dialog();
int do_bend(){
    int obj;
    float  BA;
    VTX BPd;
    VECTOR BX;
    if(do_bend_dialog()){
      BPd.x=bend_point2.x-bend_point1.x;
      BPd.y=bend_point2.y-bend_point1.y;
      BPd.z=bend_point2.z-bend_point1.z;
      if(  (bend_direction.x==0 && bend_direction.y==0 && bend_direction.z==0)
           ||(BPd.x==0 && BPd.y==0 && BPd.z==0)
         ){
            return D_O_K; /* fail */
      }


      BA = bend_angle/vector_length_f(BPd.x,BPd.y,BPd.z);

      normalize_vector_f(&bend_direction.x,&bend_direction.y,&bend_direction.z);
      normalize_vector_f(&BPd.x,&BPd.y,&BPd.z);
      cross_product_f ( BPd.x,BPd.y,BPd.z,bend_direction.x,bend_direction.y,bend_direction.z,&BX.x,&BX.y,&BX.z);
      if(BX.x==0 && BX.y==0 && BX.z==0){
          return D_O_K; /* fail */
      }
      save_undo();
      if(bend_selected_p)
         bend_selected(bend_point1,BX,BA);
      else
         bend_mesh(mesh,bend_point1,BX,BA);

      redraw_flatviews();
      dialog_message(&the_dialog[THREEDPROC], MSG_DRAW, 0, &obj);
    }
    return D_O_K;
}

int bendangleslidercallback(void *dp3 UNUSED, int d2);
char bendanglestring[80];
char bendpoint1xstring[12];
char bendpoint1ystring[12];
char bendpoint1zstring[12];
char bendpoint2xstring[12];
char bendpoint2ystring[12];
char bendpoint2zstring[12];

char benddirectionxstring[12];
char benddirectionystring[12];
char benddirectionzstring[12];
#define BENDANGLESLIDERLENGTH 99

DIALOG bend_dialog[]={
/* (dialog proc)     (x)   (y)   (w)   (h)     (fg)  (bg)  (key)    (flags)           (d1)  (d2)  (dp) */
  { d_box_proc,        0,   0,   260, 220,      0,   DBG,    0,         0,              0,    0,   NULL,NULL,NULL},
  

  { d_text_proc,      10,  22,     0,   0,      0,   DBG,    0,         0,              0,    0,    "Angle:" ,NULL,NULL},
  { d_slider_proc,    10,  32,   120,  10,      0,   DBG,    0,         0, BENDANGLESLIDERLENGTH,    0,    NULL,     bendangleslidercallback,NULL},
  { d_text_proc,      70,  44,    60,  10,      0,   DBG,    0,         0,              0,    0,    bendanglestring ,NULL,NULL},

  { d_text_proc,      10,  90,     0,   0,      0,   DBG,    0,         0,              0,    0,    "X:" ,NULL,NULL},
  { d_text_proc,      10, 106,     0,   0,      0,   DBG,    0,         0,              0,    0,    "Y:" ,NULL,NULL},
  { d_text_proc,      10, 122,     0,   0,      0,   DBG,    0,         0,              0,    0,    "Z:" ,NULL,NULL},
  
  { d_text_proc,     30,  76,     0,   0,      0,   DBG,    0,         0,              0,    0,    "Point 1" ,NULL,NULL},
  { d_edit_proc,     30,  90,    50,  12,      0,   255,     0,         0,               8,    0,   bendpoint1xstring,NULL,NULL},
  { d_edit_proc,     30, 106,    50,  12,      0,   255,     0,         0,               8,    0,   bendpoint1ystring,NULL,NULL},
  { d_edit_proc,     30, 122,    50,  12,      0,   255,     0,         0,               8,    0,   bendpoint1zstring,NULL,NULL},

  { d_text_proc,     100, 76,     0,   0,      0,   DBG,    0,         0,              0,    0,    "Point 2" ,NULL,NULL},
  { d_edit_proc,     100, 90,    50,  12,      0,   255,     0,         0,               8,    0,   bendpoint2xstring,NULL,NULL},
  { d_edit_proc,     100,106,    50,  12,      0,   255,     0,         0,               8,    0,   bendpoint2ystring,NULL,NULL},
  { d_edit_proc,     100,122,    50,  12,      0,   255,     0,         0,               8,    0,   bendpoint2zstring,NULL,NULL},
  
  { d_text_proc,     170, 76,     0,   0,      0,   DBG,    0,         0,              0,    0,    "Direction" ,NULL,NULL},
  { d_edit_proc,     170, 90,    50,  12,      0,   255,    0,         0,               8,    0,   benddirectionxstring,NULL,NULL},
  { d_edit_proc,     170,106,    50,  12,      0,   255,    0,         0,               8,    0,   benddirectionystring,NULL,NULL},
  { d_edit_proc,     170,122,    50,  12,      0,   255,    0,         0,               8,    0,   benddirectionzstring,NULL,NULL},

  { d_check_proc,     10, 150,    150,  12,      0,   DBG,    's',         0,              0,    0,    "only Selected" ,NULL,NULL},
  
  { d_button_proc,    10,  190,    20,  20,      0,   255,   13,    D_EXIT,              0,    0,   "OK",NULL,NULL},
  { d_button_proc,    40,  190,    60,  20,      0,   255,    0,    D_EXIT,              0,    0,   "Cancel",NULL,NULL},
  
  { NULL,              0,    0,    0,   0,      0,     0,    0,         0,              0,    0,    NULL ,NULL,NULL}
};
#define BENDANGLESLIDER 2
#define BENDANGLETEXT 3
#define BENDSELECTEDCHECK 19
#define BENDCANCEL 21

int bendangleslidercallback(void *dp3 UNUSED, int d2)
{
        int obj;
        DIALOG *dtext=&bend_dialog[BENDANGLETEXT];
        bend_angle=15*(float)(d2)/(float)BENDANGLESLIDERLENGTH;
        sprintf(bendanglestring,"%1.3f",bend_angle);
        rectfill(screen,dtext->x,dtext->y, dtext->x + dtext->w,dtext->y + dtext->h,dtext->bg);
        dialog_message(dtext, MSG_DRAW,0, &obj);
        return(D_O_K);
}
int do_bend_dialog(){
    int popupret;
    float holdbendangle=bend_angle;
    sprintf(bendanglestring,"%1.3f",bend_angle);
    sprintf(bendpoint1xstring,"%1.1f",bend_point1.x);
    sprintf(bendpoint1ystring,"%1.1f",-bend_point1.y);
    sprintf(bendpoint1zstring,"%1.1f",bend_point1.z);
    sprintf(bendpoint2xstring,"%1.1f",bend_point2.x);
    sprintf(bendpoint2ystring,"%1.1f",-bend_point2.y);
    sprintf(bendpoint2zstring,"%1.1f",bend_point2.z);
    sprintf(benddirectionxstring,"%1.1f",bend_direction.x);
    sprintf(benddirectionystring,"%1.1f",-bend_direction.y);
    sprintf(benddirectionzstring,"%1.1f",bend_direction.z);
    
    bend_dialog[BENDANGLESLIDER].d2=(int)(bend_angle/15*BENDANGLESLIDERLENGTH);

    centre_dialog(bend_dialog);
    popupret=popup_dialog(bend_dialog,2);
    if(popupret==BENDCANCEL){
       bend_angle=holdbendangle;

       return 0;
    }
    bend_selected_p=bend_dialog[BENDSELECTEDCHECK].flags&D_SELECTED;
    bend_point1.x=atof(bendpoint1xstring);
    bend_point1.y=-atof(bendpoint1ystring);
    bend_point1.z=atof(bendpoint1zstring);
    bend_point2.x=atof(bendpoint2xstring);
    bend_point2.y=-atof(bendpoint2ystring);
    bend_point2.z=atof(bendpoint2zstring);
    bend_direction.x=atof(benddirectionxstring);
    bend_direction.y=-atof(benddirectionystring);
    bend_direction.z=atof(benddirectionzstring);

    return 1;
}

/***/

char repulsepointxstring[12];
char repulsepointystring[12];
char repulsepointzstring[12];
char repulsepointcstring[12];
char repulsepointgexpstring[12];
char repulsepointexpmultstring[12];

DIALOG repulse_point_dialog[]={
/* (dialog proc)     (x)   (y)   (w)   (h)     (fg)  (bg)  (key)    (flags)           (d1)  (d2)  (dp) */
  { d_box_proc,        0,   0,   260, 220,      0,   DBG,    0,         0,              0,    0,   NULL,NULL,NULL},
  
  { d_text_proc,      10,  20,     0,   0,      0,   DBG,    0,         0,              0,    0,    "X -> X+c(X-P)e^(-a|X-P|^b)" ,NULL,NULL},

  { d_text_proc,      10,  90,     0,   0,      0,   DBG,    0,         0,              0,    0,    "X:" ,NULL,NULL},
  { d_text_proc,      10, 106,     0,   0,      0,   DBG,    0,         0,              0,    0,    "Y:" ,NULL,NULL},
  { d_text_proc,      10, 122,     0,   0,      0,   DBG,    0,         0,              0,    0,    "Z:" ,NULL,NULL},
  
  { d_text_proc,     30,  76,     0,   0,      0,   DBG,    0,         0,              0,    0,    "point" ,NULL,NULL},
  { d_edit_proc,     30,  90,    50,  12,      0,   255,     0,         0,               8,    0,   repulsepointxstring,NULL,NULL},
  { d_edit_proc,     30, 106,    50,  12,      0,   255,     0,         0,               8,    0,   repulsepointystring,NULL,NULL},
  { d_edit_proc,     30, 122,    50,  12,      0,   255,     0,         0,               8,    0,   repulsepointzstring,NULL,NULL},

  { d_text_proc,     100, 44,     0,   0,      0,   DBG,    0,         0,              0,    0,    "c" ,NULL,NULL},
  { d_edit_proc,     100, 60,    70,  12,      0,   255,     0,         0,               8,    0,   repulsepointcstring,NULL,NULL},

  { d_text_proc,     100, 76,     0,   0,      0,   DBG,    0,         0,              0,    0,    "a" ,NULL,NULL},
  { d_edit_proc,     100, 90,    70,  12,      0,   255,     0,         0,               8,    0,   repulsepointexpmultstring,NULL,NULL},
  
  { d_text_proc,     180, 76,     0,   0,      0,   DBG,    0,         0,              0,    0,    "b" ,NULL,NULL},
  { d_edit_proc,     180, 90,    50,  12,      0,   255,    0,         0,               8,    0,   repulsepointgexpstring,NULL,NULL},

  { d_check_proc,     10,  150,    150,   12,      0,   DBG,   's',         0,              0,    0,    "only Selected" ,NULL,NULL},
  
  { d_button_proc,    10,  190,    20,  20,      0,   255,   13,    D_EXIT,              0,    0,   "OK",NULL,NULL},
  { d_button_proc,    40,  190,    60,  20,      0,   255,    0,    D_EXIT,              0,    0,   "Cancel",NULL,NULL},
  
  { NULL,              0,    0,    0,   0,      0,     0,    0,         0,              0,    0,    NULL ,NULL,NULL}
};

#define REPULSEPOINTXTEXT 2
#define REPULSEPOINTYTEXT 3
#define REPULSEPOINTZTEXT 4
#define REPULSEPOINTpointTEXT 5

#define REPULSEPOINTXEDIT 6
#define REPULSEPOINTYEDIT 7
#define REPULSEPOINTZEDIT 8

#define REPULSEPOINTSELECTEDCHECK 15
#define REPULSEPOINTCANCEL 17

int do_repulse_point_dialog(){
    int popupret;
    
    sprintf(repulsepointxstring,"%1.1f",repulse_point.x);
    sprintf(repulsepointystring,"%1.1f",-repulse_point.y);
    sprintf(repulsepointzstring,"%1.1f",repulse_point.z);
    sprintf(repulsepointgexpstring,"%1.2f",repulse_point_gauss_exp);
    sprintf(repulsepointcstring,"%1.2f",repulse_point_c);
    sprintf(repulsepointexpmultstring,"%1.4f",repulse_point_expmult);

    centre_dialog(repulse_point_dialog);
    popupret=popup_dialog(repulse_point_dialog,2);
    if(popupret==REPULSEPOINTCANCEL){


       return 0;
    }
    repulsepoint_selected_p=repulse_point_dialog[REPULSEPOINTSELECTEDCHECK].flags&D_SELECTED;
    repulse_point.x=atof(repulsepointxstring);
    repulse_point.y=-atof(repulsepointystring);
    repulse_point.z=atof(repulsepointzstring);
    repulse_point_c=atof(repulsepointcstring);
    repulse_point_gauss_exp=atof(repulsepointgexpstring);
    repulse_point_expmult=fabs(atof(repulsepointexpmultstring));
    return 1;
}

void do_repulse_mouse_dialog(){
   repulse_point_dialog[REPULSEPOINTXTEXT].flags|=D_HIDDEN;
   repulse_point_dialog[REPULSEPOINTYTEXT].flags|=D_HIDDEN;
   repulse_point_dialog[REPULSEPOINTZTEXT].flags|=D_HIDDEN;
   repulse_point_dialog[REPULSEPOINTpointTEXT].flags|=D_HIDDEN;
   repulse_point_dialog[REPULSEPOINTXEDIT].flags|=D_HIDDEN;
   repulse_point_dialog[REPULSEPOINTYEDIT].flags|=D_HIDDEN;
   repulse_point_dialog[REPULSEPOINTZEDIT].flags|=D_HIDDEN;


   do_repulse_point_dialog();
   
   repulse_point_dialog[REPULSEPOINTXTEXT].flags&=~D_HIDDEN;
   repulse_point_dialog[REPULSEPOINTYTEXT].flags&=~D_HIDDEN;
   repulse_point_dialog[REPULSEPOINTZTEXT].flags&=~D_HIDDEN;
   repulse_point_dialog[REPULSEPOINTpointTEXT].flags&=~D_HIDDEN;
   repulse_point_dialog[REPULSEPOINTXEDIT].flags&=~D_HIDDEN;
   repulse_point_dialog[REPULSEPOINTYEDIT].flags&=~D_HIDDEN;
   repulse_point_dialog[REPULSEPOINTZEDIT].flags&=~D_HIDDEN;

}


int repulse_mouse(int view, int mx,int my){
    int vi,si;
    VTX *v;

    switch(view){
        case XYVIEW:
             repulse_mouse_vtx.x=(float)mx;
             repulse_mouse_vtx.y=(float)my;
        break;
        case ZYVIEW:
             repulse_mouse_vtx.z=(float)mx;
             repulse_mouse_vtx.y=(float)my;
        break;
        case XZVIEW:
             repulse_mouse_vtx.x=(float)mx;
             repulse_mouse_vtx.z=(float)my;
        break;

    }
    if(repulsepoint_selected_p){
      for(si=0;si<num_selected;si++){
         v=&(mesh->points[curVertex[si]]);
         vtx_repulse_point(v,repulse_mouse_vtx,repulse_point_c,repulse_point_expmult, repulse_point_gauss_exp);
      }

    }
    else{
      for(vi=0;vi<mesh->num_vertices;vi++){
         v=&(mesh->points[vi]);
         vtx_repulse_point(v,repulse_mouse_vtx,repulse_point_c,repulse_point_expmult, repulse_point_gauss_exp);
      }
    }
    return 1;
}

MESH * deform_connected_selection_copy(){
    MESH *dmesh;

    int vi,fi;
    int s1,s2,s3,n_s;
    TRIANGLE T;

    if(num_selected==0){
        return NULL;
    }
    vi=curVertex[0];
    num_selected=0;
    select_connected_component(vi);
    n_s=num_selected;
    dmesh=(MESH *)malloc(sizeof(MESH));
    if(dmesh==NULL){
    
          return NULL;
    }

    dmesh->points=NULL;
    dmesh->faces=NULL;
    dmesh->edges=NULL;

    dmesh->output_points=NULL;
    dmesh->output_faces=NULL;

    dmesh->num_vertices=0;
    dmesh->num_faces=0;
      
    dmesh->num_vertices_allocated=mesh_vtx_alloc_size;
    dmesh->num_faces_allocated=mesh_face_alloc_size;
    dmesh->num_edges_allocated=mesh_edge_alloc_size;

    dmesh->points=(VTX *)malloc(sizeof( VTX)*(mesh_vtx_alloc_size));
    dmesh->faces=(TRIANGLE *)malloc(sizeof( TRIANGLE)*(mesh_face_alloc_size));
    dmesh->edges=(EDGE *)malloc(sizeof( EDGE)*(mesh_edge_alloc_size));

      
    if(dmesh->points==NULL || dmesh->faces==NULL || dmesh->edges==NULL)
    {
           if(dmesh->points)
                free(dmesh->points);
           if(dmesh->faces)
                free(dmesh->faces);
           if(dmesh->edges)
                free(dmesh->edges);
           free(dmesh);
           return NULL;
    }

    for(vi=0;vi<num_selected;vi++){
         dmesh->points[vi]=mesh->points[curVertex[vi]];
         dmesh->num_vertices++;
    }

    for(fi=0;fi<mesh->num_faces;fi++){
            T=mesh->faces[fi];
            if(triangle_in_selected(T,&s1,&s2,&s3)){
                new_triangle(dmesh,s1,s2,s3,0);
            }
    }
    return dmesh;
}

float smoothing=1;
int smoothingavg=0;
int smooth_selected_p=0;
int do_smoothing_dialog();
int do_smoothsubdivide(){
    int obj;
    int Lret=0,alertret,goahead=0;
    MESH *subdmesh;

    int vi,n_s,n_v;
    if(do_smoothing_dialog()==0)
        return D_O_K;
    if(smooth_selected_p){
      subdmesh=deform_connected_selection_copy();
      n_s=subdmesh->num_vertices;
      if(subdmesh==NULL){
          return D_O_K;
      }
    }
    else {
      subdmesh=mesh;
      n_s=num_selected;
    }
    while(Lret==0){

        save_undo();
        waiting_call();
        Lret=SmoothSubdivide(subdmesh,smoothing,smoothingavg,goahead);
        
        waiting_done();
        if(Lret==0){
               alertret=alert("Mesh may be singular","Continue?",NULL,"OK","Cancel",13,0);
               if (alertret==2)
                  break;
               else goahead=1;
        }        
    }
    if(Lret>0){
       if(smooth_selected_p){
         delete_selected();
         n_v=mesh->num_vertices;
       
         add_mesh_to_mesh(mesh,subdmesh);
         for(vi=0;vi<n_s;vi++){
            add_to_selected(n_v+vi, 0);
         }
       }
       enabledisableitems();
       update_numbers();
       redraw_flatviews();
       dialog_message(&the_dialog[THREEDPROC], MSG_DRAW, 0, &obj);
    }    
    else if(Lret==-1){
               alertret=alert("Too many triangles to subdivide!",NULL,NULL,"OK",NULL,13,0);
    }
    else if(Lret==-2){
               alertret=alert("Too many vertices to subdivide!",NULL,NULL,"OK",NULL,13,0);
    }
    else if(Lret==-3){
               alertret=alert("Too many edges to subdivide!",NULL,NULL,"OK",NULL,13,0);
    }    
    else if(Lret==-4){
              error_exit("memory allocation error in Smooth subdivision",NULL);
    }
    if(smooth_selected_p)
      delete_mesh(subdmesh);
    return D_O_K;
}

int do_extrude_selected(){
    int obj;
    int s,f,ns,nv,nf;
    int s1,s2,s3;
    int v1=-1,v2=-1,v3=-1,v4=-1;//supress compiler warning
    int ei,e,o;
    int fcount=0;
    VTX vtx;
    TRIANGLE T;
    int ok=0;
    
    ns=num_selected;
    nv=mesh->num_vertices;
    nf=mesh->num_faces;
    if(ns+nv>=mesh->num_vertices_allocated-1){
       alert("Too many vertices to duplicate!",NULL,NULL,"OK",NULL,13,0);
       return D_O_K;
    }
    for(f=0;f<nf;f++){
    /* find out how many new triangles we will need */
            T=mesh->faces[f];
            if(triangle_in_selected(T,&s1,&s2,&s3))
                fcount++;
    }
    if(nf+fcount>=mesh->num_faces_allocated-1){
       alert("Too many triangles to duplicate!",NULL,NULL,"OK",NULL,13,0);
       return D_O_K;
    }
    waiting_call();
    make_edge_list(mesh);

    for(s=0;s<num_selected;s++){
        vtx=mesh->points[curVertex[s]];
        new_vertex(mesh,vtx,0);
    }
    
    for(f=0;f<nf;f++){
            T=mesh->faces[f];
            if(triangle_in_selected(T,&s1,&s2,&s3)){
                new_triangle(mesh,nv+s2,nv+s1,nv+s3,T.oflag);
            }
    }
    for(f=0;f<nf;f++){
       T=mesh->faces[f];
       for(e=0;e<3;e++){
              switch(e){
                case 0:
                  v1=T.v1;v2=T.v2;

                  break;
                case 1:
                  v1=T.v2;v2=T.v3;

                  break;
                case 2:
                  v1=T.v3;v2=T.v1;

                  break;
              }
              s1=vertex_in_selected(v1);
              s2=vertex_in_selected(v2);
              if(s1<0||s2<0) continue;
              v3=nv+s2;v4=nv+s1;
              ei=find_edge_in_meshedges(mesh,v1,v2, &o);
              if(ei<0) continue;
              if(mesh->edges[ei].t1>=0 && mesh->edges[ei].t2<0){
                 new_triangle(mesh,v1,v2,v4,0);
                 new_triangle(mesh,v4,v3,v2,0);
                 ok=1;
              }
       }
    }

    if(!ok){
       waiting_done();
       mesh->num_vertices=nv;
       mesh->num_faces=nf;
       mesh->num_edges=0;
       alert("The selection contains",
             "no boundary to extrude",
             NULL,"OK",NULL,13,0);
       return D_O_K;
    }
    try_orient(mesh);
    num_selected=0;
    for(s=0;s<ns;s++)
        add_to_selected(nv+s,0);
    waiting_done();
    update_numbers();
    redraw_flatviews();
    dialog_message(&the_dialog[THREEDPROC], MSG_DRAW, 0, &obj);
    
    set_mouse_mode_flat(MOUSEMODEFLAT_MOVESELECTED);
    enabledisableitems();

    return D_O_K;
}

int do_paste_copy(){
      int obj,vi;
      int prev_n_v=mesh->num_vertices;

      int new_n_v=add_mesh_to_mesh(mesh,mesh_copy);
      if(new_n_v==-1){
          alert("Too many vertices to paste!",NULL,NULL,"OK",NULL,13,0);
          return D_O_K;
      }
      if(new_n_v==-2){
          alert("Too many triangles to paste!",NULL,NULL,"OK",NULL,13,0);
          return D_O_K;
      }
      if(mesh_copy==NULL){
          return D_O_K;
      }
      num_selected=0;
      for(vi=prev_n_v;vi<mesh->num_vertices;vi++){
          add_to_selected(vi,0);
      }
      update_numbers();
      redraw_flatviews();
      dialog_message(&the_dialog[THREEDPROC], MSG_DRAW, 0, &obj);

      enabledisableitems();

      set_mouse_mode_flat(MOUSEMODEFLAT_MOVESELECTED);
      return D_O_K;
}

MESH * copy_selected_to_new_mesh(){
      MESH *newmesh=NULL;
      int n_v=0,
          n_f=0;
      int vi,fi;
      int s1,s2,s3;
      TRIANGLE T;

      if(num_selected==0)return NULL;
      newmesh=(MESH *)malloc(sizeof(MESH));
      if(newmesh==NULL){
          error_exit("memory allocation error for mesh copy",NULL);
      }
      n_v=num_selected;
      newmesh->output_points=NULL;
      newmesh->output_faces=NULL;
      newmesh->edges=NULL;

      for(fi=0;fi<mesh->num_faces;fi++){
      // find out how many triangles we need
            T=mesh->faces[fi];
            if(triangle_in_selected(T,&s1,&s2,&s3))
                n_f++;
      }
      newmesh->num_vertices=0;
      newmesh->num_faces=0;
      
      newmesh->num_vertices_allocated=n_v+1;
      newmesh->num_faces_allocated=n_f+1;
      newmesh->num_vertices=0;
      newmesh->num_faces=0;

      newmesh->points=(VTX *)malloc(sizeof( VTX)*(n_v+1));
      newmesh->faces=(TRIANGLE *)malloc(sizeof( TRIANGLE)*(n_f+1));

      
      if(newmesh->points==NULL || (n_f && newmesh->faces==NULL))
      {
          error_exit("memory allocation error for mesh copy",NULL);
      }

      for(vi=0;vi<n_v;vi++){
         newmesh->points[vi]=mesh->points[curVertex[vi]];
         newmesh->num_vertices++;
      }

      for(fi=0;fi<mesh->num_faces;fi++){

            T=mesh->faces[fi];
            if(triangle_in_selected(T,&s1,&s2,&s3)){
                new_triangle(newmesh,s1,s2,s3,0);
            }
      }
      return newmesh;
}

int do_make_copy_of_selected(){
      if(num_selected==0)return D_O_K;
      if(mesh_copy){
         delete_mesh(mesh_copy);
      }
      mesh_copy=copy_selected_to_new_mesh();
      if(mesh_copy!=NULL){
         edit_menu[PASTECOPYITEM].flags&=~D_DISABLED;
      }

      return D_O_K;
}

int do_dup_selected(){
    int obj;
    int s,f,ns,nv,nf;
    int s1,s2,s3;
    int fcount=0;
    VTX vtx;
    TRIANGLE T;

    ns=num_selected;
    nv=mesh->num_vertices;
    nf=mesh->num_faces;
    if(ns+nv>=mesh->num_vertices_allocated-1){
       alert("Too many vertices to duplicate!",NULL,NULL,"OK",NULL,13,0);
       return D_O_K;
    }
    for(f=0;f<nf;f++){
    // find out how many new triangles we will need
            T=mesh->faces[f];
            if(triangle_in_selected(T,&s1,&s2,&s3))
                fcount++;
    }
    if(nf+fcount>=mesh->num_faces_allocated-1){
       alert("Too many triangles to duplicate!",NULL,NULL,"OK",NULL,13,0);
       return D_O_K;
    }

    for(s=0;s<num_selected;s++){
        vtx=mesh->points[curVertex[s]];
        new_vertex(mesh,vtx,0);
    }

    for(f=0;f<nf;f++){
            T=mesh->faces[f];
            if(triangle_in_selected(T,&s1,&s2,&s3)){
                new_triangle(mesh,nv+s1,nv+s2,nv+s3,T.oflag);
            }
    }
    num_selected=0;
    for(s=0;s<ns;s++)
        add_to_selected(nv+s,0);
        
    update_numbers();
    redraw_flatviews();
    dialog_message(&the_dialog[THREEDPROC], MSG_DRAW, 0, &obj);

    enabledisableitems();
    set_mouse_mode_flat(MOUSEMODEFLAT_MOVESELECTED);
    return D_O_K;
}

int do_detach_selected(){
/* detach the selection from the rest of the mesh*/
    int obj;
    int s,ns,nv,nf;
    int newvtx,ti;
    int s1,s2,s3;

    VTX vtx;
    TRIANGLE T;

    ns=num_selected;
    nv=mesh->num_vertices;
    nf=mesh->num_faces;    

    for(s=0;s<num_selected;s++){
        newvtx=-1;
        vtx=mesh->points[curVertex[s]];
        /* if a triangle contains a selected vertex,
           but the triangle itself is not contained
           in the selection, make a copy of that vertex
           to 'leave behind' with the rest of the mesh so
           that triangles won't be destroyed */
        for(ti=0;ti<mesh->num_faces;ti++){
           T=mesh->faces[ti];
           if( triangle_has_vertex_p(mesh,ti,curVertex[s])){
              if(!triangle_in_selected(T,&s1,&s2,&s3)){

                 if(newvtx==-1)
                   newvtx=new_vertex(mesh,vtx,0);
                 if(curVertex[s]==T.v1)
                     mesh->faces[ti].v1=newvtx;
                 if(curVertex[s]==T.v2)
                     mesh->faces[ti].v2=newvtx;
                 if(curVertex[s]==T.v3)
                     mesh->faces[ti].v3=newvtx;
              }
           }
        }
    }


        
    update_numbers();
    redraw_flatviews();
    dialog_message(&the_dialog[THREEDPROC], MSG_DRAW, 0, &obj);

    enabledisableitems();
    set_mouse_mode_flat(MOUSEMODEFLAT_MOVESELECTED);
    return D_O_K;
}

int do_select_all(){
    int i;
    num_selected=0;
    for(i=0;i<mesh->num_vertices;i++){
          add_to_selected(i,0);
    }
    update_numbers();
    redraw_flatviews();
    enabledisableitems();
    return D_O_K;
}

int do_unselect_all(){
    num_selected=0;
    update_numbers();
    redraw_flatviews();
    enabledisableitems();
    return D_O_K;
}


int do_reduce_one(){
      int obj;
      ProgressiveMesh(mesh);
      num_selected=0;
      update_numbers();
      redraw_flatviews();
      dialog_message(&the_dialog[THREEDPROC], MSG_DRAW, 0, &obj);
      enabledisableitems();


      return D_O_K;
}
int do_invert_selection(){
    int i,j,found;
    int temp[num_selected];
    int ntemp=num_selected;

    for(i=0;i<num_selected;i++){
        temp[i]=curVertex[i];
    }
    num_selected=0;
    for(i=0;i<mesh->num_vertices;i++){
          found=0;
          for(j=0;j<ntemp;j++){
             if(i==temp[j])
                  found=1;
          }
          if(found==0)
            add_to_selected(i,0);
    }
    update_numbers();
    redraw_flatviews();
    enabledisableitems();
    return D_O_K;
}

MENU reflect_menu[]={
   {"Reflect &X",NULL,NULL,0,NULL},
   {"Reflect &Y",NULL,NULL,0,NULL},
   {"Reflect &Z",NULL,NULL,0,NULL},
   { NULL,NULL,NULL,0,NULL}
};


int do_reflect_selected(){
    int menuret;

    menuret=do_menu(reflect_menu,mouse_x,mouse_y);
    if(menuret>=0){
      mirror_selected(menuret);
      redraw_flatviews();
      redraw_3dview();
    }
    return D_O_K;
}

int ok_timer=0;
volatile int wait_counter;
BITMAP * wm[4]={NULL,NULL,NULL,NULL};
void wait_proc(){
     wait_counter = (wait_counter+1)%4;
     set_mouse_sprite(wm[wait_counter]);
     //if(esc_pressed()) escape=1; /***/
}
END_OF_FUNCTION(wait_proc);

int waiting=0;
void waiting_call(){
     ok_timer=0;
     if(install_int_ex(wait_proc, BPS_TO_TIMER(8))==0)
        ok_timer=1;
     else
       set_mouse_sprite(waitmouse);
     waiting=1;
}
void waiting_done(){
     if(ok_timer){
        remove_int(wait_proc);
     }
     set_mouse_sprite(NULL);
     waiting=0;
}


int read_ini_file(char *filename){
   FILE *inifp=NULL;
   int s_w=def_s_w,
       s_h=def_s_h;
   char res_string[256]="";
   char *wptr,*hptr;
   inifp=fopen(filename,"rt");
   if(inifp){
      fgets(res_string, 255, inifp);
      wptr=res_string;
      hptr=strchr(res_string,'x');hptr++;
      s_w=atoi(wptr);
      s_h=atoi(hptr);
      fclose(inifp);
   }
   else{
     set_dialog_view_sizes(s_w,s_h);
     return 0;
   }
   set_dialog_view_sizes(s_w,s_h);
   return 1;
}
int write_ini_file(char *filename){
   FILE *inifp=NULL;
   char res_string[256]="";
   inifp=fopen(filename,"wt");
   if(inifp){
      sprintf(res_string,"%dx%d",SCREENWIDE,SCREENHIGH);
      fputs(res_string,  inifp);
      fclose(inifp);
   }
   else return 0;

   return 1;
}

char pathdatafilename[256]="";
char allegrodatafilename[256]="triangles.dat";

FILE *Dfp;

int main(int argc, char *argv[])
{
   char *argptr;
   int numargsleft=argc-1;
   int argnum=1;
   int loadret;
   int readinteger;

   ShadingPaletteBottom=GreyScaleBottom;
   ShadingPaletteTop=GreyScaleTop;

   read_ini_file(def_ini_filename);
   if(argc>1){
         while( numargsleft){
           argptr=argv[argnum];
           switch(*argptr){
              case '-':
                 argptr++;
                 switch(*argptr){
                     case 'a':
                        printf("\nTriangles build number %d\n", BUILD_NUMBER);
                        exit(0);
                     break;
                     case 'h':

                     break;
                     case 'm':
                        argptr++;
                        readinteger=atoi(argptr);
                        if(readinteger){
                           mesh_vtx_alloc_size=readinteger*1000;
                           mesh_face_alloc_size =readinteger*1000;
                           mesh_edge_alloc_size=2*readinteger*1000;
                           curVertex_alloc_size=mesh_vtx_alloc_size;
                        }
                     break;
                 }
                 break;
              case '+':
                 argptr++;
                 switch(*argptr){
                     case 'w':
                        argptr++;

                        
                     break;
                     case 'h':
                        argptr++;

                     break;
                 }
                 break;
              default:
                 strcpy(savetripath,argptr);
           }
           numargsleft--;
           argnum++;
         }

   }

   /* setting up */
   allegro_init();
   install_keyboard();
   install_timer();
   install_mouse();

   if (set_gfx_mode(GFX_AUTODETECT, SCREENWIDE, SCREENHIGH, 0, 0) != 0) {
      allegro_exit();
      printf("Error setting graphics mode\n%s\n\n", allegro_error);
      return 1;
   }
   LOCK_VARIABLE(ok_timer);
   LOCK_VARIABLE(wm);
   LOCK_VARIABLE(wait_counter);
   LOCK_FUNCTION(wait_proc);
#ifdef USEEXEDAT
   allegrodata=load_datafile("#");
#else
   allegrodata=load_datafile(allegrodatafilename);
#endif
   if(!allegrodata){
       error_exit("Couldn't find the program's 'data' file",NULL);
   }
   else{
      set_palette(allegrodata[THEPALETTE].dat);
      notribmp=allegrodata[NOTRISELBMP].dat;
      threetribmp=allegrodata[THREETRISELBMP].dat;
      threenotribmp=allegrodata[THREENOTRISELBMP].dat;
      waitmouse=allegrodata[WAITMOUSEBMP].dat;
      wm[0]=allegrodata[WM0BMP].dat;
      wm[1]=allegrodata[WM1BMP].dat;
      wm[2]=allegrodata[WM2BMP].dat;
      wm[3]=allegrodata[WM3BMP].dat;
      the_dialog[TRISELBMP].dp=notribmp;
      the_dialog[XYPROC].dp2=allegrodata[XYAXISBMP].dat;
      the_dialog[ZYPROC].dp2=allegrodata[ZYAXISBMP].dat;
      the_dialog[XZPROC].dp2=allegrodata[XZAXISBMP].dat;
      the_dialog[DEFORMBUTTON].dp=allegrodata[DEFORMICON].dat;
      the_dialog[DEFORMBUTTON].w=((BITMAP*)allegrodata[DEFORMICON].dat)->w;
      the_dialog[DEFORMBUTTON].h=((BITMAP*)allegrodata[DEFORMICON].dat)->h;
      
   }
   gui_fg_color=0;
   gui_bg_color=230;

   /* bitmap buffers */
   threedbmp = create_bitmap(VIEWS_WIDTH, VIEWS_HEIGHT);
   xybmp     = create_bitmap(VIEWS_WIDTH, VIEWS_HEIGHT);
   zybmp     = create_bitmap(VIEWS_WIDTH, VIEWS_HEIGHT);
   xzbmp     = create_bitmap(VIEWS_WIDTH, VIEWS_HEIGHT);

   if(threedbmp==NULL||xybmp==NULL||zybmp==NULL||xzbmp==NULL){
      error_exit("memory allocation error for bitmaps",NULL);
   }
   the_dialog[XYPROC].dp=xybmp;
   the_dialog[ZYPROC].dp=zybmp;
   the_dialog[XZPROC].dp=xzbmp;
   the_dialog[THREEDPROC].dp=threedbmp;
   set_mouse_mode_flat(MOUSEMODEFLAT_CREATEVERTEX);
   set_mouse_mode_3d(MOUSEMODE3d_ROTATE0);

   /* set up the viewport for the perspective projection */
   set_projection_viewport(0, 0, VIEWS_WIDTH, VIEWS_HEIGHT);

   /* make room for the mesh */
   mesh=new_mesh(mesh_vtx_alloc_size,mesh_face_alloc_size,mesh_edge_alloc_size,
                 Initial_v,Initial_r,Initial_dz,Initial_dr,Initial_scale);
   if(mesh==NULL){
          error_exit("memory allocation error for new mesh",NULL);
   }
   /* the 'selected vertices' index array */
   curVertex=(int*)malloc(sizeof(int)*curVertex_alloc_size);
   if(curVertex==NULL){
          error_exit("memory allocation error for selection buffer",NULL);
   }
   
   /* load a mesh if one was given on the command line */
   if(*savetripath){
       waiting_call();
       if(!stricmp(get_extension(savetripath),"raw")){
           loadret=load_RAW_triangle_mesh(mesh,savetripath);
       }
       else{
           loadret=load_TRI_triangle_mesh2(mesh,savetripath);
       }
       waiting_done();
       if(loadret<0){
         if(loadret>-5)
            alert("Mesh is too large!",NULL,NULL,"OK",NULL,13,0);
         else
            alert("Mesh mess!",NULL,NULL,"OK",NULL,13,0);
       }
       else
         enabledisableitems();

   }

   show_mouse(screen);
   //stackbytes=stackavail();
   do_dialog(the_dialog,-1);

   /* save whatever happens to be left */
   if(mesh->num_faces>0){
      save_TRI_triangle_mesh(mesh,"lasttri.tri",3);
   }

   write_ini_file(def_ini_filename);
   
   /* free the pieces */
   destroy_bitmap(threedbmp);
   destroy_bitmap(xybmp);
   destroy_bitmap(zybmp);
   destroy_bitmap(xzbmp);
   delete_mesh(mesh);

   free(curVertex);

   return 0;
}

MENU saveas_menu[]={
   {"&Tri",NULL,NULL,0,NULL},
   {"&POV triangle mesh",NULL,NULL,0,NULL},
   {"POV &smooth_triangle mesh",NULL,NULL,0,NULL},
   {"MegaPOV mesh2 (&flat)",NULL,NULL,0,NULL},
   {"MegaPOV mesh2 (s&Mooth)",NULL,NULL,0,NULL},
   {"&RAW triangles",NULL,NULL,0,NULL},
   {"R&AW smoothed triangles",NULL,NULL,0,NULL},
   {"PCM(1) triangles",NULL,NULL,0,NULL},
   {"PCM(1) smoothed triangles",NULL,NULL,0,NULL},
//   {"",NULL,NULL,0,NULL},
   {"VRML (1.0)",NULL,NULL,0,NULL},
   {"FE triangles",NULL,NULL,0,NULL},
   
   { NULL,NULL,NULL,0,NULL}
};

VECTOR *VNormals=NULL;
int save_smooth_p=1;
char *fs_extensions[]={
"TRI",
"INC;POV",
"INC;POV",
"INC;POV",
"INC;POV",

"RAW",
"RAW",
"PCM",
"PCM",
"WRL",
"FE"
};

int save_triangles_as(){
    int fsret,savret=-1;
    char ext[36];
    char *fptr;
    char hold[512];
    int savetype=0;
    int menuret;
    menuret=do_menu(saveas_menu,600,10);
    if(menuret==-1) return D_O_K;
    else {
       savetype=menuret;
    }
    strncpy(ext, fs_extensions[menuret], 3);
    ext[3]='\0';
    if(*savepath=='\0'){
       replace_extension(savepath, savetripath, ext, 511);
    }
    else{
       strcpy(hold,savepath);
       replace_extension(hold, savepath, ext, 511);
       strcpy(savepath,hold);
       replace_extension(savetripath, savepath, "tri", 511);

    }

    fsret=do_savefiledialog("Save to:", savepath,fs_extensions[menuret]);
    if(fsret==0) return D_O_K;
    fptr=get_filename(savepath);
    if(*fptr=='.'){
         alert("Can't save to", fptr, NULL, "OK",NULL, 13, 0);
         return D_O_K;
    }
    if(savetype==POVSMOOTH||savetype==MESH2SMOOTH||savetype==RAWSMOOTH||savetype==PCM1SMOOTH){
        save_smooth_p=1;
    }
    else save_smooth_p=0;
    
    strcpy(ext,get_extension(savepath));
    strlwr(ext);
    replace_extension(savetripath, savepath, "TRI", 511);

    if(VNormals){
       free(VNormals);
       VNormals=NULL;
    }


    if(save_smooth_p){
         VNormals=(VECTOR *)malloc(sizeof(VECTOR)*mesh->num_vertices);
         get_vertex_normals(mesh->faces,mesh->num_faces, mesh->num_vertices,VNormals,1);
    }
    if(savetype==POV||savetype==POVSMOOTH){
         savret=save_POV_triangle_mesh(mesh,savepath,VNormals);
    }
    else if(savetype==MESH2FLAT||savetype==MESH2SMOOTH){
         savret=save_mesh2(mesh,savepath,VNormals);
    }

    else if(savetype==PCM1||savetype==PCM1SMOOTH){
         savret=save_PCM_triangle_mesh(mesh,savepath,VNormals);
    }
    else if(savetype==RAW||savetype==RAWSMOOTH){
         savret=save_RAW_triangle_mesh(mesh,savepath,NULL);
    }
    else if(savetype==FE){
         savret=save_FE_triangle_mesh(mesh,savepath);
    }
    else if(savetype==WRL){
         savret=save_WRL_triangle_mesh(mesh,savepath);
    }

    else{
         if(savetype!=TRI){

           if(alert("Save as TRI?", NULL, NULL, "OK",NULL, 13, 0)==2)
               return D_O_K;

         }
         
         savret=save_TRI_triangle_mesh(mesh,savepath,0);
         strcpy(savetripath,savepath);
    }

    if(savret==-1){
         alert("Couldn't save", savepath, NULL, "OK",NULL, 13, 0);
    }
    if(VNormals){
       free(VNormals);
       VNormals=NULL;
    }
    return D_O_K;
}
int save_tri(){
    char temppath[256];
    
    char *fptr;
    int fsret,savret=-1;
    strcpy(temppath,savetripath);
    if(*savetripath=='\0'){
       replace_extension(savetripath, savepath, "TRI", 255);
    }
    else
       replace_extension(savetripath, temppath, "TRI", 255);

    fsret=do_savefiledialog("Save triangles to (.TRI)", savetripath, "TRI");
    if(fsret==0) return D_O_K;
    fptr=get_filename(savetripath);
    if(*fptr=='.'){
         alert("Can't save to", fptr, NULL, "OK",NULL, 13, 0);
         return D_O_K;
    }

    savret=save_TRI_triangle_mesh(mesh,savetripath,0);

    if(savret==-1){
         alert("Couldn't save", savetripath, NULL, "OK",NULL, 13, 0);
    }

    return D_O_K;
}
int save_selected_tri(){
    char temppath[256];
    char *fptr;
    int fsret,savret=-1;
    MESH *selected;
    strcpy(temppath,savetripath);
    if(*savetripath=='\0'){
       replace_extension(savetripath, savepath, "TRI", 255);
    }
    else
       replace_extension(savetripath, temppath, "TRI", 255);

    fsret=do_savefiledialog("Save triangles to (.TRI)", savetripath, "TRI");
    if(fsret==0) return D_O_K;
    fptr=get_filename(savetripath);
    if(*fptr=='.'){
         alert("Can't save to", fptr, NULL, "OK",NULL, 13, 0);
         return D_O_K;
    }

    selected=copy_selected_to_new_mesh();
    if(selected!=NULL){

         savret=save_TRI_triangle_mesh(selected,savetripath,0);

         delete_mesh(selected);
    }
    if(savret==-1){
         alert("Couldn't save", savetripath, NULL, "OK",NULL, 13, 0);
    }
    return D_O_K;
}

int load_tri(){
    int fsret=-1,loadret,obj;
    int i,nv=mesh->num_vertices;
    FILE *fp=NULL;
    char ext[36];
    while(fp==NULL&&fsret!=0){
       fsret=popup_file_select("Open (.TRI/.RAW)", savetripath, "TRI;RAW");
       if(fsret!=0){
          fp=fopen(savetripath,"rt");
          if(fp==NULL){
             alert("Unable to open file",NULL,NULL,"Ok",NULL,13,0);
             continue;
          }
          fclose(fp);
       }
    }
    if(fsret==0) return D_O_K;

    strcpy(ext,get_extension(savetripath));
    waiting_call();
    if(!stricmp(ext,"RAW")){
      loadret=load_RAW_triangle_mesh(mesh,savetripath);
    }
    else{
      loadret=load_TRI_triangle_mesh2(mesh,savetripath);
    }
    waiting_done();
    if(loadret<0){
       if(loadret>-5){
         alert("Mesh is too large!",NULL,NULL,"OK",NULL,13,0);
         return D_O_K;
       }
       else
         alert("Mesh mess!",NULL,NULL,"OK",NULL,13,0);

    }
    strcpy(savepath,savetripath);
    num_selected=0;
    for (i=nv;i<mesh->num_vertices;i++)
            add_to_selected(i,0);
    enabledisableitems();
    update_numbers();
    redraw_flatviews();
    dialog_message(&the_dialog[THREEDPROC], MSG_DRAW, 0, &obj);

    return D_O_K;
}

int do_get_undo(){
    int loadret,obj;

    num_selected=0;
    mesh->num_vertices=0;
    mesh->num_faces=0;

    init_shape(mesh,Initial_v,Initial_r,Initial_dz,Initial_dr,Initial_scale);

    waiting_call();
    loadret=load_TRI_triangle_mesh2(mesh,"lasttri.tri");
    waiting_done();
    if(loadret<0){
       if(loadret>-5)
         alert("Mesh is too large!",NULL,NULL,"OK",NULL,13,0);
       else
         alert("Mesh mess!",NULL,NULL,"OK",NULL,13,0);
    }
    enabledisableitems();
    update_numbers();
    redraw_flatviews();
    dialog_message(&the_dialog[THREEDPROC], MSG_DRAW, 0, &obj);

    return D_O_K;
}


int tabbedshiftselected(){

    int i;
    int direction=key_shifts & KB_SHIFT_FLAG ? LEFT:RIGHT;
    if(num_selected){
        for(i=0;i<num_selected;i++){
             if(direction==RIGHT){
                curVertex[i]=(curVertex[i]+1) %(mesh->num_vertices);
             }
             else{
                 if(curVertex[i]>0){
                    curVertex[i]=curVertex[i]-1;
                 }
                 else{
                    curVertex[i]=mesh->num_vertices-1;
                 }

             }
        }
        redraw_flatviews();
    }
    return D_O_K;
}
int orientedlighting(){
    int obj;
    normal_orient=normal_orient?0:1;
    dialog_message(&the_dialog[THREEDPROC], MSG_DRAW, 0, &obj);
    return D_O_K;
}

int quitter()
{
   if (alert("Really want to quit?", NULL, NULL, "Yes", "Cancel", 'y', 27) == 1)
      return D_CLOSE;
   else
      return D_O_K;
}


int smoothingslidercallback(void *dp3 UNUSED, int d2);
char smoothingstring[80];
#define SMOOTHINGSLIDERLENGTH 99

DIALOG smoothing_dialog[]={
/* (dialog proc)     (x)   (y)   (w)   (h)     (fg)  (bg)  (key)    (flags)           (d1)  (d2)  (dp) */
  { d_box_proc,        0,   0,   300, 150,      0,   DBG,    0,         0,              0,    0,   NULL,NULL,NULL},
  

  { d_text_proc,      10,  22,     0,   0,      0,   DBG,    0,         0,              0,    0,    "smoothing:" ,NULL,NULL},
  { d_slider_proc,    10,  32,   110,  10,      0,   DBG,    0,         0, SMOOTHINGSLIDERLENGTH,    0,    NULL,     smoothingslidercallback,NULL},
  { d_text_proc,      70,  44,    60,  10,      0,   DBG,    0,         0,              0,    0,    smoothingstring ,NULL,NULL},
  
  { d_check_proc,     10,  64,   200,  12,      0,   DBG,   'a',    0,         0,    0,   "Average between marked",NULL,NULL},

  { d_check_proc,     10,  90,   280,  12,      0,   DBG,   's',    0,         0,    0,   "apply only to Selected component",NULL,NULL},
  { d_button_proc,    10,  120,    20,  20,      0,   255,   13,    D_EXIT,              0,    0,   "OK",NULL,NULL},
  { d_button_proc,    40,  120,    60,  20,      0,   255,    0,    D_EXIT,              0,    0,   "Cancel",NULL,NULL},
  
  { NULL,              0,    0,    0,   0,      0,     0,    0,         0,              0,    0,    NULL ,NULL,NULL}
};
#define SMOOTHINGSLIDER 2
#define SMOOTHINGTEXT 3
#define SMOOTHINGAVGCHECK 4
#define SMOOTHSELECTEDCHECK 5
#define SMOOTHINGCANCEL 7

int smoothingslidercallback(void *dp3 UNUSED, int d2)
{
        int obj;
        DIALOG *dtext=&smoothing_dialog[SMOOTHINGTEXT];
        smoothing=3*(float)(d2)/(float)SMOOTHINGSLIDERLENGTH;
        sprintf(smoothingstring,"%1.3f",smoothing);
        rectfill(screen,dtext->x,dtext->y, dtext->x + dtext->w,dtext->y + dtext->h,dtext->bg);
        dialog_message(dtext, MSG_DRAW,0, &obj);
        return(D_O_K);
}
int do_smoothing_dialog(){
    int popupret;
    float holdsmoothing=smoothing;
    sprintf(smoothingstring,"%1.3f",smoothing);
    smoothing_dialog[SMOOTHINGSLIDER].d2=(int)(smoothing/3*SMOOTHINGSLIDERLENGTH);
    if(smoothingavg){
        smoothing_dialog[SMOOTHINGAVGCHECK].flags |= D_SELECTED;
    }
    else{
        smoothing_dialog[SMOOTHINGAVGCHECK].flags &= ~D_SELECTED;
    }
    centre_dialog(smoothing_dialog);
    popupret=popup_dialog(smoothing_dialog,2);
    if(popupret==SMOOTHINGCANCEL){
       smoothing=holdsmoothing;
       return 0;
    }
    smoothingavg=smoothing_dialog[SMOOTHINGAVGCHECK].flags & D_SELECTED;
    smooth_selected_p=smoothing_dialog[SMOOTHSELECTEDCHECK].flags & D_SELECTED;

    return 1;
}

int pullaslidercallback(void *dp3 UNUSED, int d2);
char pullastring[80];
char maxnedgestring[80];
#define PULLASLIDERLENGTH 99

DIALOG pull_dialog[]={
/* (dialog proc)     (x)   (y)   (w)   (h)     (fg)  (bg)  (key)    (flags)           (d1)  (d2)  (dp) */
  { d_box_proc,        0,   0,   200, 150,      0,   DBG,    0,         0,              0,    0,   NULL,NULL,NULL},
  

  { d_text_proc,      10,  22,     0,   0,      0,   DBG,    0,         0,              0,    0,    "Pull fall-off rate:" ,NULL,NULL},
  { d_slider_proc,    10,  32,   110,  10,      0,   DBG,    0,         0, PULLASLIDERLENGTH,    0,    NULL,     pullaslidercallback,NULL},
  { d_text_proc,      70,  44,    60,  10,      0,   DBG,    0,         0,              0,    0,    pullastring ,NULL,NULL},

  { d_text_proc,      50, 68,     0,   0,      0,   DBG,    0,         0,              0,    0,    "Max edges" ,NULL,NULL},
  { d_edit_proc,      50, 82,    32,  12,      0,   255,     0,         0,              4,    0,  maxnedgestring,NULL,NULL},

  { d_button_proc,    10,  120,    20,  20,      0,   255,   13,    D_EXIT,              0,    0,   "OK",NULL,NULL},
  { d_button_proc,    40,  120,    60,  20,      0,   255,    0,    D_EXIT,              0,    0,   "Cancel",NULL,NULL},
  
  { NULL,              0,    0,    0,   0,      0,     0,    0,         0,              0,    0,    NULL ,NULL,NULL}
};
#define PULLASLIDER 2
#define PULLATEXT 3

#define PULLCANCEL 7

int pullaslidercallback(void *dp3 UNUSED, int d2)
{
        int obj;
        DIALOG *dtext=&pull_dialog[PULLATEXT];
        pull_a=.001*(float)(d2)/(float)PULLASLIDERLENGTH;
        sprintf(pullastring,"%1.3f",1000*pull_a);
        rectfill(screen,dtext->x,dtext->y, dtext->x + dtext->w,dtext->y + dtext->h,dtext->bg);
        dialog_message(dtext, MSG_DRAW,0, &obj);
        return(D_O_K);
}
int do_pull_dialog(){
    int popupret;
    float holdpull_a=pull_a;
    sprintf(pullastring,"%1.3f",1000*pull_a);
    sprintf(maxnedgestring,"%d",maxnedge);
    pull_dialog[PULLASLIDER].d2=(int)(1/.001*(pull_a*PULLASLIDERLENGTH));
    centre_dialog(pull_dialog);
    popupret=popup_dialog(pull_dialog,2);
    if(popupret==PULLCANCEL){
       pull_a=holdpull_a;

    }
    maxnedge=abs(atoi(maxnedgestring));
    return D_O_K;
}

int perturbingslidercallback(void *dp3 UNUSED, int d2);
char perturbingstring[80];
#define PERTURBINGSLIDERLENGTH 99

DIALOG perturb_dialog[]={
/* (dialog proc)     (x)   (y)   (w)   (h)     (fg)  (bg)  (key)    (flags)           (d1)  (d2)  (dp) */
  { d_box_proc,        0,   0,   150, 130,      0,   DBG,    0,         0,              0,    0,   NULL,NULL,NULL},
  

  { d_text_proc,      10,  22,     0,   0,      0,   DBG,    0,         0,              0,    0,    "Roughening:" ,NULL,NULL},
  { d_slider_proc,    10,  32,   110,  10,      0,   DBG,    0,         0, PERTURBINGSLIDERLENGTH,    0,    NULL,     perturbingslidercallback,NULL},
  { d_text_proc,      70,  44,    60,  10,      0,   DBG,    0,         0,              0,    0,    perturbingstring ,NULL,NULL},

  { d_check_proc,     10,  70,    150,  12,      0,   DBG,   's',         0,              0,    0,    "only Selected" ,NULL,NULL},
  { d_button_proc,    10,  100,    20,  20,      0,   255,   13,    D_EXIT,              0,    0,   "OK",NULL,NULL},
  { d_button_proc,    40,  100,    60,  20,      0,   255,    0,    D_EXIT,              0,    0,   "Cancel",NULL,NULL},
  
  { NULL,              0,    0,    0,   0,      0,     0,    0,         0,              0,    0,    NULL ,NULL,NULL}
};
#define PERTURBINGSLIDER 2
#define PERTURBINGTEXT 3
#define PERTURBSELECTEDCHECK 4
#define PERTURBINGCANCEL 6

int perturbingslidercallback(void *dp3 UNUSED, int d2)
{
        int obj;
        DIALOG *dtext=&perturb_dialog[PERTURBINGTEXT];
        perturbing=(float)(d2)/(float)PERTURBINGSLIDERLENGTH;
        sprintf(perturbingstring,"%1.3f",perturbing);
        rectfill(screen,dtext->x,dtext->y, dtext->x + dtext->w,dtext->y + dtext->h,dtext->bg);
        dialog_message(dtext, MSG_DRAW,0, &obj);
        return(D_O_K);
}
int do_perturb_dialog(){
    int popupret;
    float holdperturbing=perturbing;
    sprintf(perturbingstring,"%1.3f",perturbing);
    perturb_dialog[PERTURBINGSLIDER].d2=(int)(perturbing*PERTURBINGSLIDERLENGTH);
    centre_dialog(perturb_dialog);
    popupret=popup_dialog(perturb_dialog,2);
    if(popupret==PERTURBINGCANCEL){
       perturbing=holdperturbing;
       return 0;
    }
    perturb_selected_p=perturb_dialog[PERTURBSELECTEDCHECK].flags&D_SELECTED;
    return 1;
}

int reduceslidercallback(void *dp3 UNUSED, int d2);
char reducestring[80];
#define REDUCESLIDERLENGTH 99

DIALOG reduce_dialog[]={
/* (dialog proc)     (x)   (y)   (w)   (h)     (fg)  (bg)  (key)    (flags)           (d1)  (d2)  (dp) */
  { d_box_proc,        0,   0,   240, 120,      0,   DBG,    0,         0,              0,    0,   NULL,NULL,NULL},
  

  { d_text_proc,      10,  22,     0,   0,      0,   DBG,    0,         0,              0,    0,    "Fraction:" ,NULL,NULL},
  { d_slider_proc,    10,  32,   110,  10,      0,   DBG,    0,         0, REDUCESLIDERLENGTH,    0,    NULL,     reduceslidercallback,NULL},
  { d_text_proc,      70,  44,    60,  10,      0,   DBG,    0,         0,              0,    0,    reducestring ,NULL,NULL},

  { d_check_proc,     10,  60,    200,  12,      0,   DBG,   's',         0,              0,    0,    "apply only to Selected" ,NULL,NULL},

  { d_button_proc,    10,  90,    20,  20,      0,   255,   13,    D_EXIT,              0,    0,   "OK",NULL,NULL},
  { d_button_proc,    40,  90,    60,  20,      0,   255,    0,    D_EXIT,              0,    0,   "Cancel",NULL,NULL},
  
  { NULL,              0,    0,    0,   0,      0,     0,    0,         0,              0,    0,    NULL ,NULL,NULL}
};
#define REDUCESLIDER 2
#define REDUCETEXT 3
#define REDUCESELECTEDCHECK 4
#define REDUCECANCEL 6

int reduceslidercallback(void *dp3 UNUSED, int d2)
{
        int obj;
        DIALOG *dtext=&reduce_dialog[REDUCETEXT];
        reducepercent=.9*(float)(d2)/(float)REDUCESLIDERLENGTH;
        sprintf(reducestring,"%1.3f",reducepercent);
        rectfill(screen,dtext->x,dtext->y, dtext->x + dtext->w,dtext->y + dtext->h,dtext->bg);
        dialog_message(dtext, MSG_DRAW,0, &obj);
        return(D_O_K);
}

int do_reduce_dialog(){
    int popupret;
    float holdreducepercent=reducepercent;
    sprintf(reducestring,"%1.3f",reducepercent);
    reduce_dialog[REDUCESLIDER].d2=1/.9*(int)(reducepercent*REDUCESLIDERLENGTH);
    centre_dialog(reduce_dialog);
    popupret=popup_dialog(reduce_dialog,2);
    if(popupret==REDUCECANCEL){
       reducepercent=holdreducepercent;
       return 0;
    }
    reduce_selected_p=reduce_dialog[REDUCESELECTEDCHECK].flags&D_SELECTED;
    return 1;
}

int meld_coincident_selected_p=0;
float coincident_epsilon=0;
int do_meld_coincident_dialog();
int meld_all_coincident_vertices();
int meld_selected_coincident_vertices();

int do_meld_coincident(){
    int obj;


    if(do_meld_coincident_dialog()){
       save_undo();
       waiting_call();
       if(meld_coincident_selected_p)
         meld_selected_coincident_vertices();
       else
         meld_all_coincident_vertices();
       
       waiting_done();

       update_numbers();
       redraw_flatviews();
       dialog_message(&the_dialog[THREEDPROC], MSG_DRAW, 0, &obj);
    }
    return D_O_K;
}

int coincidentepilonslidercallback(void *dp3 UNUSED, int d2);
char coincidentepsilonstring[80];

#define COINCIDENTEPSILONSLIDERLENGTH 99
DIALOG meld_coincident_dialog[]={
/* (dialog proc)     (x)   (y)   (w)   (h)     (fg)  (bg)  (key)    (flags)           (d1)  (d2)  (dp) */
  { d_box_proc,        0,   0,   180, 130,      0,   DBG,    0,         0,              0,    0,   NULL,NULL,NULL},

  { d_text_proc,      10,  22,     0,   0,      0,   DBG,    0,         0,              0,    0,    "epsilon:" ,NULL,NULL},
  { d_slider_proc,    10,  32,   110,  10,      0,   DBG,    0,         0, COINCIDENTEPSILONSLIDERLENGTH,    0,    NULL,    coincidentepilonslidercallback,NULL},
  { d_text_proc,      70,  44,    60,  10,      0,   DBG,    0,         0,              0,    0,    coincidentepsilonstring ,NULL,NULL},

  { d_check_proc,     10,  70,    170,  12,      0,   DBG,   's',         0,              0,    0,    "only to Selected" ,NULL,NULL},
  { d_button_proc,    10,  100,    20,  20,      0,   255,   13,    D_EXIT,              0,    0,   "OK",NULL,NULL},
  { d_button_proc,    40,  100,    60,  20,      0,   255,    0,    D_EXIT,              0,    0,   "Cancel",NULL,NULL},
  
  { NULL,              0,    0,    0,   0,      0,     0,    0,         0,              0,    0,    NULL ,NULL,NULL}
};
#define MELDCOINCIDENTSLIDER 2
#define MELDCOINCIDENTTEXT 3
#define MELDCOINCIDENTSELECTEDCHECK 4
#define MELDCOINCIDENTCANCEL 6

int coincidentepilonslidercallback(void *dp3 UNUSED, int d2)
{
        int obj;
        DIALOG *dtext=&meld_coincident_dialog[MELDCOINCIDENTTEXT];
        coincident_epsilon=10.0*(float)(d2)/(float)COINCIDENTEPSILONSLIDERLENGTH;
        sprintf(coincidentepsilonstring,"%1.3f",coincident_epsilon);
        rectfill(screen,dtext->x,dtext->y, dtext->x + dtext->w,dtext->y + dtext->h,dtext->bg);
        dialog_message(dtext, MSG_DRAW,0, &obj);
        return(D_O_K);
}

int do_meld_coincident_dialog(){
    int popupret;
    float holdcoincident_epsilon=coincident_epsilon;
    sprintf(coincidentepsilonstring,"%1.3f",coincident_epsilon);
    meld_coincident_dialog[MELDCOINCIDENTSLIDER].d2=(int)(coincident_epsilon/10.0*COINCIDENTEPSILONSLIDERLENGTH);
    centre_dialog(meld_coincident_dialog);
    popupret=popup_dialog(meld_coincident_dialog,2);
    if(popupret==MELDCOINCIDENTCANCEL){
       coincident_epsilon=holdcoincident_epsilon;
       return 0;
    }
    meld_coincident_selected_p=meld_coincident_dialog[MELDCOINCIDENTSELECTEDCHECK].flags&D_SELECTED;
    return 1;
}

int meld_all_coincident_vertices(){
    int vi=0,vj,ti;
    int oktomeld;
    TRIANGLE *T;
    float dist2;
    float eps=coincident_epsilon*coincident_epsilon;
    float vix,viy,viz,vjx,vjy,vjz;
    while (vi<mesh->num_vertices){
          if(esc_pressed()){
               waiting_done();
               break;
          }
          vix=mesh->points[vi].x;viy=mesh->points[vi].y;viz=mesh->points[vi].z;
          vj=vi+1;
          while(vj<mesh->num_vertices){
              oktomeld=1;
              vjx=mesh->points[vj].x;vjy=mesh->points[vj].y;vjz=mesh->points[vj].z;
              dist2=(vix-vjx)*(vix-vjx)+(viy-vjy)*(viy-vjy)+(viz-vjz)*(viz-vjz);
              if(dist2<=eps){
                  for(ti=0;ti<mesh->num_faces;ti++){
                    T=&(mesh->faces[ti]);

                    if(find_edge_in_triangle(T,vi,vj)){
                          oktomeld=0;break;
                    }
                  }
                  if(oktomeld)
                     meld_vertex_pair(vi,vj);
                  else vj++;

              }
              else{
                 vj++;
              }
          }
          vi++;
    }
    return 1;
}
int meld_selected_coincident_vertices(){
    int si=0,vi,vj,ti;
    int oktomeld;
    float dist2;
    float eps=coincident_epsilon*coincident_epsilon;
    float vix,viy,viz,vjx,vjy,vjz;
    TRIANGLE *T;
    while (si<num_selected){
          if(esc_pressed()){
               waiting_done();
               break;
          }
          vi=curVertex[si];
          vix=mesh->points[vi].x;viy=mesh->points[vi].y;viz=mesh->points[vi].z;
          vj=0;
          while(vj<mesh->num_vertices){
              if(vi==vj) {vj++;continue;}
              vjx=mesh->points[vj].x;vjy=mesh->points[vj].y;vjz=mesh->points[vj].z;
              dist2=(vix-vjx)*(vix-vjx)+(viy-vjy)*(viy-vjy)+(viz-vjz)*(viz-vjz);
              oktomeld=1;
              if(dist2<=eps){
                  for(ti=0;ti<mesh->num_faces;ti++){
                    T=&(mesh->faces[ti]);

                    if(find_edge_in_triangle(T,vi,vj)){
                          oktomeld=0;break;
                    }
                  }
                  if(oktomeld)
                     meld_vertex_pair(vi,vj);
                  else vj++;
              }
              else{
                 vj++;
              }
          }
          si++;
    }
    return 1;
}

/* misc. utility functions */
int dump_screen(){
          BITMAP *bmp;
          PALETTE pal;
          
          get_palette(pal);
          bmp = create_sub_bitmap(screen, 0, 0, SCREEN_W, SCREEN_H);
          save_bitmap("dump.pcx", bmp, pal);
          destroy_bitmap(bmp);
          return D_O_K;

}

int error_exit(char *message1,char *message2){
        if(waiting){
            waiting_done();
        }
        alert(message1,message2,"Exiting program...","OK",NULL,13,0);
        allegro_exit();
        if(message1)
           printf("\n%s\n",message1);
        if(message2)
           printf("%s\n",message2);
        exit(0);
}

int esc_pressed(){
     return (keypressed()&& ((readkey() >> 8) == KEY_ESC));
}


void set_dialog_view_sizes(int s_w,int s_h){
     int BUT_A_L;
     
     SCREENWIDE=s_w;
     SCREENHIGH=s_h;
     /* size of the mesh view windows */
     VIEWS_WIDTH  = (SCREENHIGH/2-1);
     VIEWS_HEIGHT = (SCREENHIGH/2-1);

     the_dialog[MAINLAYOUT].w=SCREENWIDE;
     the_dialog[MAINLAYOUT].h=SCREENHIGH;
     the_dialog[XYPROC].x=0;the_dialog[XYPROC].y=0;
     the_dialog[XYPROC].w=VIEWS_WIDTH; the_dialog[XYPROC].h=VIEWS_HEIGHT;
     the_dialog[ZYPROC].x=VIEWS_WIDTH+2;the_dialog[ZYPROC].y=0;
     the_dialog[ZYPROC].w=VIEWS_WIDTH; the_dialog[ZYPROC].h=VIEWS_HEIGHT;
     the_dialog[XZPROC].x=0;the_dialog[XZPROC].y=VIEWS_HEIGHT+2;
     the_dialog[XZPROC].w=VIEWS_WIDTH; the_dialog[XZPROC].h=VIEWS_HEIGHT;
     the_dialog[THREEDPROC].x=VIEWS_WIDTH+2;the_dialog[THREEDPROC].y=VIEWS_HEIGHT+2;
     the_dialog[THREEDPROC].w=VIEWS_WIDTH; the_dialog[THREEDPROC].h=VIEWS_HEIGHT;
     the_dialog[RIGHTBIGBOX].x=2*(VIEWS_WIDTH+1);the_dialog[RIGHTBIGBOX].y=0;
     the_dialog[RIGHTBIGBOX].w=SCREENWIDE-2*(VIEWS_WIDTH+1);the_dialog[RIGHTBIGBOX].h=SCREENHIGH;

     /* left edge of 'button area' */
     BUT_A_L = 2*(VIEWS_WIDTH+1)+10;
     
     the_dialog[NUMVTXTEXT].x=BUT_A_L;
     the_dialog[NUMTRITEXT].x=BUT_A_L;
     the_dialog[NUMSELTEXT].x=BUT_A_L;
     the_dialog[MOUSETEXT].x=BUT_A_L;
     the_dialog[TRISELBMP].x=BUT_A_L;
     the_dialog[MAINMENU].x=2*(VIEWS_WIDTH+1);
     the_dialog[POLYGONBUTTON].x=BUT_A_L+60;
     the_dialog[DEFORMBUTTON].x=BUT_A_L+60;
     the_dialog[REDRAWBUTTON].x=BUT_A_L+60;

}
int do_grey_shading(){
           int obj;
           ShadingPaletteBottom=GreyScaleBottom;
           ShadingPaletteTop=GreyScaleTop;
           dialog_message(&the_dialog[THREEDPROC], MSG_DRAW, 0, &obj);
           return D_O_K;
}
int do_green_shading(){
           int obj;
           ShadingPaletteBottom=GreenScaleBottom;
           ShadingPaletteTop=GreenScaleTop;
           dialog_message(&the_dialog[THREEDPROC], MSG_DRAW, 0, &obj);
           return D_O_K;
}
int do_red_shading(){
           int obj;
           ShadingPaletteBottom=RedScaleBottom;
           ShadingPaletteTop=RedScaleTop;
           dialog_message(&the_dialog[THREEDPROC], MSG_DRAW, 0, &obj);
           return D_O_K;
}


MENU shading_palette_menu[]={
   {"&Grey",          do_grey_shading,    NULL,0,NULL},

   {"&gree&N",        do_green_shading,    NULL,0,NULL},

   {"&Red",           do_red_shading,    NULL,0,NULL},
   {NULL,NULL,NULL,0,NULL}
};
#define GREYITEM 0
#define GREENITEM 1
#define REDITEM 2


int do_display_resolution_menu(){
   int s_w,s_h;
   int old_s_w=SCREENWIDE,old_s_h=SCREENHIGH;
   int mret=do_menu(display_resolution_menu,mouse_x,mouse_y);
   switch(mret){
       case RES640x480:
           s_w=640;s_h=480;
       break;
       case RES800x600:
           s_w=800;s_h=600;
       break;
       case RES1024x768:
           s_w=1024;s_h=768;
       break;
       case RES1280x1024:
           s_w=1280;s_h=1024;
       break;
       case RES1600x1200:
           s_w=1600;s_h=1200;
       break;
       
       default:
           s_w=640;s_h=480;
       
   }
   set_dialog_view_sizes(s_w,s_h);

   if (set_gfx_mode(GFX_AUTODETECT, SCREENWIDE, SCREENHIGH, 0, 0) != 0) {
      set_dialog_view_sizes(old_s_w,old_s_h);
      if (set_gfx_mode(GFX_AUTODETECT, SCREENWIDE, SCREENHIGH, 0, 0) != 0) {
         allegro_exit();
         printf("Error setting graphics mode\n%s\n\n", allegro_error);
         exit(0);
      }

      alert("Error setting graphics mode",NULL, NULL, "OK",NULL, 13, 0);
   }
   set_palette(allegrodata[THEPALETTE].dat);
   destroy_bitmap(threedbmp);
   destroy_bitmap(xybmp);
   destroy_bitmap(zybmp);
   destroy_bitmap(xzbmp);
   
   threedbmp = create_bitmap(VIEWS_WIDTH, VIEWS_HEIGHT);
   xybmp     = create_bitmap(VIEWS_WIDTH, VIEWS_HEIGHT);
   zybmp     = create_bitmap(VIEWS_WIDTH, VIEWS_HEIGHT);
   xzbmp     = create_bitmap(VIEWS_WIDTH, VIEWS_HEIGHT);

   if(threedbmp==NULL||xybmp==NULL||zybmp==NULL||xzbmp==NULL){
      error_exit("memory allocation error for bitmaps",NULL);
   }
   the_dialog[XYPROC].dp=xybmp;
   the_dialog[ZYPROC].dp=zybmp;
   the_dialog[XZPROC].dp=xzbmp;
   the_dialog[THREEDPROC].dp=threedbmp;

   
   set_projection_viewport(0, 0, VIEWS_WIDTH, VIEWS_HEIGHT);
   gui_fg_color=0;
   gui_bg_color=220;

   show_mouse(screen);
   return D_REDRAW;
}



